/*******************************************************************************

    File:  analyse.c
    Authors: Stephanie Warrick, Pete Yule, Rick Cooper
    Contents: Routines to score an action sequence

*******************************************************************************/

#include "cs.h"
#include "analyse.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define MaxBuffer 256

/******************************************************************************/
/* For debugging: *************************************************************/

void fprint_action_list(FILE *fp, ActionList *tmp)
{
    fprintf(fp, "Actions:\n");
    while (tmp != NULL) {
        fprintf(fp, "%d: %s", tmp->stepNumber, tmp->name);
        fprintf(fp, " %s", ((tmp->argument1 == NULL) ? "NULL" : tmp->argument1));
        fprintf(fp, " %s", ((tmp->argument2 == NULL) ? "NULL" : tmp->argument2));
        fprintf(fp, " %s", ((tmp->argument3 == NULL) ? "NULL" : tmp->argument3));
        fprintf(fp, "\n");
        tmp = tmp->next;
    }
}

/******************************************************************************/
/* Formatted output for action-related errors/warnings: ***********************/

static void write_action(FILE *fp, ActionList *action)
{
    fprintf(fp, "%s", action->name);
    if (action->argument1 != NULL) {
        fprintf(fp, "(%s", action->argument1);
        if (action->argument2 != NULL) {
            fprintf(fp, ", %s", action->argument2);
        }
        if (action->argument3 != NULL) {
            fprintf(fp, ", %s", action->argument3);
        }
        fprintf(fp, ")");
    }
}

void report_unrecognised_object(FILE *fp, char *name)
{
    fprintf(fp, "WARNING: Object not recognised: %s\n", name);
    abort();
}

void report_unrecognised_action(FILE *fp, ActionList *action)
{
    fprintf(fp, "WARNING: Unrecognised action: ");
    write_action(fp, action);
    fprintf(fp, "\n");
    abort();
}

/******************************************************************************/

static Boolean argcmp(char *a, char *b)
{
    if ((a == NULL) || (b == NULL)) {
        return(a == b);
    }
    else {
        return(strcmp(a, b) == 0);
    }
}	

void ignore_consecutive_actions(ActionList *this)
{
    ActionList *next = this->next;

    while ((next != NULL) && (strcmp(this->name, next->name) == 0)) {
        if (!argcmp(this->argument1, next->argument1)) {
            break;
        }
        else if (!argcmp(this->argument2, next->argument2)) {
            break;
        }
        else if (!argcmp(this->argument3, next->argument3)) {
            break;
        }
        else {
            next->checked = TRUE;
            next = next->next;
        }
    }    
}

void score_consecutive_actions_as_perseverations(DataAnalysis *la, ActionList *this)
{
    ActionList *next = this->next;
    Boolean perseveration = FALSE;

    while ((next != NULL) && (!next->checked) && (strcmp(this->name, next->name) == 0)) {
        if (!argcmp(this->argument1, next->argument1)) {
            break;
        }
        else if (!argcmp(this->argument2, next->argument2)) {
            break;
        }
        else if (!argcmp(this->argument3, next->argument3)) {
            break;
        }
        else {
            perseveration = TRUE;
            next->checked = TRUE;
            next = next->next;
        }
    }    
    if (perseveration) {
        analyseErrorLogAppend(la, ERROR_PERSEVERATION, this->name, this->argument1, this->argument2);
    }
}

/******************************************************************************/
/* The analysis routine for the null task: ************************************/
void analyse_null_task(DataAnalysis *la)
{
}

/*******************************************************************************
** Read action file and install actions: ***************************************
********************************************************************************

The input file consists of a series of lines of the form:
 
         action arg1 [arg2] [arg3]
 
and is terminated by a line of the form

         { complete | incomplete } 2035 

where 2035 (e.g.) is the cycle count of the last step

*******************************************************************************/

static char *allocate_copy(char *buffer, int *end)
{
    char *string = NULL;
    int i = 0;

    if ((buffer[i] != '\0') && (buffer[i] != '\n') && (buffer[i] != ' ')) {
        do {
            i++;
        } while ((buffer[i] != '\0') && (buffer[i] != '\n') && (buffer[i] != ' '));
        buffer[i++] = '\0';
        *end += i;
        if ((string = (char *)malloc(sizeof(char) * i)) != NULL) {
            strcpy(string, buffer);
        }
    }
    return(string);
}

/*----------------------------------------------------------------------------*/

static ActionList *create_action(char *buffer, int n)
{
    ActionList *new = (ActionList *)malloc(sizeof(ActionList));
    if (new != NULL) {
        int j = 0;
        new->checked = FALSE;
        new->stepNumber = n;
        new->name = allocate_copy(&buffer[j], &j);
        new->argument1 = allocate_copy(&buffer[j], &j);
        new->argument2 = allocate_copy(&buffer[j], &j);
        new->argument3 = allocate_copy(&buffer[j], &j);
        new->next = NULL;
    }
    return(new);
}

/*----------------------------------------------------------------------------*/

LineList *free_line_list(LineList *ll)
{
    LineList *tmp;

    while ((tmp = ll) != NULL) {
        if (ll->line != NULL) { free(ll->line); }
        ll = ll->next;
        free(tmp);
    }
    return(ll);
}

/*----------------------------------------------------------------------------*/

static ActionList *free_action_list(ActionList *al)
{
    ActionList *tmp;

    while ((tmp = al) != NULL) {
        if (al->name != NULL) { free(al->name); }
        if (al->argument1 != NULL) { free(al->argument1); }
        if (al->argument2 != NULL) { free(al->argument2); }
        if (al->argument3 != NULL) { free(al->argument3); }
        al = al->next;
        free(tmp);
    }
    return(al);
}

static void disposeActionSequence(DataAnalysis *la)
{
    la->actionSequence = free_action_list(la->actionSequence);
}

/*----------------------------------------------------------------------------*/

static void append_action(DataAnalysis *la, ActionList *new)
{
    if (la->actionSequence == NULL) {
        la->actionSequence = new;
    }
    else {
        ActionList *tmp = la->actionSequence;
        while (tmp->next != NULL) {
            tmp = tmp->next;
        }
        tmp->next = new;
    }
}

/*----------------------------------------------------------------------------*/

static void installActionSequence(FILE *fp, DataAnalysis *la)
{
    char buffer[MaxBuffer];
    ActionList *new;
    int n = 0;

    la->numActions = 0;
    la->numAccomplished = 0;
    la->numActionAdditions = 0;
    la->numActionAnticipations = 0;
    la->numStepOmissions = 0;
    la->numPerseverations = 0;
    la->numToolOmissions = 0;
    la->numMisUse1 = 0;
    la->numMisUse2 = 0;
    la->numMisLocation1 = 0;
    la->numMisLocation2 = 0;
    la->numObjectSubstitution = 0;
    la->numToying = 0;
    la->numPantomime = 0;
    la->numPerplexity = 0;
    disposeActionSequence(la);
    la->log = free_line_list(la->log);
    la->numCycles = 0;
    la->propCommission = 0.0;
    la->numErrors = 0;
    la->numInterventions = 0;
    la->complete = FALSE;

    while (fgets(buffer, MaxBuffer, fp) != NULL) {
        if (strchr(buffer, '\n') == NULL) {
            printf("STRING BUFFER OVERFLOW: %s\n", buffer);
            printf("STRING BUFFER OVERFLOW: Action ignored\n");
        }
        else if (strncmp(buffer, "complete", 8) == 0) {
            la->numCycles = (int) atol(&buffer[9]);
            la->complete = TRUE;
        }
        else if (strncmp(buffer, "incomplete", 10) == 0) {
            la->numCycles = (int) atol(&buffer[10]);
            la->complete = FALSE;
        }
        else if (strncmp(buffer, "intervention", 12) == 0) {
            la->numInterventions++;
        }
        else if ((new = create_action(buffer, n++)) != NULL) {
            append_action(la, new);
            la->numActions++;
        }
    }
}

/******************************************************************************/
    
static void set_commission_prop(DataAnalysis *la)
{
    double commit = la->numActionAnticipations + la->numPerseverations + la->numActionAdditions;
    double conceptual = la->numMisUse1 + la->numMisUse2 + la->numMisLocation1 + la->numMisLocation2 + la->numObjectSubstitution + la->numToolOmissions + la->numPantomime + la->numPerplexity + la->numToying;
    double omit = la->numStepOmissions;

    la->numErrors = commit + conceptual + omit;
    la->propCommission = ((la->numErrors > 0) ? ((commit + conceptual) / la->numErrors) : 0.0);
}

static void quantise_dvs(DataAnalysis *la)
{
    /* For the New scoring system (as at Mon Nov 19 15:47:33 2001):     */
    /* Just score absence/presence for most dvs...                      */

    la->numAccomplished = (la->numAccomplished * 100) / 6;

    if (la->numActionAdditions > 0) { la->numActionAdditions = 1; }
    if (la->numActionAnticipations > 0) { la->numActionAnticipations = 1; }
    if (la->numStepOmissions > 0) { la->numStepOmissions = 1; }
    if (la->numPerseverations > 0) { la->numPerseverations = 1; }
    if (la->numToolOmissions > 0) { la->numToolOmissions = 1; }
    if (la->numMisUse1 > 0) { la->numMisUse1 = 1; }
    if (la->numMisUse2 > 0) { la->numMisUse2 = 1; }
    if (la->numMisLocation1 > 0) { la->numMisLocation1 = 1; }
    if (la->numMisLocation2 > 0) { la->numMisLocation2 = 1; }
    if (la->numObjectSubstitution > 0) { la->numObjectSubstitution = 1; }
}

/******************************************************************************/
/* Entry Point 1: *************************************************************/

Boolean analyse_results(char *filename, TaskType *task, DataAnalysis *la)
{
    FILE *fp;

    if ((fp = fopen(filename, "r")) == NULL) {
        fprintf(stderr, "INPUT FILE NOT FOUND: %s missing!\n", filename);
        return(FALSE);
    }
    else {
        fprintf(stdout, "PROCESSING: %s ----------------------------------------\n", filename);
        installActionSequence(fp, la);
        fclose(fp);
        task->analyse(la);
        set_commission_prop(la);
        if (FALSE) {             /* quantise_dvs only if we want to */
            quantise_dvs(la);    /* score the percentage of trials  */
        }                        /* showing each error type         */
        disposeActionSequence(la);
        la->actionSequence = free_action_list(la->actionSequence);
        return(TRUE);
    }
}

/******************************************************************************/
/* Entry Point 2: *************************************************************/

void analyseErrorLogAppend(DataAnalysis *la, ErrorType et, char *action, char *arg1, char *arg2)
{
    char buffer[256];
    char *error  = NULL;
    LineList *tmp;

    switch (et) {
        case ERROR_ACTION_ADDITION:
            { error = "Action Addition"; la->numActionAdditions++; break; }
        case ERROR_ACTION_ANTICIPATION:
            { error = "Anticipation"; la->numActionAnticipations++; break; }
        case ERROR_STEP_OMISSION:
            { error = "Omission"; la->numStepOmissions++; break; }
        case ERROR_PERSEVERATION:
            { error = "Perseveration"; la->numPerseverations++; break; }
        case ERROR_TOOL_OMISSION:
            { error = "Tool Omission"; la->numToolOmissions++; break; }
        case ERROR_MISUSE_1:
            { error = "Misuse (type 1)"; la->numMisUse1++; break; }
        case ERROR_MISUSE_2:
            { error = "Misuse (type 2)"; la->numMisUse2++; break; }
        case ERROR_MISLOCATION_1:
            { error = "Mislocation (type 1)"; la->numMisLocation1++; break; }
        case ERROR_MISLOCATION_2:
            { error = "Mislocation (type 2)"; la->numMisLocation2++; break; }
        case ERROR_OBJECT_SUBSTITUTION:
            { error = "Object Substitution"; la->numObjectSubstitution++; break; }
        case ERROR_TOYING:
            { error = "Toying"; la->numToying++; break; }
        case ERROR_PERPLEXITY:
            { error = "Perplexity"; la->numPerplexity++; break; }
        case ERROR_PANTOMIME:
            { error = "Pantomime"; la->numPantomime++; break; }
    }

    if (arg1 == NULL) {
        g_snprintf(buffer, 256, "%s: %s.", error, action);
    }
    else if (arg2 == NULL) {
        g_snprintf(buffer, 256, "%s: %s(%s).", error, action, arg1);
    }
    else {
        g_snprintf(buffer, 256, "%s: %s(%s, %s).", error, action, arg1, arg2);
    }

    if ((tmp = (LineList *)malloc(sizeof(LineList))) != NULL) {
        LineList *prev, *next;
        if ((tmp->line = (char *)malloc(sizeof(char) * (strlen(buffer) + 1))) != NULL) {
            // Copy the string:
            strcpy(tmp->line, buffer);
            // Now insert the line in alphanumeric order

            if (la->log == NULL) {
                la->log = tmp;
                tmp->next = NULL;
            }
#ifdef ALPHA_ORDERING
            else if (strcmp(tmp->line, la->log->line) < 0) {
                tmp->next = la->log;
                la->log = tmp;
            }
            else {
                prev = la->log;
                next = prev->next;
                while ((next != NULL) && (strcmp(tmp->line, next->line) > 0)) {
                    prev = next;
                    next = prev->next;
                }
                prev->next = tmp;
                tmp->next = next;
            }
#else
            else {
                prev = la->log;
                next = prev->next;
                while (next != NULL) {
                    prev = next;
                    next = prev->next;
                }
                prev->next = tmp;
                tmp->next = next;
            }
#endif
        }
        else {
            free(tmp);
        }
    }
}

/******************************************************************************/
