/*******************************************************************************

    File: bottle/action.c
    AUTHOR: R. Cooper
    CREATION DATE: Wed Aug  4 12:33:37 2004
    CONTENTS: Actions for the bottle task

*******************************************************************************/

#include "../cs.h"
#include "task.h"

/******************************************************************************/

/* A local buffer used for actionMessage:                                     */
static char buffer[128]; 

/******************************************************************************/

void uncork_bottle(ContentionScheduler *cs, Effector *opener_hand)
{
    /* Should be holding the bottle opener and fixated on the bottle. */

    Object *bottle = cs->fixated_object;
    Object *opener = effectorGetContents(opener_hand);
    Object *cork = objectGetAttachment(bottle);

    char *bottle_name = (bottle == NULL ? "nothing" : bottle->name);
    char *opener_name = (opener == NULL ? "nothing" : opener->name);
    char *opener_hand_name = (opener_hand == NULL ? "nothing" : opener_hand->name);

    g_snprintf(buffer, 128, "Uncorking %s with %s (held by %s)", bottle_name, opener_name, opener_hand_name);
    actionMessage(cs, "UncorkBottle", buffer);
    record_action(cs, "UncorkBottle", bottle_name, opener_name, opener_hand_name);

    if (opener_hand == NULL) {
        errorEffectorNull(cs, "UncorkBottle");
    }
    else if (opener == NULL) {
        errorEffectorState(cs, "UncorkBottle: hand (%s) is empty", opener_hand->name);
    }
    else if (isBottleOpener(opener) < 0.9) {
        errorObjectIncorrect(cs, "%s: implement (%s) is not a bottle operner", "UncorkBottle", opener->name, NULL);
    }
    else if (isBottle(bottle) < 0.9) {
        errorObjectIncorrect(cs, "%s: target (%s) is not a bottle", "UncorkBottle", bottle->name, NULL);
    }
    else if (isOpenBottle(bottle) > 0.9) {
        errorObjectIncorrect(cs, "%s: target (%s) is already uncorked", "UncorkBottle", bottle->name, NULL);
    }
    else if (cork == NULL) {
        errorObjectIncorrect(cs, "%s: target (%s) has no attachement!", "UncorkBottle", bottle->name, NULL);
    }
    else if (isCork(cork) < 0.9) {
        errorObjectIncorrect(cs, "%s: attachment to target is (%s) is not a cork!", "UncorkBottle", cork->name, NULL);
    }
    else {
        traceObjectPreAction(cs, bottle);
        bottle->state = (bottle->state | STATE_OPEN) & ~STATE_CLOSED;
        traceObjectPostAction(cs, bottle);

        traceObjectPreAction(cs, cork);
        objectDeleteAttached(bottle, cork);
        traceObjectPostAction(cs, cork);
    }
}

/******************************************************************************/

void recork_bottle(ContentionScheduler *cs, Effector *hand)
{
    /* Should be holding the cork and fixated on the bottle. */

    Object *bottle = cs->fixated_object;
    Object *cork = effectorGetContents(hand);

    char *bottle_name = (bottle == NULL ? "nothing" : bottle->name);
    char *cork_name = (cork == NULL ? "nothing" : cork->name);
    char *hand_name = (hand == NULL ? "nothing" : hand->name);

    g_snprintf(buffer, 128, "Recorking %s with %s (held by %s)", bottle_name, cork_name, hand_name);
    actionMessage(cs, "RecorkBottle", buffer);
    record_action(cs, "RecorkBottle", bottle_name, cork_name, hand_name);

    if (hand == NULL) {
        errorEffectorNull(cs, "RecorkBottle");
    }
    else if (cork == NULL) {
        errorEffectorState(cs, "RecorkBottle: hand (%s) is empty", hand->name);
    }
    else if (isCork(cork) < 0.9) {
        errorObjectIncorrect(cs, "%s: %s is not a cork", "RecorkBottle", cork->name, NULL);
    }
    else if (isBottle(bottle) < 0.9) {
        errorObjectIncorrect(cs, "%s: target (%s) is not a bottle", "RecorkBottle", bottle->name, NULL);
    }
    else if (isClosedBottle(bottle) > 0.9) {
        errorObjectIncorrect(cs, "%s: target (%s) is already corked", "RecorkBottle", bottle->name, NULL);
    }
    else {
        traceObjectPreAction(cs, bottle);
        bottle->state = (bottle->state | STATE_CLOSED) & ~STATE_OPEN;
        traceObjectPostAction(cs, bottle);

        traceObjectPreAction(cs, cork);
        hand->itemsHeld = objectListDeleteElement(hand->itemsHeld, cork);
        cork->holder = NULL;
        objectInstallAttachment(bottle, cork);
        traceObjectPostAction(cs, cork);
    }
}

/******************************************************************************/
/********************** drink(HandContainingObject) ***************************/

/* When a container's contents are drunk we assume that only some of the      */
/* fluid disappears. This version drinks from all objects in the given        */
/* effector.                                                                  */

void drink(ContentionScheduler *cs, Effector *effector)
{
    Object *theme = effectorGetContents(effector);
    Object *fluid = objectGetContent(theme);

    char *fluid_name = (fluid == NULL ? "nothing" : fluid->name);
    char *theme_name = (theme == NULL ? "nothing" : theme->name);
    char *effector_name = (effector == NULL ? "nothing" : effector->name);

    g_snprintf(buffer, 128, "Drinking %s from %s (held by %s)", fluid_name, theme_name, effector_name);
    actionMessage(cs, "Drink", buffer);
    record_action(cs, "Drink", fluid_name, theme_name, effector_name);

    if (effector == NULL) {
        errorEffectorNull(cs, "Drink");
    }
    else if (theme == NULL) {
        errorEffectorState(cs, "Drink: effector (%s) is empty", effector->name);
    }
    else if (fluid == NULL) {
        errorEffectorState(cs, "Drink: container (%s) is empty", theme->name);
    }
    else if (!isFluidBoolean(fluid)) {
        errorEffectorState(cs, "Drink: Contents of %s are not drinkable", theme->name);
    }
    else {
        traceObjectPreAction(cs, theme);
        theme->state = theme->state | STATE_USED;
        traceObjectPostAction(cs, theme);
    }
}

/******************************************************************************/
