/*******************************************************************************

    FILE: bottle/analyse.c
    AUTHOR: R. Cooper
    CREATION DATE: Fri Aug 13 12:41:35 2004
    CONTENTS: Analysis routine for the bottle task

*******************************************************************************/

#include "../cs.h"
#include "../analyse.h"

/******************************************************************************/

typedef enum bottle_object {
    OBJECT_NONE, OBJECT_GLASS, OBJECT_BOTTLE, OBJECT_CORK, OBJECT_OPENER,
    OBJECT_WATER, OBJECT_TABLE, OBJECT_UNKNOWN
} BottleObject;

typedef enum bottle_location {
    LOC_NONE, LOC_HELD, LOC_ON_TABLE, LOC_ON_BOTTLE
} BottleLocation;

typedef struct bottle_analysis_object_list {
    BottleObject                        id;        /* Enumerated id */
    BottleLocation                      location;
    int                                 state;
    struct bottle_analysis_object_list *next;      /* Next object in the list  */
} BottleAnalysisObjectList;

typedef struct bottle_state {
    // Information on all possible objects in the world
    BottleAnalysisObjectList *items; 

    //Fixated item:
    BottleObject              fixated;

    // Other relevant state information:
    Boolean                  glass_filled;
    Boolean                  drink_drunk;
    Boolean                  bottle_recorked;

} BottleState;

#define BOTTLE_STATE_CLOSED            1
#define BOTTLE_STATE_CONTAINS_FLUID    2

/******************************************************************************/

static BottleObject object_id_from_string(char *arg)
{
    if (arg == NULL) {
        return(OBJECT_NONE);
    }
    else if (strcmp(arg, "nothing") == 0) {
        return(OBJECT_NONE);
    }
    else if (strcmp(arg, "glass") == 0) {
        return(OBJECT_GLASS);
    }
    else if (strcmp(arg, "bottle") == 0) {
        return(OBJECT_BOTTLE);
    }
    else if (strcmp(arg, "cork") == 0) {
        return(OBJECT_CORK);
    }
    else if (strcmp(arg, "bottle_opener") == 0) {
        return(OBJECT_OPENER);
    }
    else if (strcmp(arg, "water") == 0) {
        return(OBJECT_WATER);
    }
    else if (strcmp(arg, "table") == 0) {
        return(OBJECT_TABLE);
    }
    else {    /* Final catch-all case: */
        report_unrecognised_object(stdout, arg);
        return(OBJECT_UNKNOWN);
    }
}

static BottleAnalysisObjectList *object_ptr_from_id(BottleState *state, BottleObject id)
{
    BottleAnalysisObjectList *tmp = state->items;

    while ((tmp != NULL) && (tmp->id != id)) {
        tmp = tmp->next;
    }
    if ((tmp == NULL) || (tmp->id != id)) {
        return(NULL);
    }
    else {
        return(tmp);
    }
}

static BottleAnalysisObjectList *object_ptr_from_string(BottleState *state, char *arg)
{
    return(object_ptr_from_id(state, object_id_from_string(arg)));
}

static Boolean action_compare(ActionList *this, char *action, BottleObject arg1)
{
    if (this == NULL) {
        return(FALSE);
    }
    else if (strcmp(this->name, action) != 0) {
        return(FALSE);
    }
    else if (object_id_from_string(this->argument1) != arg1) {
        return(FALSE);
    }
    else {
        return(TRUE);
    }
}

/******************************************************************************/

static void state_add_item(BottleState *state, BottleObject id, BottleLocation loc, int item_state)
{
    BottleAnalysisObjectList *new = (BottleAnalysisObjectList *)malloc(sizeof(BottleAnalysisObjectList));

    if (new != NULL) {
        new->id = id;
        new->location = loc;
        new->state = item_state;
        new->next = state->items;
        state->items = new;
    }
}

static void state_initialise(BottleState *state)
{
    state->items = NULL;

    /* All objects must be initialised to match bottle/object.c */

    state_add_item(state, OBJECT_GLASS, LOC_ON_TABLE, 0);
    state_add_item(state, OBJECT_BOTTLE, LOC_ON_TABLE, BOTTLE_STATE_CONTAINS_FLUID);
    state_add_item(state, OBJECT_CORK, LOC_ON_BOTTLE, 0);
    state_add_item(state, OBJECT_OPENER, LOC_ON_TABLE, 0);
    state_add_item(state, OBJECT_WATER, LOC_NONE, 0);
    state_add_item(state, OBJECT_TABLE, LOC_NONE, 0);

    state->fixated = OBJECT_NONE;

    state->glass_filled = FALSE;
    state->drink_drunk = FALSE;
    state->bottle_recorked = FALSE;
}

static void state_destroy(BottleState *state)
{
    while (state->items != NULL) {
        BottleAnalysisObjectList *tmp = state->items;
        state->items = tmp->next;
        free(tmp);
    }
}

/******************************************************************************/

static void check_consecutive_toying_actions(ActionList *this, BottleObject arg1)
{
    if (action_compare(this->next, "PickUp", arg1)) {
        if (action_compare(this->next->next, "PutDown", arg1)) {
            this->next->checked = TRUE;
            this->next->next->checked = TRUE;
            check_consecutive_toying_actions(this->next->next, arg1);
        }
        else if (action_compare(this->next->next, "Discard", arg1)) {
            this->next->checked = TRUE;
            this->next->next->checked = TRUE;
            check_consecutive_toying_actions(this->next->next, arg1);
        }
    }
}

/******************************************************************************/

static void score_action_fixate(BottleState *state, DataAnalysis *la, ActionList *action, char *arg1)
{
    BottleAnalysisObjectList *item1;

    if (object_id_from_string(arg1) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "Fixate", arg1, NULL);
        ignore_consecutive_actions(action);
    }
    else if ((item1 = object_ptr_from_string(state, arg1)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else {
        state->fixated = item1->id;
    }
}

/*----------------------------------------------------------------------------*/

static void score_action_pickup(BottleState *state, DataAnalysis *la, ActionList *action, char *arg1)
{
    BottleAnalysisObjectList *item1;

    if (object_id_from_string(arg1) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "PickUp", arg1, NULL);
        ignore_consecutive_actions(action);
    }
    else if ((item1 = object_ptr_from_string(state, arg1)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if (item1->location == LOC_HELD) {
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "PickUp", arg1, NULL);
        ignore_consecutive_actions(action);
    }
    else if (item1->id == OBJECT_TABLE) {
        // Score PERPLEXITY: Object cannot be picked up
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "PickUp", arg1, NULL);
        ignore_consecutive_actions(action);
    }
    else {
        item1->location = LOC_HELD;
        if (action_compare(action->next, "PutDown", item1->id)) {
            // Score TOYING error and ignore any consecutive toying
            analyseErrorLogAppend(la, ERROR_TOYING, "PickUp/PutDown", arg1, NULL);
            check_consecutive_toying_actions(action->next, item1->id);
        }
        else if (action_compare(action->next, "Discard", item1->id)) {
            // Score TOYING error and ignore any consecutive toying
            analyseErrorLogAppend(la, ERROR_TOYING, "PickUp/Discard", arg1, NULL);
            check_consecutive_toying_actions(action->next, item1->id);
        }
    }
}

/*----------------------------------------------------------------------------*/

static void score_action_putdown(BottleState *state, DataAnalysis *la, ActionList *action, char *arg1)
{
    BottleAnalysisObjectList *item1;

    if (object_id_from_string(arg1) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "PutDown", arg1, NULL);
        ignore_consecutive_actions(action);
    }
    else if ((item1 = object_ptr_from_string(state, arg1)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if (item1->location != LOC_HELD) {
        // Score PERPLEXITY: Object is not held
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "PutDown", arg1, NULL);
        ignore_consecutive_actions(action);
    }
    else {
        item1->location = LOC_ON_TABLE;
    }
}

/*----------------------------------------------------------------------------*/

static void score_action_discard(BottleState *state, DataAnalysis *la, ActionList *action, char *arg1)
{
    BottleAnalysisObjectList *item1;

    if (object_id_from_string(arg1) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "Discard", arg1, NULL);
        ignore_consecutive_actions(action);
    }
    else if ((item1 = object_ptr_from_string(state, arg1)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if (item1->location != LOC_HELD) {
        // Score PERPLEXITY: Object is not held
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "Discard", arg1, NULL);
        ignore_consecutive_actions(action);
    }
    else {
        item1->location = LOC_ON_TABLE;
    }
}

/*----------------------------------------------------------------------------*/

static void score_action_uncork_bottle(BottleState *state, DataAnalysis *la, ActionList *action, char *arg1, char *arg2)
{
    BottleAnalysisObjectList *item1;     // The container
    BottleAnalysisObjectList *item2;     // The opener
    BottleAnalysisObjectList *cork = object_ptr_from_id(state, OBJECT_CORK);

    if (object_id_from_string(arg1) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PANTOMIME, "UncorkBottle", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if ((item1 = object_ptr_from_string(state, arg1)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if (object_id_from_string(arg2) == OBJECT_NONE) {
        if (item1->id == OBJECT_BOTTLE) {
            // Score TOOL OMISSION: opening without a tool!
            analyseErrorLogAppend(la, ERROR_TOOL_OMISSION, "UncorkBottle", arg1, arg2);

            item1->state = item1->state & ~BOTTLE_STATE_CLOSED;
            if (cork != NULL) {
                cork->location = LOC_ON_TABLE;
            }
        }
        else {
            analyseErrorLogAppend(la, ERROR_MISUSE_1, "UncorkBottle", arg1, arg2);
        }
    }
    else if ((item2 = object_ptr_from_string(state, arg2)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if (item1 == item2) {
        // Score PERPLEXITY: trying to pour container into itself
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "UncorkBottle01", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if (item2->location != LOC_HELD) {
        // Score PERPLEXITY: opener isn't held
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "UncorkBottle02", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if (item1->id != OBJECT_BOTTLE) {
        // Score PERPLEXITY: container isn't the bottle
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "UncorkBottle03", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if ((item1->state & BOTTLE_STATE_CLOSED) != 0) {
        // Score PERSEVERATION: Bottle is already open
        analyseErrorLogAppend(la, ERROR_PERSEVERATION, "UncorkBottle04", arg1, arg2);
    }
    else {
        if (item2->id != OBJECT_OPENER) {
            // Score MISUSE 1: "opener" isn't an opener
            analyseErrorLogAppend(la, ERROR_MISUSE_1, "UncorkBottle06", arg1, arg2);
        }
        item1->state = item1->state & ~BOTTLE_STATE_CLOSED;
        if (cork != NULL) {
            cork->location = LOC_ON_TABLE;
        }
    }
}

/*----------------------------------------------------------------------------*/

static void score_action_recork_bottle(BottleState *state, DataAnalysis *la, ActionList *action, char *arg1, char *arg2)
{
    BottleAnalysisObjectList *item1;     // The container
    BottleAnalysisObjectList *item2;     // The cork

    if (object_id_from_string(arg1) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "RecorkBottle01", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if ((item1 = object_ptr_from_string(state, arg1)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if (object_id_from_string(arg2) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PANTOMIME, "RecorkBottle02", arg1, arg2);
    }
    else if ((item2 = object_ptr_from_string(state, arg2)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if (item1 == item2) {
        // Score PERPLEXITY: trying to pour container into itself
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "RecorkBottle03", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if (item2->location != LOC_HELD) {
        // Score PERPLEXITY: cork isn't held
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "RecorkBottle04", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if (item1->id != OBJECT_BOTTLE) {
        // Score MISUSE 1: trying to cork non-bottle
        analyseErrorLogAppend(la, ERROR_MISUSE_1, "RecorkBottle05", arg1, arg2);
    }
    else if (item1->state & BOTTLE_STATE_CLOSED) {
        // Score PERSEVERATION: Bottle is already closed
        analyseErrorLogAppend(la, ERROR_PERSEVERATION, "RecorkBottle06", arg1, arg2);
    }
    else if (item2->id == OBJECT_OPENER) {
        // Score MISUSE 2: trying to cork with opener -- would be ok if we were uncorking!
        analyseErrorLogAppend(la, ERROR_MISUSE_2, "RecorkBottle07", arg1, arg2);
    }
    else if (item2->id != OBJECT_CORK) {
        // Score PERPLEXITY: trying to cork with non-cork
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "RecorkBottle08", arg1, arg2);
    }
    else {
        item2->location = LOC_ON_BOTTLE;
        item1->state = item1->state | BOTTLE_STATE_CLOSED;
        state->bottle_recorked = TRUE;
    }
}

/*----------------------------------------------------------------------------*/

static void score_action_pour(BottleState *state, DataAnalysis *la, ActionList *action, char *arg1, char *arg2)
{
    BottleAnalysisObjectList *item1;     // The source container
    BottleAnalysisObjectList *item2;     // The target container

    if (object_id_from_string(arg1) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PANTOMIME, "PourSome00", arg1, arg2);
    }
    else if ((item1 = object_ptr_from_string(state, arg1)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if ((item2 = object_ptr_from_string(state, arg2)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if (item1 == item2) {
        // Score PERPLEXITY: trying to pour container into itself
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "PourSome01", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if (item1->location != LOC_HELD) {
        // Score PERPLEXITY: source container isn't held
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "PourSome02", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if (item1->state & BOTTLE_STATE_CLOSED) {
        // Score ANTICIPATION error: attempting to pour from closed container
        analyseErrorLogAppend(la, ERROR_ACTION_ANTICIPATION, "PourSome03", arg1, arg2);
    }
    else if ((item1->state & BOTTLE_STATE_CONTAINS_FLUID) == 0) {
        if ((item1->id == OBJECT_BOTTLE) || (item1->id == OBJECT_GLASS)) {
            // Score ANTICIPATION error: sensible source but it is empty!
            analyseErrorLogAppend(la, ERROR_ACTION_ANTICIPATION, "PourSome04", arg1, arg2);
        }
        else {
            // Score MISUSE_1 error:  Pouring bizarre object
            analyseErrorLogAppend(la, ERROR_MISUSE_1, "PourSome06", arg1, arg2);
        }
    }
    else {
        if (item2->id != OBJECT_GLASS) {
            // Score MISLOCATION1 error: target isn't the glass
            analyseErrorLogAppend(la, ERROR_MISLOCATION_1, "PourSome07", arg1, arg2);
        }
        else {
            state->glass_filled = TRUE;
        }
        item2->state = item2->state | BOTTLE_STATE_CONTAINS_FLUID;
    }
}

/*----------------------------------------------------------------------------*/

static void score_action_drink(BottleState *state, DataAnalysis *la, ActionList *action, char *arg1, char *arg2)
{
    BottleAnalysisObjectList *item2;    // The container

    if (object_id_from_string(arg2) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PANTOMIME, "Drink00", arg1, arg2);
    }
    else if ((item2 = object_ptr_from_string(state, arg2)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if (item2->location != LOC_HELD) {
        // Score PERPLEXITY: trying to drink from a container that isn't held
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "Drink01", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if ((item2->state & BOTTLE_STATE_CONTAINS_FLUID) == 0) {
        if ((item2->id == OBJECT_GLASS) || (item2->id == OBJECT_BOTTLE)) {
            // Score ANTICIPATION error: Drinking from sensible but empty container!
            analyseErrorLogAppend(la, ERROR_ACTION_ANTICIPATION, "Drink02", arg1, arg2);
        }
        else {
            // Score MISUSE_1 error: Attempting to drink from a non-container!
            analyseErrorLogAppend(la, ERROR_MISUSE_1, "Drink03", arg1, arg2);
        }
    }
    else if (item2->state & BOTTLE_STATE_CLOSED) {
        // Score ANTICIPATION error: Drinking from closed container
        analyseErrorLogAppend(la, ERROR_ACTION_ANTICIPATION, "Drink04", arg1, arg2);
    }
    else {
        if (item2->id != OBJECT_GLASS) {
            // Score MISUSE_2: Drinking from some container other than the glass
            analyseErrorLogAppend(la, ERROR_MISUSE_2, "Drink05", arg1, arg2);
        }
        state->drink_drunk = TRUE;
    }
}

/******************************************************************************/

static void score_action(BottleState *state, DataAnalysis *la, ActionList *action)
{
    if (action == NULL) {
        fprintf(stderr, "INTERNAL ERROR: NULL action in score_action/3 (file is bottle/analyse.c)\n");
    }
    else if (strcmp(action->name, "Fixate") == 0) {
        score_action_fixate(state, la, action, action->argument1);
    }
    else if (strcmp(action->name, "PickUp") == 0) {
        score_action_pickup(state, la, action, action->argument1);
    }
    else if (strcmp(action->name, "PutDown") == 0) {
        score_action_putdown(state, la, action, action->argument1);
    }
    else if (strcmp(action->name, "Discard") == 0) {
        score_action_discard(state, la, action, action->argument1);
    }
    else if (strcmp(action->name, "UncorkBottle") == 0) {
        score_action_uncork_bottle(state, la, action, action->argument1, action->argument2);
    }
    else if (strcmp(action->name, "RecorkBottle") == 0) {
        score_action_recork_bottle(state, la, action, action->argument1, action->argument2);
    }
    else if (strcmp(action->name, "PourSome") == 0) {
        score_action_pour(state, la, action, action->argument1, action->argument2);
    }
    else if (strcmp(action->name, "Drink") == 0) {
        score_action_drink(state, la, action, action->argument1, action->argument2);
    }
    else {
        report_unrecognised_action(stdout, action);
    }
}

/******************************************************************************/

static void score_omissions(BottleState *state, DataAnalysis *la)
{
    if (!state->glass_filled) {
        analyseErrorLogAppend(la, ERROR_STEP_OMISSION, "FillGlass", "glass", "bottle");
    }
    if (!state->drink_drunk) {
        analyseErrorLogAppend(la, ERROR_STEP_OMISSION, "DrinkSomething", "glass", NULL);
    }
    if (!state->bottle_recorked) {
        analyseErrorLogAppend(la, ERROR_STEP_OMISSION, "RecorkBottle", "bottle", "cork");
    }
}

/******************************************************************************/

void analyse_bottle_task(DataAnalysis *la)
{
    ActionList *as;
    BottleState state;

    state_initialise(&state);

    for (as = la->actionSequence; as; as = as->next) {
        as->checked = FALSE;
    }

    for (as = la->actionSequence; as; as = as->next) {
        if (!as->checked) {
            score_action(&state, la, as);
            as->checked = TRUE;
        }
    }

    score_omissions(&state, la);

    state_destroy(&state);
}

/******************************************************************************/
