/*******************************************************************************

    File: candle/action.c
    AUTHOR: R. Cooper
    CREATION DATE: Tue Aug  3 13:46:53 2004
    CONTENTS: Actions for the candle task

*******************************************************************************/

#include "../cs.h"
#include "task.h"

/******************************************************************************/

/* A local buffer used for actionMessage:                                     */
static char buffer[128]; 

extern Object *head_of_match(Object *match_base);

/******************************************************************************/

void candle_into_holder(ContentionScheduler *cs, Effector *hand)
{
    Object *target = cs->fixated_object;
    Object *candle = effectorGetContents(hand);

    char *target_name = (target == NULL ? "nothing" : target->name);
    char *candle_name = (candle == NULL ? "nothing" : candle->name);
    char *hand_name = (hand == NULL ? "nothing" : hand->name);

    g_snprintf(buffer, 128, "Inserting %s (held by %s) into %s", candle_name, hand_name, target_name);
    actionMessage(cs, "CandleIntoHolder", buffer);
    record_action(cs, "CandleIntoHolder", candle_name, target_name, hand_name);

    if (hand == NULL) {
        errorEffectorNull(cs, "CandleIntoHolder");
    }
    else if (candle == NULL) {
        errorEffectorState(cs, "CandleIntoHolder: hand (%s) is empty", hand->name);
    }
    else if (isCandle(candle) < 0.9) {
        errorObjectIncorrect(cs, "%s: hand (%s) holds incorrect object %s", "CandleIntoHolder", hand->name, candle->name);
    }
    else if (isCandleStick(target) < 0.9) {
        errorObjectIncorrect(cs, "%s: target (%s) is incorrect", "CandleIntoHolder", target->name, NULL);
    }
    else {

        if (candle->base != NULL) {
            /* We seem to be putting the candle in upsidedown. Rejig the    */
            /* representation so the base is attached to the top to prevent */
            /* crashing later when we reinitialise.                         */
            Object *bottom = candle->base;

            objectDeleteAttached(bottom, candle);
            objectInstallAttachment(candle, bottom);
        }

        traceObjectPreAction(cs, candle);
        candle->base = target;
        traceObjectPostAction(cs, candle);

        traceObjectPreAction(cs, target);
        objectInstallAttachment(target, candle);
        traceObjectPostAction(cs, target);

        traceEffectorPreAction(cs, hand);
        hand->itemsHeld = objectListDeleteElement(hand->itemsHeld, candle);
        candle->holder = NULL;
        traceEffectorPostAction(cs, hand);
    }
}

/*----------------------------------------------------------------------------*/

void slide_open(ContentionScheduler *cs, Effector *boxHand)
{
    /* We assume that the match box is held by one hand and can be opened   */
    /* even if there is something in the other hand.                        */

    Object *box = effectorGetContents(boxHand);

    char *box_name = (box == NULL ? "nothing" : box->name);
    char *hand_name = (boxHand == NULL ? "nothing" : boxHand->name);

    g_snprintf(buffer, 128, "Sliding open %s (held by %s)", box_name, hand_name);
    actionMessage(cs, "SlideOpen", buffer);
    record_action(cs, "SlideOpen", box_name, hand_name, NULL);

    if (boxHand == NULL) {
        errorEffectorNull(cs, "SlideOpen");
    }
    else if ((box = effectorGetContents(boxHand)) == NULL) {
        errorEffectorState(cs, "SlideOpen: hand (%s) is empty", boxHand->name);
    }
    else if ((isMatchBox(box) < 0.9) && (isMatchBoxDrawer(box) < 0.9)) {
        errorObjectIncorrect(cs, "%s: hand (%s) holds incorrect object %s", "SlideOpen", boxHand->name, box->name);
    }
    else if (isOpenBoolean(box)) {
        errorObjectIncorrect(cs, "%s: box (%s) is already open", "SlideOpen", box->name, NULL);
    }
    else {
        Object *drawer = NULL;
        ObjectList *themes;
        ObjectRole role;

        /* Possibly we're opening by the drawer, instead of the box. If so  */
        /* reset the object pointer to point to the box:                    */
        if (isMatchBoxDrawer(box) > 0.9) {
            box = box->base;
        }

        traceObjectPreAction(cs, box);
        box->state = (box->state | STATE_OPEN) & ~STATE_CLOSED;
        traceObjectPostAction(cs, box);

        if ((drawer = objectGetAttachment(box)) != NULL) {

            traceObjectPreAction(cs, drawer);
            drawer->state = (drawer->state | STATE_OPEN) & ~STATE_CLOSED;
            traceObjectPostAction(cs, drawer);


            /* Move all contents of the box's drawer to the primary object list */
            for (themes = drawer->itemsContained; themes != NULL; themes = themes->tail) {
                cs->objects = objectListAppendElement(cs->objects, themes->head);
                themes->head->time0 = cs->cycle;
                for (role = 0; role < (MaxRoles-1) ; role++) {
                    themes->head->activation[role][cs->cycle] = initial_activation(OBJECT);
                }
            }
        }
    }
}

/*----------------------------------------------------------------------------*/

void slide_closed(ContentionScheduler *cs, Effector *boxHand)
{
    /* We assume that the match box is held by one hand and can be closed   */
    /* even if there is something in the other hand.                        */

    Object *box = effectorGetContents(boxHand);

    char *box_name = (box == NULL ? "nothing" : box->name);
    char *hand_name = (boxHand == NULL ? "nothing" : boxHand->name);

    g_snprintf(buffer, 128, "Sliding closed %s (held by %s)", box_name, hand_name);
    actionMessage(cs, "SlideClosed", buffer);
    record_action(cs, "SlideClosed", box_name, hand_name, NULL);

    if (boxHand == NULL) {
        errorEffectorNull(cs, "SlideClosed");
    }
    else if (box == NULL) {
        errorEffectorState(cs, "SlideClosed: hand (%s) is empty", boxHand->name);
    }
    else if ((isMatchBox(box) < 0.9) && (isMatchBoxDrawer(box) < 0.9)) {
        errorObjectIncorrect(cs, "%s: hand (%s) holds incorrect object %s", "SlideClosed", boxHand->name, box->name);
    }
    else if (isClosedBoolean(box)) {
        errorObjectIncorrect(cs, "%s: box (%s) is already closed", "SlideClosed", box->name, NULL);
    }
    else {
        ObjectList *themes;
        Object *drawer = NULL;

        /* Possibly we're closing by the drawer, instead of the box. If so  */
        /* reset the object pointer to point to the box:                    */
        if (isMatchBoxDrawer(box) > 0.9) {
            box = box->base;
        }

        traceObjectPreAction(cs, box);
        box->state = (box->state | STATE_CLOSED) & ~STATE_OPEN;
        traceObjectPostAction(cs, box);

        if ((drawer = objectGetAttachment(box)) != NULL) {

            traceObjectPreAction(cs, drawer);
            drawer->state = (drawer->state | STATE_CLOSED) & ~STATE_OPEN;
            traceObjectPostAction(cs, drawer);

            /* Remove all contents of the box's drawer from the primary object list */
            for (themes = drawer->itemsContained; themes != NULL; themes = themes->tail) {
                cs->objects = objectListDeleteElement(cs->objects, themes->head);
            }
        }
    }
}

/*----------------------------------------------------------------------------*/

void tear_open(ContentionScheduler *cs, Effector *boxHand)
{
    Object *box = effectorGetContents(boxHand);

    char *box_name = (box == NULL ? "nothing" : box->name);
    char *hand_name = (boxHand == NULL ? "nothing" : boxHand->name);

    g_snprintf(buffer, 128, "Tearing open %s (held by %s)", box_name, hand_name);
    actionMessage(cs, "TearOpen", buffer);
    record_action(cs, "TearOpen", box_name, hand_name, NULL);
}

/*----------------------------------------------------------------------------*/

void peel_open(ContentionScheduler *cs, Effector *boxHand)
{
    Object *box = effectorGetContents(boxHand);

    char *box_name = (box == NULL ? "nothing" : box->name);
    char *hand_name = (boxHand == NULL ? "nothing" : boxHand->name);

    g_snprintf(buffer, 128, "Peeling open %s (held by %s)", box_name, hand_name);
    actionMessage(cs, "PeelOpen", buffer);
    record_action(cs, "PeelOpen", box_name, hand_name, NULL);
}

/*----------------------------------------------------------------------------*/

void strike_match(ContentionScheduler *cs, Effector *matchHand)
{
    Object *box = cs->fixated_object;
    Object *match_base = effectorGetContents(matchHand);
    Object *match_head = head_of_match(match_base);

    char *box_name = (box == NULL ? "nothing" : box->name);
    char *match_base_name = (match_base == NULL ? "nothing" : match_base->name);

    g_snprintf(buffer, 128, "Striking %s against %s", match_base_name, box_name);
    actionMessage(cs, "StrikeMatch", buffer);
    record_action(cs, "StrikeMatch", match_base_name, box_name, NULL);

    if (matchHand == NULL) {
        errorEffectorNull(cs, "StrikeMatch");
    }
    else if ((match_base = effectorGetContents(matchHand)) == NULL) {
        errorEffectorState(cs, "StrikeMatch: hand (%s) is empty", matchHand->name);
    }
    else if (isMatch(match_base) < 0.9) {
        errorObjectState(cs, "%s: %s isn't a match!", "StrikeMatch", match_base->name, NULL);
    }
    else if (match_head == NULL) {
        errorObjectState(cs, "%s: striking wrong end of match (%s)!", "StrikeMatch", match_base->name, NULL);
    }
    else if (isUsedMatch(match_head) > 0.9) {
        errorObjectState(cs, "%s: match (%s) is used", "StrikeMatch", match_head->name, NULL);
    }
    else if (isLitMatch(match_head) > 0.9) {
        errorObjectState(cs, "%s: match (%s) is already alight", "StrikeMatch", match_head->name, NULL);
    }
    else if (isMatchBox(box) < 0.9) {
        errorObjectState(cs, "%s: %s isn't a match box!", "StrikeMatch", box->name, NULL);
    }
    else {
        traceObjectPreAction(cs, match_head);
        match_head->state = match_head->state | STATE_LIT;
        traceObjectPostAction(cs, match_head);
    }
}

/*----------------------------------------------------------------------------*/

void match_to_candle(ContentionScheduler *cs, Effector *matchHand)
{
    Object *candle = cs->fixated_object;
    Object *match_base = effectorGetContents(matchHand);
    Object *match_head = head_of_match(match_base);

    char *candle_name = (candle == NULL ? "nothing" : candle->name);
    char *match_base_name = (match_base == NULL ? "nothing" : match_base->name);

    g_snprintf(buffer, 128, "Moving %s to %s", match_base_name, candle_name);
    actionMessage(cs, "MatchToCandle", buffer);
    record_action(cs, "MatchToCandle", match_base_name, candle_name, NULL);

    if (matchHand == NULL) {
        errorEffectorNull(cs, "MatchToCandle");
    }
    else if (match_base == NULL) {
        errorEffectorState(cs, "MatchToCandle: hand (%s) is empty", matchHand->name);
    }
    else if (candle == NULL) {
        errorObjectNull(cs, "Target", "MatchToCandle");
    }
    else if (isMatch(match_base) < 0.9) {
        errorObjectState(cs, "%s: %s isn't a match!", "MatchToCandle", match_base->name, NULL);
    }
    else if (match_head == NULL) {
        errorObjectState(cs, "%s: holding wrong end of match (%s)!", "MatchToCandle", match_base->name, NULL);
    }
    else if (isLitCandle(candle) > 0.9) {
        errorObjectState(cs, "%s: candle (%s) is already alight", "MatchToCandle", candle->name, NULL);
    }
    else if (isLitMatch(match_head) < 0.9) {
        errorObjectState(cs, "%s: match (%s) is not alight", "MatchToCandle", match_base->name, NULL);
    }
     else {
        traceObjectPreAction(cs, candle);
        candle->state = candle->state | STATE_LIT;
        traceObjectPostAction(cs, candle);
    }
}

/*----------------------------------------------------------------------------*/

void blow_out_match(ContentionScheduler *cs)
{
    Object *match = cs->fixated_object;

    char *match_name = (match == NULL ? "nothing" : match->name);

    g_snprintf(buffer, 128, "Blowing out %s", match_name);
    actionMessage(cs, "BlowOutFlame", buffer);
    record_action(cs, "BlowOutFlame", match_name, NULL, NULL);

    if (match == NULL) {
        errorObjectNull(cs, "Implement", "BlowOutMatch");
    }
    else if (isLit(match) < 0.9) {
        errorObjectState(cs, "%s: %s is not alight", "BlowOutMatch", match->name, NULL);
    }
    else {
        traceObjectPreAction(cs, match);
        match->state = (match->state & ~STATE_LIT) | STATE_USED;
        traceObjectPostAction(cs, match);
    }
}

/******************************************************************************/

