/*******************************************************************************

    FILE: candle/object.c
    AUTHOR: R. Cooper
    CREATION DATE: Tue Aug  3 13:46:21 2004
    CONTENTS: Object-related routines for the candle task

*******************************************************************************/

#include "../cs.h"
#include "task.h"

extern double feature_overlap(int af, int bf);

/******************************************************************************/

double isLit(Object *o)
{
    if (o->state & STATE_LIT) {
        return(1.0);
    }
    else if ((o->base != NULL) && (o->base->state & STATE_LIT)) {
        return(1.0);
    }
    else {
        return(0.0);
    }
}

/******************************************************************************/

double isCandleBase(Object *o)
{
    return(feature_overlap(o->features, IS_CYLINDRICAL | IS_EXTENDED | IS_MEDIUM | HOLESIZE_1));
}

double isCandleTop(Object *o)
{
    return(feature_overlap(o->features, IS_CYLINDRICAL | IS_EXTENDED | IS_MEDIUM | HAS_WICK));
}

double isCandle(Object *o)
{
    return(FZOR(isCandleBase(o), isCandleTop(o)));
}

double isMatch(Object *o)
{
    return(FZOR(isMatchBase(o), isMatchHead(o)));
}

double isMatchBase(Object *o)
{
    return(feature_overlap(o->features, IS_TOOL | IS_EXTENDED | IS_SMALL | IS_HANDLE));
}

double isMatchHead(Object *o)
{
    return(feature_overlap(o->features, IS_TOOL | IS_EXTENDED | IS_SMALL));
}

double isMatchBox(Object *o)
{
    return(feature_overlap(o->features, IS_CONTAINER | HAS_DRAWER | IS_MEDIUM | IS_SQUARE));
}

double isMatchBoxDrawer(Object *o)
{
    return(feature_overlap(o->features, IS_CONTAINER | IS_DRAWER | IS_MEDIUM | IS_SQUARE));
}

double isCandleStick(Object *o)
{
    return(feature_overlap(o->features, IS_METAL | IS_MEDIUM | HOLESIZE_1));
}

/******************************************************************************/

Object *head_of_match(Object *match_base)
{
    if (match_base == NULL) {
        return(NULL);
    }
    else if (isMatchBase(match_base)) {
        return(objectGetAttachment(match_base));
    }
    else {
        return(NULL);
    }
}

/******************************************************************************/

double isOpenMatchBox(Object *o)
{
    return(FZAND(isMatchBox(o), isOpen(o)));
}

double isClosedMatchBox(Object *o)
{
    return(FZAND(isMatchBox(o), FZNOT(isOpen(o))));
}

double isUsedMatch(Object *o)
{
    return(FZAND(isUsed(o), isMatch(o)));
}

double isFullCandleStick(Object *o)
{
    return(FZAND(isCandleStick(o), isFullObject(o)));
}

double isLitMatch(Object *o)
{
    return(FZAND(isLit(o), isMatch(o)));
}

double isLitCandle(Object *o)
{
    return(FZAND(isLit(o), isCandle(o)));
}

/******************************************************************************/

void write_candle_object_attributes(FILE *fp, Object *o)
{
    if (o->state & STATE_OPEN) { 
        fprintf(fp, "open, ");
    }
    else if (o->state & STATE_CLOSED) { 
        fprintf(fp, "closed, ");
    }

    if (o->state & STATE_LIT) { 
        fprintf(fp, "lit, ");
    }

    if (o->state & STATE_USED) { 
        fprintf(fp, "used, ");
    }

    if (o->itemsAttached) {
        fprintf(fp, "has attachments, ");
    }
}

/******************************************************************************/

void install_candle_basic_objects(ContentionScheduler *cs)
{
    Object *object, *drawer;
    Object *match_base, *match_top;
    char buffer[24];
    int i;

    object = objectInstall(cs, "candle_base", IS_EXTENDED | IS_CYLINDRICAL | IS_MEDIUM | HOLESIZE_1, 0);
    objectSetPosition(object,  3, 3);
    objectInstallAttachment(object, objectInstall(cs, "candle_top", IS_EXTENDED | IS_CYLINDRICAL | IS_MEDIUM | HAS_WICK, 0));

    object = objectInstall(cs, "candle_stick", IS_METAL | HOLESIZE_1 | IS_MEDIUM, 0);
    objectSetPosition(object,  0, 3);

    object = objectInstall(cs, "matchbox", IS_CONTAINER | HAS_DRAWER | IS_MEDIUM | IS_SQUARE, STATE_CLOSED);
    objectSetPosition(object, -3, 3);

    drawer = objectInstall(cs, "matchbox_drawer", IS_CONTAINER | IS_DRAWER | IS_MEDIUM | IS_SQUARE, STATE_CLOSED);
    objectInstallAttachment(object, drawer);

    for (i = 0; i < 20; i++) {
        g_snprintf(buffer, 22, "match_stick_%d", i);
        match_base = objectInstallItem(cs, buffer, IS_TOOL | IS_EXTENDED | IS_SMALL | IS_HANDLE);
        g_snprintf(buffer, 22, "match_stick_head_%d", i);
        match_top = objectInstallItem(cs, buffer, IS_TOOL | IS_EXTENDED | IS_SMALL);
        objectInstallAttachment(match_base, match_top);
        objectInstallContents(drawer, match_base);
        objectInstallContents(drawer, match_top);
    }

    object = objectInstall(cs, "table", 0, 0);
}

/******************************************************************************/
