/*******************************************************************************

    FILE: effector.c
    AUTHOR: R. Cooper (translated from C++ written by SVW)
    CREATION DATE: 
    CONTENTS: 

*******************************************************************************/

#include "cs.h"

/******************************************************************************/

static void initialise_effector_activations(ContentionScheduler *cs)
{
    EffectorList *el;

    for (el = cs->effectors; el != NULL; el = el->tail) {
        el->head->activation[cs->cycle] = initial_activation(EFFECTOR);
    }

    if (maths == NORMALISED) {
        normalise_effector_activations(cs);
    }
}

void destroy_effectors(ContentionScheduler *cs)
{
    while (cs->effectors != NULL) {
        EffectorList *tmp = cs->effectors->tail;

        while (cs->effectors->head->itemsHeld != NULL) {
            ObjectList *tmp = cs->effectors->head->itemsHeld;
            cs->effectors->head->itemsHeld = tmp->tail;
            free(tmp);
        }
        free(cs->effectors->head);
        free(cs->effectors);
        cs->effectors = tmp;
    }
}

void initialise_effectors(ContentionScheduler *cs)
{
    cs->task.install_effectors(cs);
    initialise_effector_activations(cs);
}

void reinitialise_effectors(ContentionScheduler *cs)
{
    destroy_effectors(cs);
    initialise_effectors(cs);
}

/******************************************************************************/

Effector *effectorInstall(ContentionScheduler *cs, char *effector_name)
{
    EffectorList *el;

    if ((el = (EffectorList *)malloc(sizeof(EffectorList))) == NULL) {
	fprintf(stderr, "MEMORY ALLOCATION FAILURE: install_effector.1\n");
    }
    else if ((el->head = (Effector *)malloc(sizeof(Effector))) == NULL) {
	fprintf(stderr, "MEMORY ALLOCATION FAILURE: install_effector.2\n");
        free(el);
    }
    else {
        el->head->name = effector_name;
        el->head->highlighted = FALSE;
        el->head->itemsHeld = NULL;
        el->tail = cs->effectors;
        cs->effectors = el;
    }
    return(el->head);
}

/******************************************************************************/
/* Display functions: *********************************************************/

void traceEffectorPreAction(ContentionScheduler *cs, Effector *this)
{
    if (messages[WORLD] != NULL) {
        fprintf(messages[WORLD], "%5d\t%s: [", cs->cycle, this->name);
	cs->task.write_effector_attributes(messages[WORLD], this);
	fprintf(messages[WORLD], "]");
	fflush(messages[WORLD]);
    }
}

/*----------------------------------------------------------------------------*/

void traceEffectorPostAction(ContentionScheduler *cs, Effector *this)
{
    if (messages[WORLD] != NULL) {
	fprintf(messages[WORLD], " --> [");
	cs->task.write_effector_attributes(messages[WORLD], this);
	fprintf(messages[WORLD], "]\n");
    }
}

/******************************************************************************/

double isEmptyEffectorFuzzy(Effector *this)
{
    return((this->itemsHeld == NULL) ? 1.0 : 0.0);
}

double isFullEffectorFuzzy(Effector *this)
{
    return((this->itemsHeld != NULL) ? 1.0 : 0.0);
}

Boolean isEmptyEffectorBoolean(Effector *this)
{
    return(this->itemsHeld == NULL);
}

Boolean isFullEffectorBoolean(Effector *this)
{
    return(this->itemsHeld != NULL);
}

/*----------------------------------------------------------------------------*/
/* holds(e, f) returns TRUE iff e holds an object o and f(o) is TRUE          */

double holds(Effector *effector, double objectTest(Object *))
{
    ObjectList *os;
    double found = 0.0;

    for (os = effector->itemsHeld; os != NULL; os = os->tail) {
        found = FZOR(found, objectTest(os->head));
    }
    return(found);
}

/*----------------------------------------------------------------------------*/
/* bestEmptyHand(cs) returns the empty hand with the greatest activation, or  */
/* NULL                                                                       */

Effector *bestEmptyHand(ContentionScheduler *cs)
{
    EffectorList *es;
    Effector *hand = NULL;
    double tmp = minAct;
    
    for (es = cs->effectors; es; es = es->tail) {
        if ((isEmptyEffectorBoolean(es->head)) && (es->head->activation[cs->cycle-1] > tmp)) {
            tmp = es->head->activation[cs->cycle-1];
            hand = es->head;
        }
    }
    return(hand);
}

/*----------------------------------------------------------------------------*/
/* bestNonemptyHand(cs) returns the nonempty hand with the greatest           */
/* activation, or NULL                                                        */

Effector *bestNonemptyHand(ContentionScheduler *cs)
{
    EffectorList *es;
    Effector *hand = NULL;
    double tmp = minAct;
    
    for (es = cs->effectors; es; es = es->tail) {
        if ((isFullEffectorBoolean(es->head)) && (es->head->activation[cs->cycle-1] > tmp)) {
            tmp = es->head->activation[cs->cycle-1];
            hand = es->head;
	}
    }
    return(hand);
}

/******************************************************************************/
/* State change functions: ****************************************************/

int effectorDiscardCoordinate(Effector *hand)
{
    return((strcmp(hand->name, "leftHand") == 0) ? -1 : +1);
}

/******************************************************************************/

Effector *effectorGetOtherHand(ContentionScheduler *cs, Effector *hand)
{
    EffectorList *es = cs->effectors;
    while ((es != NULL) && (es->head == hand)) {
        es = es->tail;
    }
    return( es == NULL ? NULL : es->head );
}

/******************************************************************************/

Object *effectorGetContents(Effector *hand)
{
    if ((hand == NULL) || (hand->itemsHeld == NULL)) {
        return(NULL);
    }
    else {
        return(hand->itemsHeld->head);
    }
}

/******************************************************************************/

void effectorAddItemToContents(Effector *hand, Object *item)
{
    hand->itemsHeld = objectListAppendElement(hand->itemsHeld, item);
}

/******************************************************************************/

void install_null_task_effectors(ContentionScheduler *cs)
{
}

void write_null_task_effector_attributes(FILE *fp, Effector *effector)
{
}

/******************************************************************************/
