/*******************************************************************************

    FILE: espresso/action.c
    AUTHOR: R. Cooper
    CREATION DATE: Fri Apr 16 08:47:18 2004
    CONTENTS: Actions for the espresso pot task

*******************************************************************************/

#include "../cs.h"
#include "task.h"

/******************************************************************************/

/* A local buffer used for actionMessage:                                     */
static char buffer[128]; 

/******************************************************************************/

static Boolean isEspressoPotBoolean(Object *object)
{
    return(isEspressoBase(object) > 0.90);
}

static Boolean isEspressoTopBoolean(Object *object)
{
    return(isEspressoTop(object) > 0.90);
}

/******************************************************************************/

void fill_scoop(ContentionScheduler *cs, Effector *hand)
{
    Object *source = cs->fixated_object;
    Object *scoop = effectorGetContents(hand);

    char *scoop_name = (scoop == NULL ? "nothing" : scoop->name);
    char *source_name = (source == NULL ? "nothing" : source->name);
    char *hand_name = (hand == NULL ? "nothing" : hand->name);

    g_snprintf(buffer, 128, "Scooping with %s held by %s from %s", scoop_name, hand_name, source_name);
    actionMessage(cs, "FillScoop", buffer);
    record_action(cs, "FillScoop", scoop_name, source_name, hand_name);

    if (hand == NULL) {
        errorEffectorNull(cs, "FillScoop");
    }
    else if (scoop == NULL) {
        errorEffectorState(cs, "FillScoop: hand (%s) is empty", hand->name);
    }
    else if (isTool(scoop) < 0.9) {
        errorObjectIncorrect(cs, "%s: hand (%s) holds incorrect object (%s)", "FillScoop", hand->name, scoop->name);
    }
    else if (!isContainerBoolean(source)) {
        errorObjectIncorrect(cs, "%s: source (%s) is not a container", "FillScoop", source->name, NULL);
    }
    else if (!isOpen(source)) {
        errorObjectIncorrect(cs, "%s: source (%s) is not open", "FillScoop", source->name, NULL);
    }
    else if (isEmptyObject(source)) {
        errorObjectIncorrect(cs, "%s: source (%s) is empty", "FillScoop", source->name, NULL);
    }
    else {
        Object *stuff = objectGetContent(source);

        /* stuff is a MASS object so source contents are unaltered */

        traceObjectPreAction(cs, scoop);
        objectInstallContents(scoop, objectCreate(cs, stuff->name, stuff->features, 0));
        traceObjectPostAction(cs, scoop);

        objectBindToRole(cs, source, SOURCE);
    }
}

/******************************************************************************/

void empty_scoop(ContentionScheduler *cs, Effector *hand)
{
    Object *target = cs->fixated_object;
    Object *scoop = effectorGetContents(hand);

    char *scoop_name = (scoop == NULL ? "nothing" : scoop->name);
    char *target_name = (target == NULL ? "nothing" : target->name);
    char *hand_name = (hand == NULL ? "nothing" : hand->name);

    g_snprintf(buffer, 128, "Emptying %s (held by %s) into %s", scoop_name, hand_name, target_name);
    actionMessage(cs, "EmptyScoop", buffer);
    record_action(cs, "EmptyScoop", scoop_name, target_name, hand_name);

    if (hand == NULL) {
        errorEffectorNull(cs, "FillScoop");
    }
    else if (scoop == NULL) {
        errorEffectorState(cs, "FillScoop: hand (%s) is empty", hand->name);
    }
    else if (isTool(scoop) < 0.9) {
        errorObjectIncorrect(cs, "%s: hand (%s) holds incorrect object (%s)", "FillScoop", hand->name, scoop->name);
    }
    else if (isEmptyObject(scoop)) {
        errorObjectIncorrect(cs, "%s: implement (%s) is empty", "FillScoop", scoop->name, NULL);
    }
    else if (!isContainerBoolean(target)) {
        errorObjectIncorrect(cs, "%s: target (%s) is not a container", "FillScoop", target->name, NULL);
    }
    else if (!isOpen(target)) {
        errorObjectIncorrect(cs, "%s: target (%s) is not open", "FillScoop", target->name, NULL);
    }
    else {
        Object *theme = objectGetContent(scoop);

        traceObjectPreAction(cs, scoop);
        scoop->itemsContained = objectListDeleteElement(scoop->itemsContained, theme);
        traceObjectPostAction(cs, scoop);

        traceObjectPreAction(cs, target);
        objectInstallContents(target, theme);
        traceObjectPostAction(cs, target);

        objectBindToRole(cs, target, TARGET);
    }
}

/******************************************************************************/
/*************** pour(OpenContainer, HandContainingOpenSource) ****************/

/* Pouring moves some of the source to the target. Pouring fails if the thing */
/* held by the hand isn't an open non-empty container or if the target is not */
/* an open container, or if the source and target are the same.               */

void pour_some(ContentionScheduler *cs, Effector *effector)
{
    Object *target = cs->fixated_object;
    Object *source = effectorGetContents(effector);
    Object *juice = objectGetContent(source);
    Object *lid;

    char *source_name = (source == NULL ? "nothing" : source->name);
    char *juice_name = (juice == NULL ? "nothing" : juice->name);
    char *target_name = (target == NULL ? "nothing" : target->name);
    char *effector_name = (effector == NULL ? "nothing" : effector->name);

    g_snprintf(buffer, 128, "Pouring %s from %s (held by %s) into %s", juice_name, source_name, effector_name, target_name);
    actionMessage(cs, "Pour", buffer);
    record_action(cs, "PourSome", source_name, target_name, effector_name);

    if (effector == NULL) {
        errorEffectorNull(cs, "Pour");
    }
    else if (target == NULL) {
        errorObjectNull(cs, "Target", "Pour");
    }
    else if (source == NULL) {
        errorEffectorState(cs, "Pour: hand (%s) is empty", effector->name);
    }
    else if (isClosedBoolean(source)) {
        errorObjectState(cs, "%s: source (%s) is closed", "Pour", source->name, NULL);
    }
    else if ((lid = objectGetCover(source)) != NULL) {
        errorObjectState(cs, "%s: source (%s) is covered by %s", "Pour", source->name, lid->name);
    }
    else if (isContainedBoolean(source)) {
        errorObjectState(cs, "%s: source (%s) is inside %s", "Pour", source->name, source->container->name);
    }
    else if (!isFullObjectBoolean(source)) {
        errorObjectState(cs, "%s: source (%s) is empty", "Pour", source->name, NULL);
    }
    else if (isClosedBoolean(target)) {
        errorObjectState(cs, "%s: target (%s) is closed", "Pour", target->name, NULL);
    }
    else if ((lid = objectGetCover(target)) != NULL) {
        errorObjectState(cs, "%s: target (%s) is covered by %s", "Pour", target->name, lid->name);
    }
    else if (isContainedBoolean(target)) {
        errorObjectState(cs, "%s: target (%s) is inside %s", "Pour", target->name, target->container->name);
    }
    else if (target == source) {
        errorObjectIncorrect(cs, "%s: target and source are same (%s = %s)", "Pour", target->name, source->name);
    }
    else if (!isContainerBoolean(target)) {
        errorObjectIncorrect(cs, "%s: target (%s) is not a container", "Pour", target->name, NULL);
    }
    else {

        /* pour operates on MASS objects, so the source's contents are unaltered */

        source->state = source->state | STATE_USED;

        traceObjectPreAction(cs, target);
        objectInstallContents(target, objectCreate(cs, juice->name, juice->features, 0));
        traceObjectPostAction(cs, target);

        objectBindToRole(cs, target, TARGET);
    }
}

/******************************************************************************/
/****************** screwOpen(HandContainingJar, EmptyHand) *******************/

/* Unscrewing results in the held jar becoming open, emptyHand holding the    */
/* lid, and the position of the lid being heldBy(emptyHand). Unscrewing fails */
/* if the jarHand doesn't hold a closed jar or if the emptyHand is not        */
/* initially empty.                                                           */

void screwPotOpen(ContentionScheduler *cs, Effector *jarHand, Effector *emptyHand)
{
    Object *jar = effectorGetContents(jarHand);
    Object *lid;

    char *jar_name = (jar == NULL ? "nothing" : jar->name);
    char *empty_hand_name = (emptyHand == NULL ? "nothing" : emptyHand->name);

    g_snprintf(buffer, 128, "Screwing open %s with %s", jar_name, empty_hand_name);
    actionMessage(cs, "ScrewPotOpen", buffer);
    record_action(cs, "ScrewPotOpen", jar_name, empty_hand_name, NULL);

    if (jarHand == NULL) {
        errorEffectorNull(cs, "ScrewPotOpen");
    }
    else if (jar == NULL) {
        errorEffectorState(cs, "ScewOpen: jarHand (%s) is empty", jarHand->name);
    }
    else if (!isEspressoPotBoolean(jar)) {
        errorObjectIncorrect(cs, "%s: jarHand (%s) holds incorrect object (%s)", "ScrewPotOpen", jarHand->name, jar->name);
    }
    else if (isOpenBoolean(jar)) {
        errorObjectState(cs, "%s: desired object (%s) is already open", "ScrewPotOpen", jar->name, NULL);
    }
    else if ((lid = objectGetCover(jar)) == NULL) {
        errorObjectState(cs, "%s: %s doesn't carry a lid\n", "ScrewPotOpen", jar->name, NULL);
    }
    else if (!isEspressoTopBoolean(lid)) {
        errorObjectState(cs, "%s: %s carries incorrect lid (%s)\n", "ScrewPotOpen", jar->name, lid->name);
    }
    else {
        if (emptyHand == NULL) {
            emptyHand = effectorGetWorst(cs);
            effectorDiscardContents(cs, emptyHand);
        }
        else if (isFullEffectorBoolean(emptyHand)) {
            if (emptyHand == jarHand) {
                emptyHand = effectorGetOtherHand(cs, jarHand);
            }
            effectorDiscardContents(cs, emptyHand);
        }

        traceEffectorPreAction(cs, emptyHand);
        effectorAddItemToContents(emptyHand, lid);
        traceEffectorPostAction(cs, emptyHand);

        traceObjectPreAction(cs, lid);
        lid->holder = emptyHand;
        traceObjectPostAction(cs, lid);

        traceObjectPreAction(cs, jar);
        objectRemoveCover(jar, lid);
        traceObjectPostAction(cs, jar);

        objectBindToRole(cs, effectorGetContents(emptyHand), THEME);
    }
}

/******************************************************************************/
/*********** screwPotClosed(HandContainingJar, HandContainingLid) ****************/

/* Screwing closed results in the jar becoming closed, and the lidHand        */
/* becoming empty. Screwing closed fails if the jarHand doesn't contain an    */
/* open jar or if the lidHand doesn't contain the lid of that jar.            */

void screwPotClosed(ContentionScheduler *cs, Effector *jarHand, Effector *lidHand)
{
    Object *jar = effectorGetContents(jarHand);
    Object *lid = effectorGetContents(lidHand);

    char *jar_name = (jar == NULL ? "nothing" : jar->name);
    char *lid_name = (lid == NULL ? "nothing" : lid->name);
    char *lid_hand_name = (lidHand == NULL ? "nothing" : lidHand->name);

    g_snprintf(buffer, 128, "Screwing closed %s with %s", jar_name, lid_hand_name);
    actionMessage(cs, "ScrewPotClosed", buffer);
    record_action(cs, "ScrewPotClosed", jar_name, lid_name, lid_hand_name);

    if (jarHand == NULL) {
        errorEffectorNull(cs, "ScrewPotClosed");
    }
    else if (lidHand == NULL) {
        errorEffectorNull(cs, "ScrewPotClosed");
    }
    else if (jarHand == lidHand) {
        errorEffectorIncorrect(cs, "ScrewPotClosed: jarHand and lidHand are same (%s, %s)", jarHand->name, lidHand->name);
    }
    else if (jar == NULL) {
        errorEffectorState(cs, "ScrewPotClosed: jarHand (%s) is empty", jarHand->name);
    }
    else if (lid == NULL) {
        errorEffectorState(cs, "ScrewPotClosed: lidHand (%s) is empty", lidHand->name);
    }
    else if ((jar->holder == NULL) || (lid->holder == NULL)) {
        errorEffectorNull(cs, "ScrewPotClosed");
    }
    else if (!isEspressoPotBoolean(jar)) {
        errorObjectIncorrect(cs, "%s: jarHand (%s) holds incorrect object (%s)", "ScrewPotClosed", jar->holder->name, jar->name);
    }
    else if (!isEspressoTopBoolean(lid)) {
        errorObjectIncorrect(cs, "%s: lidHand (%s) holds incorrect object (%s)", "ScrewPotClosed", lid->holder->name, lid->name);
    }
    else if (lid == objectGetCover(jar)) {
        errorObjectState(cs, "%s: %s is already covered by %s", "ScrewPotClosed", jar->name, lid->name);
    }
    else {
        traceEffectorPreAction(cs, lidHand);
        lidHand->itemsHeld = objectListDeleteElement(lidHand->itemsHeld, lid);
        traceEffectorPostAction(cs, lidHand);

        traceObjectPreAction(cs, lid);
        lid->holder = NULL;
        traceObjectPostAction(cs, lid);

        traceObjectPreAction(cs, jar);
        objectInstallCover(jar, lid);
        traceObjectPostAction(cs, jar);

        objectBindToRole(cs, lid, THEME);
    }
}

/******************************************************************************/
/********************* cover(Target, HandContainingTheme) *********************/

/* Covering moves the held theme to an open target. Covering fails if the     */
/* hand is empty, the target is closed, or if the theme is the same as the    */
/* target. If both target and theme are slices of bread, cover also unifies   */
/* the bread-meat-mustard construct into a single object.                     */

void cover(ContentionScheduler *cs, Effector *effector)
{
    Object *target = cs->fixated_object;
    Object *theme = effectorGetContents(effector);

    char *target_name = (target == NULL ? "nothing" : target->name);
    char *theme_name = (theme == NULL ? "nothing" : theme->name);
    char *effector_name = (effector == NULL ? "nothing" : effector->name);

    g_snprintf(buffer, 128, "Putting %s (held by %s) onto %s", theme_name, effector_name, target_name);
    actionMessage(cs, "Cover", buffer);
    record_action(cs, "Cover", target_name, theme_name, effector_name);

    if (effector == NULL) {
        errorEffectorNull(cs, "Cover");
    }
    else if (target == NULL) {
        errorObjectNull(cs, "Target", "Cover");
    }
    else if (theme == NULL) {
        errorEffectorState(cs, "Cover: hand (%s) is empty", effector->name);
    }
    else if (theme == target) {
        errorObjectIncorrect(cs, "%s: theme and target are same (%s, %s)", "Cover", theme->name, target->name);
    }
    else if (isContainedBoolean(target)) {
        errorObjectState(cs, "%s: target (%s) is inside %s", "Cover", target->name, target->container->name);
    }
    else if (isInSubordinateContactWith(target, theme)) {
        errorObjectState(cs, "%s: theme (%s) is directly or indirectly in contact with target (%s)", "Cover", theme->name, target->name);
    }
    else {
        traceEffectorPreAction(cs, effector);
        effector->itemsHeld = objectListDeleteElement(effector->itemsHeld, theme);
        traceEffectorPostAction(cs, effector);

        traceObjectPreAction(cs, theme);
        theme->holder = NULL;
        objectInstallCover(target, theme);
        traceObjectPostAction(cs, theme);

        traceObjectPreAction(cs, target);
        traceObjectPostAction(cs, target);

        objectBindToRole(cs, target, TARGET);
    }
}

/******************************************************************************/
