/*******************************************************************************

    FILE: espresso/analyse.c
    AUTHOR: R. Cooper
    CREATION DATE: Fri Aug 13 09:20:59 2004
    CONTENTS: Analysis routine for the espresso task

*******************************************************************************/

#include "../cs.h"
#include "../analyse.h"

/******************************************************************************/

typedef enum espresso_object {
    OBJECT_NONE, OBJECT_POT_BASE, OBJECT_FILTER, OBJECT_POT_TOP, OBJECT_SPOON,
    OBJECT_WATER_JUG, OBJECT_COFFEE_TIN, OBJECT_TABLE,
    OBJECT_UNKNOWN
} EspressoObject;

typedef enum espresso_location {
    LOC_NONE, LOC_HELD, LOC_ON_TABLE, LOC_ON_POT, LOC_ON_SOMETHING_ELSE
} EspressoLocation;

typedef struct espresso_analysis_object_list {
    EspressoObject                       id;        /* Enumerated id */
    EspressoLocation                     location;
    int                                  state;
    struct espresso_analysis_object_list *next;     /* Next object in the list  */
} EspressoAnalysisObjectList;

typedef struct espresso_state {
    // Information on all possible objects in the world
    EspressoAnalysisObjectList   *items; 

    // Fixated item:
    EspressoObject                fixated;

    // In case we scoop with the hand:
    Boolean                     hand_contains_water;
    Boolean                     hand_contains_grounds;

    // Other relevant state information:
    Boolean                     pot_filled;
    Boolean                     filter_filled;
    Boolean                     filter_positioned;
    Boolean                     pot_closed;
} EspressoState;

#define ESPRESSO_STATE_CONTAINS_WATER          1
#define ESPRESSO_STATE_CONTAINS_GROUNDS        2
#define ESPRESSO_STATE_CLOSED                  4

/******************************************************************************/

static EspressoObject object_id_from_string(char *arg)
{
    if (arg == NULL) {
        return(OBJECT_NONE);
    }
    else if (strcmp(arg, "nothing") == 0) {
        return(OBJECT_NONE);
    }
    else if (strcmp(arg, "espresso_pot_base") == 0) {
        return(OBJECT_POT_BASE);
    }
    else if (strcmp(arg, "espresso_pot_filter") == 0) {
        return(OBJECT_FILTER);
    }
    else if (strcmp(arg, "espresso_pot_top") == 0) {
        return(OBJECT_POT_TOP);
    }
    else if (strcmp(arg, "spoon") == 0) {
        return(OBJECT_SPOON);
    }
    else if (strcmp(arg, "water_jug") == 0) {
        return(OBJECT_WATER_JUG);
    }
    else if (strcmp(arg, "coffee_tin") == 0) {
        return(OBJECT_COFFEE_TIN);
    }
    else if (strcmp(arg, "table") == 0) {
        return(OBJECT_TABLE);
    }
    else {    /* Final catch-all just in case: */
        report_unrecognised_object(stdout, arg);
        return(OBJECT_UNKNOWN);
    }
}

static EspressoAnalysisObjectList *object_ptr_from_id(EspressoState *state, EspressoObject id)
{
    EspressoAnalysisObjectList *tmp = state->items;

    while ((tmp != NULL) && (tmp->id != id)) {
        tmp = tmp->next;
    }
    if ((tmp == NULL) || (tmp->id != id)) {
        return(NULL);
    }
    else {
        return(tmp);
    }
}

static EspressoAnalysisObjectList *object_ptr_from_string(EspressoState *state, char *arg)
{
    return(object_ptr_from_id(state, object_id_from_string(arg)));
}

static Boolean action_compare(ActionList *this, char *action, EspressoObject arg1)
{
    if (this == NULL) {
        return(FALSE);
    }
    else if (strcmp(this->name, action) != 0) {
        return(FALSE);
    }
    else if (object_id_from_string(this->argument1) != arg1) {
        return(FALSE);
    }
    else {
        return(TRUE);
    }
}

/******************************************************************************/

static void state_add_item(EspressoState *state, EspressoObject id, EspressoLocation loc, int item_state)
{
    EspressoAnalysisObjectList *new = (EspressoAnalysisObjectList *)malloc(sizeof(EspressoAnalysisObjectList));

    if (new != NULL) {
        new->id = id;
        new->location = loc;
        new->state = item_state;
        new->next = state->items;
        state->items = new;
    }
}

static void state_initialise(EspressoState *state)
{
    state->items = NULL;

    /* All objects must be initialised to match espresso/object.c */

    state_add_item(state, OBJECT_POT_BASE, LOC_ON_TABLE, 0);
    state_add_item(state, OBJECT_FILTER, LOC_ON_TABLE, 0);
    state_add_item(state, OBJECT_POT_TOP, LOC_ON_TABLE, 0);
    state_add_item(state, OBJECT_SPOON, LOC_ON_TABLE, 0);
    state_add_item(state, OBJECT_WATER_JUG, LOC_ON_TABLE, ESPRESSO_STATE_CONTAINS_WATER);
    state_add_item(state, OBJECT_COFFEE_TIN, LOC_ON_TABLE, ESPRESSO_STATE_CONTAINS_GROUNDS);
    state_add_item(state, OBJECT_TABLE, LOC_NONE, 0);

    state->fixated = OBJECT_NONE;
    state->hand_contains_water = OBJECT_NONE;
    state->hand_contains_grounds = OBJECT_NONE;

    state->pot_filled = FALSE;
    state->filter_filled = FALSE;
    state->filter_positioned = FALSE;
    state->pot_closed = FALSE;
}

static void state_destroy(EspressoState *state)
{
    while (state->items != NULL) {
        EspressoAnalysisObjectList *tmp = state->items;
        state->items = tmp->next;
        free(tmp);
    }
}

/******************************************************************************/

static void check_consecutive_toying_actions(ActionList *this, EspressoObject arg1)
{
    if (action_compare(this->next, "PickUp", arg1)) {
        if (action_compare(this->next->next, "PutDown", arg1)) {
            this->next->checked = TRUE;
            this->next->next->checked = TRUE;
            check_consecutive_toying_actions(this->next->next, arg1);
        }
        else if (action_compare(this->next->next, "Discard", arg1)) {
            this->next->checked = TRUE;
            this->next->next->checked = TRUE;
            check_consecutive_toying_actions(this->next->next, arg1);
        }
    }
}

static Boolean is_espresso_container(EspressoObject id)
{
    switch (id) {
        case OBJECT_POT_BASE: { return(TRUE); }
        case OBJECT_FILTER: { return(TRUE); }
        case OBJECT_POT_TOP: { return(TRUE); }
        case OBJECT_SPOON: { return(TRUE); }
        case OBJECT_WATER_JUG: { return(TRUE); }
        case OBJECT_COFFEE_TIN: { return(TRUE); }
        case OBJECT_TABLE: { return(FALSE); }
        case OBJECT_NONE: { return(FALSE); } 
        case OBJECT_UNKNOWN: { return(FALSE); }
    }
    return(FALSE);
}

/******************************************************************************/

static void score_action_fixate(EspressoState *state, DataAnalysis *la, ActionList *action, char *arg1)
{
    EspressoAnalysisObjectList *item1;

    if (object_id_from_string(arg1) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "Fixate", arg1, NULL);
        ignore_consecutive_actions(action);
    }
    else if ((item1 = object_ptr_from_string(state, arg1)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else {
        state->fixated = item1->id;
    }
}

/*----------------------------------------------------------------------------*/

static void score_action_pickup(EspressoState *state, DataAnalysis *la, ActionList *action, char *arg1)
{
    EspressoAnalysisObjectList *item1;

    if (object_id_from_string(arg1) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "PickUp", arg1, NULL);
        ignore_consecutive_actions(action);
    }
    else if ((item1 = object_ptr_from_string(state, arg1)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if (item1->location == LOC_HELD) {
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "PickUp", arg1, NULL);
        ignore_consecutive_actions(action);
    }
    else {
        item1->location = LOC_HELD;
        if (action_compare(action->next, "PutDown", item1->id)) {
            // Score TOYING error and ignore any consecutive toying
            analyseErrorLogAppend(la, ERROR_TOYING, "PickUp/PutDown", arg1, NULL);
            check_consecutive_toying_actions(action->next, item1->id);
        }
        else if (action_compare(action->next, "Discard", item1->id)) {
            // Score TOYING error and ignore any consecutive toying
            analyseErrorLogAppend(la, ERROR_TOYING, "PickUp/Discard", arg1, NULL);
            check_consecutive_toying_actions(action->next, item1->id);
        }
        if (item1->id == OBJECT_FILTER) {
            state->filter_positioned = FALSE;
        }
    }
}

/*----------------------------------------------------------------------------*/

static void score_action_putdown(EspressoState *state, DataAnalysis *la, ActionList *action, char *arg1)
{
    EspressoAnalysisObjectList *item1;

    if (object_id_from_string(arg1) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "PutDown", arg1, NULL);
        ignore_consecutive_actions(action);
    }
    else if ((item1 = object_ptr_from_string(state, arg1)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if (item1->location != LOC_HELD) {
        // Score PERPLEXITY: Object is not held
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "PutDown", arg1, NULL);
        ignore_consecutive_actions(action);
    }
    else {
        item1->location = LOC_ON_TABLE;
    }
}

/*----------------------------------------------------------------------------*/

static void score_action_discard(EspressoState *state, DataAnalysis *la, ActionList *action, char *arg1)
{
    EspressoAnalysisObjectList *item1;

    if (object_id_from_string(arg1) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "Discard", arg1, NULL);
        ignore_consecutive_actions(action);
    }
    else if ((item1 = object_ptr_from_string(state, arg1)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if (item1->location != LOC_HELD) {
        // Score PERPLEXITY: Object is not held
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "Discard", arg1, NULL);
        ignore_consecutive_actions(action);
    }
    else {
        item1->location = LOC_ON_TABLE;
    }
}

/*----------------------------------------------------------------------------*/

static void score_action_pour_some(EspressoState *state, DataAnalysis *la, ActionList *action, char *arg1, char *arg2)
{
    EspressoAnalysisObjectList *item1, *item2;

    if (object_id_from_string(arg1) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PANTOMIME, "Pour", arg1, arg2);
    }
    else if ((item1 = object_ptr_from_string(state, arg1)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if (object_id_from_string(arg2) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "Pour", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if ((item2 = object_ptr_from_string(state, arg2)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if (item1 == item2) {
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "Pour", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if (item1->location != LOC_HELD) {
        // Score PERPLEXITY: source container isn't held
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "Pour", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if (!is_espresso_container(item1->id)) {
        // Score PERPLEXITY: Pouring from non-container
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "Pour", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if ((item1->state & (ESPRESSO_STATE_CONTAINS_WATER | ESPRESSO_STATE_CONTAINS_GROUNDS)) == 0) {
        // Score ANTICIPATION: Item 1 has not been filled
        analyseErrorLogAppend(la, ERROR_ACTION_ANTICIPATION, "Pour", arg1, arg2);
    }
    else if (item2->id == OBJECT_TABLE) {
        // Score MISLOCATION_1: Pouring onto table 
        analyseErrorLogAppend(la, ERROR_MISLOCATION_1, "Pour", arg1, arg2);
    }
    else if (!is_espresso_container(item2->id)) {
        // Score MISLOCATION_1: Pouring into non-container
        analyseErrorLogAppend(la, ERROR_MISLOCATION_1, "Pour", arg1, arg2);
    }
    else if (item1->state & ESPRESSO_STATE_CLOSED) {
        // Score ANTICIPATION: Item 1 hasn't been opened
        analyseErrorLogAppend(la, ERROR_ACTION_ANTICIPATION, "Pour", arg1, arg2);
    }
    else if (item2->state & ESPRESSO_STATE_CLOSED) {
        // Score ANTICIPATION: Item 2 hasn't been opened
        analyseErrorLogAppend(la, ERROR_ACTION_ANTICIPATION, "Pour", arg1, arg2);
    }
    else {
        // Workable ...
        if (item1->state & ESPRESSO_STATE_CONTAINS_WATER) {
            item2->state = item2->state | ESPRESSO_STATE_CONTAINS_WATER;
            if (item2->id == OBJECT_POT_BASE) {
                state->pot_filled = TRUE;
            }
            else if ((item2->id == OBJECT_FILTER) || (item2->id == OBJECT_POT_TOP)) {
                // Score MISLOCATION_2: Pouring water into silly place
                analyseErrorLogAppend(la, ERROR_MISLOCATION_2, "Pour", arg1, arg2);
            }
            else {
                // Score MISLOCATION_1: Pouring water into silly place
                analyseErrorLogAppend(la, ERROR_MISLOCATION_1, "Pour", arg1, arg2);
            }
	}
        if (item1->state & ESPRESSO_STATE_CONTAINS_GROUNDS) {
            item2->state = item2->state | ESPRESSO_STATE_CONTAINS_GROUNDS;
            if (item1->id != OBJECT_SPOON) {
                // Score TOOL OMISSION: Failed to use spoon
                analyseErrorLogAppend(la, ERROR_TOOL_OMISSION, "Pour", arg1, arg2);
            }
            if (item2->id == OBJECT_FILTER) {
                state->filter_filled = TRUE;
                if (item2->location != LOC_ON_POT) {
                    // Score ANTICIPATION: Filter is not on the pot
                    analyseErrorLogAppend(la, ERROR_ACTION_ANTICIPATION, "Pour", arg1, arg2);
                }
            }
            else if ((item2->id == OBJECT_POT_BASE) || (item2->id == OBJECT_POT_TOP)) {
                // Score MISLOCATION_2: Pouring grounds into silly place
                analyseErrorLogAppend(la, ERROR_MISLOCATION_2, "Pour", arg1, arg2);
            }
            else {
                // Score MISLOCATION_1: Pouring grounds into silly place
                analyseErrorLogAppend(la, ERROR_MISLOCATION_1, "Pour", arg1, arg2);
            }
        }
    }
}

/*----------------------------------------------------------------------------*/

static void score_action_cover(EspressoState *state, DataAnalysis *la, ActionList *action, char *arg1, char *arg2)
{
    EspressoAnalysisObjectList *item1;      // The base
    EspressoAnalysisObjectList *item2;      // The filter

    if (object_id_from_string(arg1) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PANTOMIME, "Cover", arg1, arg2);
    }
    else if ((item1 = object_ptr_from_string(state, arg1)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if (object_id_from_string(arg2) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PANTOMIME, "Cover01", arg1, arg2);
    }
    else if ((item2 = object_ptr_from_string(state, arg2)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if (item1 == item2) {
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "Cover02", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if (item1->id == OBJECT_FILTER) {
        // Score MISUSE 1: Covering the filter with something!
        analyseErrorLogAppend(la, ERROR_MISUSE_1, "Cover03", arg1, arg2);
        item2->location = LOC_ON_SOMETHING_ELSE;
    }
    else if (item2->id != OBJECT_FILTER) {
        // Covering something with a non-filter
        if (item1->id == OBJECT_POT_BASE) {
            // Score MISUSE 1: Covering the base of the pot with a non-filter
            analyseErrorLogAppend(la, ERROR_MISUSE_1, "Cover04", arg1, arg2);
            item2->location = LOC_ON_POT;
        }
        else if (item1->id == OBJECT_POT_TOP) {
            // Score MISUSE 1: Covering the top of the pot with a non-filter
            analyseErrorLogAppend(la, ERROR_MISUSE_1, "Cover04", arg1, arg2);
            item2->location = LOC_ON_SOMETHING_ELSE;
        }
        else {
            // Score ADDITION: Cover something else with a non-filter
            analyseErrorLogAppend(la, ERROR_ACTION_ADDITION, "Cover05", arg1, arg2);
            item2->location = LOC_ON_SOMETHING_ELSE;
        }
    }
    else if (item1->id == OBJECT_POT_TOP) {
        // Score MISLOCATION 2: Covering wrong part of pot
        analyseErrorLogAppend(la, ERROR_MISLOCATION_2, "Cover06", arg1, arg2);
        item2->location = LOC_ON_SOMETHING_ELSE;
    }
    else if (item1->id != OBJECT_POT_BASE) {
        // Score MISLOCATION 1: Covering wrong object
        analyseErrorLogAppend(la, ERROR_MISLOCATION_1, "Cover07", arg1, arg2);
        item2->location = LOC_ON_SOMETHING_ELSE;
    }
    else if (item1->state & ESPRESSO_STATE_CLOSED) {
        // Score ANTICIPATION: Need to open pot first
        analyseErrorLogAppend(la, ERROR_ACTION_ANTICIPATION, "Cover08", arg1, arg2);
    }
    else {
        item2->location = LOC_ON_POT;
        state->filter_positioned = TRUE;

        if ((item1->state & ESPRESSO_STATE_CONTAINS_WATER) == 0) {
            // Score ANTICIPATION: Need to put water in first
            analyseErrorLogAppend(la, ERROR_ACTION_ANTICIPATION, "Cover09", arg1, arg2);
        }
    }
}

/*----------------------------------------------------------------------------*/

static void score_action_fill_scoop(EspressoState *state, DataAnalysis *la, ActionList *action, char *arg1, char *arg2)
{
    EspressoAnalysisObjectList *item1, *item2;

    if (object_id_from_string(arg1) == OBJECT_NONE) {
        if ((item2 = object_ptr_from_string(state, arg2)) == NULL) {
            analyseErrorLogAppend(la, ERROR_PANTOMIME, "EmptyScoop", arg1, arg2);
        }
        else {
            if (!(item2->state & ESPRESSO_STATE_CLOSED)) {
                if (item2->state & (ESPRESSO_STATE_CONTAINS_WATER | ESPRESSO_STATE_CONTAINS_GROUNDS)) {
                    analyseErrorLogAppend(la, ERROR_TOOL_OMISSION, "FillScoop", arg1, arg2);
                    if (item2->state & ESPRESSO_STATE_CONTAINS_WATER) {
                        state->hand_contains_water = TRUE;
                    }
                    if (item2->state & ESPRESSO_STATE_CONTAINS_GROUNDS) {
                        state->hand_contains_grounds = TRUE;
                    }
                }
            }
        }
    }
    else if ((item1 = object_ptr_from_string(state, arg1)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if (object_id_from_string(arg2) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "FillScoop", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if ((item2 = object_ptr_from_string(state, arg2)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if (item1 == item2) {
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "FillScoop01", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if (!is_espresso_container(item1->id)) {
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "FillScoop02", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if (!is_espresso_container(item2->id)) {
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "FillScoop03", arg1, arg2);
        ignore_consecutive_actions(action);
    }

    else if (item1->state & ESPRESSO_STATE_CONTAINS_WATER) {
        // Item 1 (the scoop) is not empty
        if (item2->state & ESPRESSO_STATE_CONTAINS_WATER) {
            // Score PERSEVERATION: Scooping same subtance again
            analyseErrorLogAppend(la, ERROR_PERSEVERATION, "FillScoop04", arg1, arg2);
        }
        else if (item2->state & ESPRESSO_STATE_CONTAINS_GROUNDS) {
            // Score ANTICIPATION: Scooping without emptying the scoop
            analyseErrorLogAppend(la, ERROR_ACTION_ANTICIPATION, "FillScoop05", arg1, arg2);
            item1->state = item1->state | ESPRESSO_STATE_CONTAINS_GROUNDS;
        }
        else {
            // Score PANTOMIME: Scooping from empty container
            analyseErrorLogAppend(la, ERROR_PANTOMIME, "FillScoop06", arg1, arg2);
        }
    }
    else if (item1->state & ESPRESSO_STATE_CONTAINS_GROUNDS) {
        // Item 1 (the scoop) is not empty
        if (item2->state & ESPRESSO_STATE_CONTAINS_GROUNDS) {
            // Score PERSEVERATION: Scooping same subtance again
            analyseErrorLogAppend(la, ERROR_PERSEVERATION, "FillScoop07", arg1, arg2);
        }
        else if (item2->state & ESPRESSO_STATE_CONTAINS_WATER) {
            // Score ANTICIPATION: Scooping without emptying the scoop
            analyseErrorLogAppend(la, ERROR_ACTION_ANTICIPATION, "FillScoop08", arg1, arg2);
            item1->state = item1->state | ESPRESSO_STATE_CONTAINS_WATER;
        }
        else {
            // Score PANTOMIME: Scooping from empty container
            analyseErrorLogAppend(la, ERROR_PANTOMIME, "FillScoop09", arg1, arg2);
        }
    }
    else if ((item2->state & (ESPRESSO_STATE_CONTAINS_WATER | ESPRESSO_STATE_CONTAINS_GROUNDS)) == 0) {
        // Score PANTOMIME: Scooping from empty container
        analyseErrorLogAppend(la, ERROR_PANTOMIME, "FillScoop10", arg1, arg2);
    }
    else if (item1->state & ESPRESSO_STATE_CLOSED) {
        // Score ANTICIPATION: Scooping with a closed container
        analyseErrorLogAppend(la, ERROR_ACTION_ANTICIPATION, "FillScoop11", arg1, arg2);
    }
    else if (item2->state & ESPRESSO_STATE_CLOSED) {
        // Score ANTICIPATION: Scooping from a closed container
        analyseErrorLogAppend(la, ERROR_ACTION_ANTICIPATION, "FillScoop12", arg1, arg2);
    }
    else {
        // Workable...
        if ((item1->id == OBJECT_FILTER) && (item2->state & ESPRESSO_STATE_CONTAINS_GROUNDS)) {
            // Score TOOL OMISSION: Filling the scoop directly!
            analyseErrorLogAppend(la, ERROR_TOOL_OMISSION, "FillScoop13", arg1, arg2);
            state->filter_filled = TRUE;
        }
        else if ((item1->id == OBJECT_POT_BASE) && (item2->state & ESPRESSO_STATE_CONTAINS_WATER)) {
            // Score MISUSE 2: Filling the base directly!
            analyseErrorLogAppend(la, ERROR_MISUSE_2, "FillScoop14", arg1, arg2);
            state->filter_filled = TRUE;
        }
        else if (item1->id != OBJECT_SPOON) {
            // Score MISUSE 1: Item 1 is not a spoon!
            analyseErrorLogAppend(la, ERROR_MISUSE_1, "FillScoop15", arg1, arg2);
        }
        else if (item2->state & ESPRESSO_STATE_CONTAINS_WATER) {
            // Score MISUSE 2: Using spoon to transfer water
            analyseErrorLogAppend(la, ERROR_MISUSE_2, "FillScoop16", arg1, arg2);
        }
        // Ok .. so fill item1 with some of what was in item2:
        if (item2->state & ESPRESSO_STATE_CONTAINS_WATER) {
            item1->state = item1->state | ESPRESSO_STATE_CONTAINS_WATER;
        }
        if (item2->state & ESPRESSO_STATE_CONTAINS_GROUNDS) {
            item1->state = item1->state | ESPRESSO_STATE_CONTAINS_GROUNDS;
        }
    }
}

/*----------------------------------------------------------------------------*/

static void score_action_empty_scoop(EspressoState *state, DataAnalysis *la, ActionList *action, char *arg1, char *arg2)
{
    EspressoAnalysisObjectList *item1, *item2;

    if (object_id_from_string(arg1) == OBJECT_NONE) {
        if ((item2 = object_ptr_from_string(state, arg2)) == NULL) {
            analyseErrorLogAppend(la, ERROR_PANTOMIME, "EmptyScoop", arg1, arg2);
	}
        else {
            if (!(item2->state & ESPRESSO_STATE_CLOSED)) {
                if (state->hand_contains_water) {
                    item2->state = item2->state | ESPRESSO_STATE_CONTAINS_WATER;
                    state->hand_contains_water = FALSE;
                    if ((item2->id == OBJECT_POT_BASE) || (item2->id == OBJECT_POT_TOP)) {
                        state->pot_filled = TRUE;
                    }
                }
                if (state->hand_contains_grounds) {
                    item2->state = item2->state | ESPRESSO_STATE_CONTAINS_GROUNDS;
                    state->hand_contains_grounds = FALSE;
                    if (item2->id == OBJECT_FILTER) {
                        state->filter_filled = TRUE;
                    }
                }
            }
        }
    }
    else if ((item1 = object_ptr_from_string(state, arg1)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if (object_id_from_string(arg2) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "EmptyScoop", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if ((item2 = object_ptr_from_string(state, arg2)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if (item1 == item2) {
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "EmptyScoop01", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if (!is_espresso_container(item1->id)) {
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "EmptyScoop02", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if ((item1->state & (ESPRESSO_STATE_CONTAINS_WATER | ESPRESSO_STATE_CONTAINS_GROUNDS)) == 0) {
        // Score ANTICIPATION: Item 1 is empty
        analyseErrorLogAppend(la, ERROR_ACTION_ANTICIPATION, "EmptyScoop03", arg1, arg2);
    }
    else if (item1->state & ESPRESSO_STATE_CLOSED) {
        // Score ANTICIPATION: Scooping with a closed container
        analyseErrorLogAppend(la, ERROR_ACTION_ANTICIPATION, "EmptyScoop04", arg1, arg2);
    }
    else {
        // Workable...
        if (item2->id == OBJECT_TABLE) {
            // Score MISLOCATION_1: Emptying scoop onto table 
            analyseErrorLogAppend(la, ERROR_MISLOCATION_1, "EmptyScoop05", arg1, arg2);
        }
        else if (!is_espresso_container(item2->id)) {
            // Score MISLOCATION_1: Emptying scoop into non-container
            analyseErrorLogAppend(la, ERROR_MISLOCATION_1, "EmptyScoop06", arg1, arg2);
        }

        if (item2->state & ESPRESSO_STATE_CLOSED) {
            // Score ANTICIPATION: Scooping into a closed container
            analyseErrorLogAppend(la, ERROR_ACTION_ANTICIPATION, "EmptyScoop07", arg1, arg2);
        }
        else {
            if (item1->state & ESPRESSO_STATE_CONTAINS_WATER) {
                if (item2->id == OBJECT_POT_BASE)  {
                    state->pot_filled = TRUE;
                }
                else if ((item2->id == OBJECT_FILTER) || (item2->id == OBJECT_POT_TOP)) {
                    if (!(item1->state & ESPRESSO_STATE_CONTAINS_GROUNDS)) {
                        analyseErrorLogAppend(la, ERROR_MISLOCATION_2, "EmptyScoop08", arg1, arg2);
                        state->pot_filled = TRUE;
		    }
                }
                else {
                    analyseErrorLogAppend(la, ERROR_MISLOCATION_1, "EmptyScoop09", arg1, arg2);
                }
            }

            if (item1->state & ESPRESSO_STATE_CONTAINS_GROUNDS) {
                if (item2->id == OBJECT_FILTER)  {
                    state->filter_filled = TRUE;
                    if (item2->location != LOC_ON_POT) {
                        // Score ANTICIPATION: Filter is not on the pot
                        analyseErrorLogAppend(la, ERROR_ACTION_ANTICIPATION, "EmptyScoop10", arg1, arg2);
                    }
                }
                else if ((item2->id == OBJECT_POT_BASE) || (item2->id == OBJECT_POT_TOP)) {
                    if (!(item1->state & ESPRESSO_STATE_CONTAINS_WATER)) {
                        analyseErrorLogAppend(la, ERROR_MISLOCATION_2, "EmptyScoop11", arg1, arg2);
                        state->filter_filled = TRUE;
                    }
                }
                else {
                    analyseErrorLogAppend(la, ERROR_MISLOCATION_1, "EmptyScoop12", arg1, arg2);
                }
            }

            if (item1->state & ESPRESSO_STATE_CONTAINS_WATER) {
                item2->state = item2->state | ESPRESSO_STATE_CONTAINS_WATER;
                item1->state = item1->state & ~ESPRESSO_STATE_CONTAINS_WATER;
            }

            if (item1->state & ESPRESSO_STATE_CONTAINS_GROUNDS) {
                item2->state = item2->state | ESPRESSO_STATE_CONTAINS_GROUNDS;
                item1->state = item1->state & ~ESPRESSO_STATE_CONTAINS_GROUNDS;
            }
        }
    }
}

/*----------------------------------------------------------------------------*/

static void score_action_screw_pot_open(EspressoState *state, DataAnalysis *la, ActionList *action, char *arg1)
{
    EspressoAnalysisObjectList *item1, *item2;

    if (object_id_from_string(arg1) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PANTOMIME, "ScrewOpen", arg1, NULL);
    }
    else if ((item1 = object_ptr_from_string(state, arg1)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if (item1->location != LOC_HELD) {
        // Score PERPLEXITY: thing isn't held
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "ScrewOpen", arg1, NULL);
        ignore_consecutive_actions(action);
    }
    else if ((item1->state & ESPRESSO_STATE_CLOSED) == 0) {
        // Score PERPLEXITY: base isn't closed
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "ScrewOpen", arg1, NULL);
        ignore_consecutive_actions(action);
    }
    else {
        item1->state = item1->state & ~ESPRESSO_STATE_CLOSED;
        if (item1->id == OBJECT_POT_BASE) {
            state->pot_closed = FALSE;
            if ((item2 = object_ptr_from_id(state, OBJECT_POT_TOP)) != NULL) {
                item2->state = item2->state & ~ESPRESSO_STATE_CLOSED;
                item2->location = LOC_HELD;
            }
        }
    }
}

/*----------------------------------------------------------------------------*/

static void score_action_screw_pot_closed(EspressoState *state, DataAnalysis *la, ActionList *action, char *arg1, char *arg2)
{
    EspressoAnalysisObjectList *item1, *item2;

    if (object_id_from_string(arg1) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PANTOMIME, "ScrewClosed", arg1, arg2);
    }
    else if ((item1 = object_ptr_from_string(state, arg1)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if (object_id_from_string(arg2) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "ScrewClosed", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if ((item2 = object_ptr_from_string(state, arg2)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if (item1->location != LOC_HELD) {
        // Score PERPLEXITY: base isn't held
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "ScrewClosed", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if (item2->location != LOC_HELD) {
        // Score PERPLEXITY: lid isn't held
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "ScrewClosed", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if (item1->id != OBJECT_POT_BASE) {
        // Score PERPLEXITY: "base" isn't a base!
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "ScrewClosed", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if (item2->id != OBJECT_POT_TOP) {
        // Score PERPLEXITY: "top" isn't a top!
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "ScrewClosed", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else {
        // Good...
        item2->location = LOC_ON_POT;
        item1->state = item1->state | ESPRESSO_STATE_CLOSED;
        item2->state = item2->state | ESPRESSO_STATE_CLOSED;
        state->pot_closed = TRUE;
    }
}

/******************************************************************************/

static void score_action(EspressoState *state, DataAnalysis *la, ActionList *action)
{
    if (action == NULL) {
        fprintf(stderr, "INTERNAL ERROR: NULL action in score_action/3 (file is espresso/analyse.c)\n");
    }
    else if (strcmp(action->name, "Fixate") == 0) {
        score_action_fixate(state, la, action, action->argument1);
    }
    else if (strcmp(action->name, "PickUp") == 0) {
        score_action_pickup(state, la, action, action->argument1);
    }
    else if (strcmp(action->name, "PutDown") == 0) {
        score_action_putdown(state, la, action, action->argument1);
    }
    else if (strcmp(action->name, "Discard") == 0) {
        score_action_discard(state, la, action, action->argument1);
    }
    else if (strcmp(action->name, "PourSome") == 0) {
        score_action_pour_some(state, la, action, action->argument1, action->argument2);
        score_consecutive_actions_as_perseverations(la, action);
    }
    else if (strcmp(action->name, "Cover") == 0) {
        score_action_cover(state, la, action, action->argument1, action->argument2);
        score_consecutive_actions_as_perseverations(la, action);
    }
    else if (strcmp(action->name, "FillScoop") == 0) {
        score_action_fill_scoop(state, la, action, action->argument1, action->argument2);
        score_consecutive_actions_as_perseverations(la, action);
    }
    else if (strcmp(action->name, "EmptyScoop") == 0) {
        score_action_empty_scoop(state, la, action, action->argument1, action->argument2);
        score_consecutive_actions_as_perseverations(la, action);
    }
    else if (strcmp(action->name, "ScrewPotOpen") == 0) {
        score_action_screw_pot_open(state, la, action, action->argument1);
        score_consecutive_actions_as_perseverations(la, action);
    }
    else if (strcmp(action->name, "ScrewPotClosed") == 0) {
        score_action_screw_pot_closed(state, la, action, action->argument1, action->argument2);
        score_consecutive_actions_as_perseverations(la, action);
    }
    else {
        report_unrecognised_action(stdout, action);
    }
}

/******************************************************************************/

static void score_omissions(EspressoState *state, DataAnalysis *la)
{
    if (!state->pot_filled) {
        analyseErrorLogAppend(la, ERROR_STEP_OMISSION, "FillPot", "espresso_pot_base", "water");
    }
    if (!state->filter_positioned) {
        analyseErrorLogAppend(la, ERROR_STEP_OMISSION, "PlaceFilter", "espresso_pot_filter", "espresso_pot_base");
    }
    if (!state->filter_filled) {
        analyseErrorLogAppend(la, ERROR_STEP_OMISSION, "FillFilter", "espresso_pot_filter", "coffee_grounds");
    }
    if (!state->pot_closed) {
        analyseErrorLogAppend(la, ERROR_STEP_OMISSION, "ClosePot", "espresso_pot_base", "espresso_pot_top");
    }
}

/******************************************************************************/

void analyse_espresso_task(DataAnalysis *la)
{
    ActionList *as;
    EspressoState state;

    state_initialise(&state);

    for (as = la->actionSequence; as; as = as->next) {
        as->checked = FALSE;
    }

    for (as = la->actionSequence; as; as = as->next) {
        if (!as->checked) {
            score_action(&state, la, as);
            as->checked = TRUE;
        }
    }

    score_omissions(&state, la);

    state_destroy(&state);
}

/******************************************************************************/
