/*******************************************************************************

    File: juice/action.c
    AUTHOR: R. Cooper
    CREATION DATE: Fri Jul 30 14:03:24 2004
    CONTENTS: Actions for the juice task

*******************************************************************************/

#include "../cs.h"
#include "task.h"

/******************************************************************************/

/* A local buffer used for actionMessage:                                     */
static char buffer[128]; 

/******************************************************************************/

void cut_by_sawing(ContentionScheduler *cs, Effector *hand)
{
    Object *target = cs->fixated_object;
    Object *knife = effectorGetContents(hand);

    char *hand_name = (hand == NULL ? "nothing" : hand->name);
    char *target_name = (target == NULL ? "nothing" : target->name);
    char *knife_name = (knife == NULL ? "nothing" : knife->name);

    g_snprintf(buffer, 128, "Cutting %s with %s by sawing", target_name, knife_name);
    actionMessage(cs, "CutBySawing", buffer);
    record_action(cs, "CutBySawing", target_name, knife_name, hand_name);

    if (hand == NULL) {
        errorEffectorNull(cs, "CutBySawing");
    }
    else if (target == NULL) {
        errorObjectNull(cs, "Target", "CutBySawing");
    }
    else if (knife == NULL) {
        errorEffectorState(cs, "CutBySawing: hand (%s) is empty", hand->name);
    }
    else if (isKnife(knife) < 0.9) {
        errorObjectIncorrect(cs, "%s: hand (%s) holds incorrect object %s", "CutBySawing", hand->name, knife->name);
    }
    else if (isKnifeBlade(knife) > 0.9) {
        errorObjectIncorrect(cs, "%s: hand (%s) holds the knife by the blade! (Holding %s.)", "CutBySawing", hand->name, knife->name);
    }
    else if (isOrange(target) < 0.9) {
        errorObjectIncorrect(cs, "%s: target (%s) is incorrect", "CutBySawing", target->name, NULL);
    }
    else {
        Object *piece = NULL;

        target->features = (target->features & ~IS_SPHERE) | IS_HEMISPHERE;

        g_snprintf(buffer, 128, "%s_2", target->name);
        piece = objectInstall(cs, buffer, target->features, target->state);
        objectSetPosition(piece, 0, 0);  /* "piece" is assumed to fall to table */
        traceObjectPostCreate(cs, piece);

        traceObjectPreAction(cs, target);
        g_snprintf(buffer, 128, "%s_1", target->name);
        free(target->name); target->name = string_copy(buffer);
        target->state = target->state | STATE_CUT;
        traceObjectPostAction(cs, target);
    }
}

/*----------------------------------------------------------------------------*/

void cut_by_pushing(ContentionScheduler *cs, Effector *hand)
{
    Object *target = cs->fixated_object;
    Object *knife = effectorGetContents(hand);

    char *hand_name = (hand == NULL ? "nothing" : hand->name);
    char *target_name = (target == NULL ? "nothing" : target->name);
    char *knife_name = (knife == NULL ? "nothing" : knife->name);

    g_snprintf(buffer, 128, "Cutting %s with %s by pushing", target_name, knife_name);
    actionMessage(cs, "CutByPushing", buffer);
    record_action(cs, "CutByPushing", target_name, knife_name, hand_name);

    if (hand == NULL) {
        errorEffectorNull(cs, "CutByPushing");
    }
    else if (knife == NULL) {
        errorEffectorState(cs, "CutByPushing: hand (%s) is empty", hand->name);
    }
    else if (isKnife(knife) < 0.9) {
        errorObjectIncorrect(cs, "%s: hand (%s) holds incorrect object %s", "CutByPushing", hand->name, knife->name);
    }
    else if (isOrange(target) < 0.9) {
        errorObjectIncorrect(cs, "%s: target (%s) is incorrect", "CutByPushing", target->name, NULL);
    }
    else {
        traceObjectPreAction(cs, target);
        /* Nothing happens ... we've selected the wrong cutting action! */
        traceObjectPostAction(cs, target);
    }
}

/*----------------------------------------------------------------------------*/

void squeeze_by_twisting(ContentionScheduler *cs, Effector *hand)
{
    Object *juicer = cs->fixated_object;
    Object *theme = effectorGetContents(hand);

    char *hand_name = (hand == NULL ? "nothing" : hand->name);
    char *juicer_name = (juicer == NULL ? "nothing" : juicer->name);
    char *theme_name = (theme == NULL ? "nothing" : theme->name);

    g_snprintf(buffer, 128, "Twisting %s on %s with %s", theme_name, juicer_name, hand_name);
    actionMessage(cs, "SqueezeByTwisting", buffer);
    record_action(cs, "SqueezeByTwisting", theme_name, juicer_name, hand_name);

    if (hand == NULL) {
        errorEffectorNull(cs, "SqueezeByTwisting");
    }
    else if (theme == NULL) {
        errorEffectorState(cs, "SqueezeByTwisting: hand (%s) is empty", hand->name);
    }
    else if (juicer == NULL) {
        errorObjectNull(cs, "Target", "SqueezeByPushing");
    }
    else if (isJuicer(juicer) < 0.9) {
        errorObjectIncorrect(cs, "%s: %s is not a sensible juicing implement", "SqueezeByTwisting", juicer->name, NULL);
    }
    else if (isOrangePiece(theme) < 0.9) {
        errorObjectIncorrect(cs, "%s: %s holds incorrect object (%s)", "SqueezeByTwisting", hand->name, theme->name);
    }
    else if (isSqueezed(theme)) {
        errorObjectIncorrect(cs, "%s: %s has already been squeezed", "SqueezeByTwisting", theme->name, NULL);
    }
    else {
        traceObjectPreAction(cs, juicer);
        objectInstallContents(juicer, objectCreate(cs, "orange_juice", IS_FLUID, 0));
        traceObjectPostAction(cs, juicer);

        traceObjectPreAction(cs, theme);
        theme->state = theme->state | STATE_USED;
        traceObjectPostAction(cs, theme);
    }
}

/*----------------------------------------------------------------------------*/

void squeeze_by_pushing(ContentionScheduler *cs, Effector *hand)
{
    Object *juicer = cs->fixated_object;
    Object *theme = effectorGetContents(hand);

    char *hand_name = (hand == NULL ? "nothing" : hand->name);
    char *juicer_name = (juicer == NULL ? "nothing" : juicer->name);
    char *theme_name = (theme == NULL ? "nothing" : theme->name);

    g_snprintf(buffer, 128, "Pushing %s on %s with %s", theme_name, juicer_name, hand_name);
    actionMessage(cs, "SqueezeByPushing", buffer);
    record_action(cs, "SqueezeByPushing", theme_name, juicer_name, hand_name);

    if (hand == NULL) {
        errorEffectorNull(cs, "SqueezeByPushing");
    }
    else if (theme == NULL) {
        errorEffectorState(cs, "SqueezeByPushing: hand (%s) is empty", hand->name);
    }
    else if (juicer == NULL) {
        errorObjectNull(cs, "Target", "SqueezeByPushing");
    }
    else if (isJuicer(juicer) < 0.9) {
        errorObjectIncorrect(cs, "%s: juicer (%s) is incorrect", "SqueezeByPushing", juicer->name, NULL);
    }
    else if (isOrangePiece(theme) < 0.9) {
        errorObjectIncorrect(cs, "%s: hand (%s) holds incorrect object %s", "SqueezeByPushing", hand->name, theme->name);
    }
    else if (isSqueezed(theme)) {
        errorObjectIncorrect(cs, "%s: orange (%s) has already been squeezed", "SqueezeByPushing", theme->name, NULL);
    }
    else {
        traceObjectPreAction(cs, juicer);
        objectInstallContents(juicer, objectCreate(cs, "orange_juice", IS_FLUID, 0));
        traceObjectPostAction(cs, juicer);

        traceObjectPreAction(cs, theme);
        theme->state = theme->state | STATE_USED;
        traceObjectPostAction(cs, theme);
    }
}

/*----------------------------------------------------------------------------*/

void squeeze_by_hand(ContentionScheduler *cs, Effector *hand)
{
    Object *target = cs->fixated_object;
    Object *theme = effectorGetContents(hand);

    char *hand_name = (hand == NULL ? "nothing" : hand->name);
    char *target_name = (target == NULL ? "nothing" : target->name);
    char *theme_name = (theme == NULL ? "nothing" : theme->name);

    g_snprintf(buffer, 128, "Squeezing %s into %s with %s", theme_name, target_name, hand_name);
    actionMessage(cs, "SqueezeByHand", buffer);
    record_action(cs, "SqueezeByHand", theme_name, target_name, hand_name);

    if (hand == NULL) {
        errorEffectorNull(cs, "SqueezeByHand");
    }
    else if (theme == NULL) {
        errorEffectorState(cs, "SqueezeByHand: hand (%s) is empty", hand->name);
    }
    else if (target == NULL) {
        errorObjectNull(cs, "Target", "SqueezeByHand");
    }
    else if (isContainer(target) < 0.9) {
        errorObjectIncorrect(cs, "%s: target (%s) is incorrect", "SqueezeByHand", target->name, NULL);
    }
    else if (isOrangePiece(theme) < 0.9) {
        errorObjectIncorrect(cs, "%s: hand (%s) holds incorrect object %s", "SqueezeByHand", hand->name, theme->name);
    }
    else if (isSqueezed(theme)) {
        errorObjectIncorrect(cs, "%s: orange (%s) has already been squeezed", "SqueezeByHand", theme->name, NULL);
    }
    else {
        traceObjectPreAction(cs, target);
        objectInstallContents(target, objectCreate(cs, "orange_juice", IS_FLUID, 0));
        traceObjectPostAction(cs, target);

        traceObjectPreAction(cs, theme);
        theme->state = theme->state | STATE_USED;
        traceObjectPostAction(cs, theme);
    }
}

/*----------------------------------------------------------------------------*/

/* Need to be holding the ball; Bounce it on the fixated object, which        */
/* hopefully is the table!                                                    */

void bounce(ContentionScheduler *cs, Effector *hand)
{
    Object *target = cs->fixated_object;
    Object *theme = effectorGetContents(hand);

    char *hand_name = (hand == NULL ? "nothing" : hand->name);
    char *target_name = (target == NULL ? "nothing" : target->name);
    char *theme_name = (theme == NULL ? "nothing" : theme->name);

    g_snprintf(buffer, 128, "Bouncing %s on %s with %s", theme_name, target_name, hand_name);
    actionMessage(cs, "Bounce", buffer);
    record_action(cs, "Bounce", theme_name, target_name, hand_name);

    if (hand == NULL) {
        errorEffectorNull(cs, "Bounce");
    }
    else if (theme == NULL) {
        errorEffectorState(cs, "Bounce: hand (%s) is empty", hand->name);
    }
    else if (target == NULL) {
        errorObjectNull(cs, "Target", "Bounce");
    }
    else {
        traceObjectPreAction(cs, theme);
        theme->state = theme->state | STATE_BOUNCED;
        traceObjectPostAction(cs, theme);
    }
}

/******************************************************************************/
/*************** pour(OpenContainer, HandContainingOpenSource) ****************/

/* Pouring moves all of the source to the target. Pouring fails if the thing  */
/* held by the hand isn't an open non-empty container or if the target is not */
/* an open container, or if the source and target are the same.               */

void pour_all(ContentionScheduler *cs, Effector *hand)
{
    Object *target = cs->fixated_object;
    Object *source = effectorGetContents(hand);
    Object *juice = objectGetContent(source);
    Object *lid;

    char *hand_name = (hand == NULL ? "nothing" : hand->name);
    char *target_name = (target == NULL ? "nothing" : target->name);
    char *source_name = (source == NULL ? "nothing" : source->name);
    char *juice_name = (juice == NULL ? "nothing" : juice->name);

    g_snprintf(buffer, 128, "Pouring %s from %s into %s", juice_name, source_name, target_name);
    actionMessage(cs, "Pour", buffer);
    record_action(cs, "Pour", source_name, target_name, hand_name);

    if (hand == NULL) {
        errorEffectorNull(cs, "Pour");
    }
    else if (target == NULL) {
        errorObjectNull(cs, "Target", "Pour");
    }
    else if (source == NULL) {
        errorEffectorState(cs, "Pour: hand (%s) is empty", hand->name);
    }
    else if (isClosedBoolean(source)) {
        errorObjectState(cs, "%s: source (%s) is closed", "Pour", source->name, NULL);
    }
    else if ((lid = objectGetCover(source)) != NULL) {
        errorObjectState(cs, "%s: source (%s) is covered by %s", "Pour", source->name, lid->name);
    }
    else if (isContainedBoolean(source)) {
        errorObjectState(cs, "%s: source (%s) is inside %s", "Pour", source->name, source->container->name);
    }
    else if (!isFullObjectBoolean(source)) {
        errorObjectState(cs, "%s: source (%s) is empty", "Pour", source->name, NULL);
    }
    else if (isClosedBoolean(target)) {
        errorObjectState(cs, "%s: target (%s) is closed", "Pour", target->name, NULL);
    }
    else if ((lid = objectGetCover(target)) != NULL) {
        errorObjectState(cs, "%s: target (%s) is covered by %s", "Pour", target->name, lid->name);
    }
    else if (isContainedBoolean(target)) {
        errorObjectState(cs, "%s: target (%s) is inside %s", "Pour", target->name, target->container->name);
    }
    else if (target == source) {
        errorObjectIncorrect(cs, "%s: target and source are same (%s = %s)", "Pour", target->name, source->name);
    }
    else if (!isContainerBoolean(target)) {
        errorObjectIncorrect(cs, "%s: target (%s) is not a container", "Pour", target->name, NULL);
    }
    else {
        ObjectList *tmp;

        /* For this task we pour the entire contents of the source into the target: */

        traceObjectPreAction(cs, source);
        tmp = source->itemsContained;
        source->itemsContained = tmp->tail;
        free(tmp);
        traceObjectPostAction(cs, source);

        traceObjectPreAction(cs, target);
        objectInstallContents(target, juice);
        traceObjectPostAction(cs, target);

        objectBindToRole(cs, target, TARGET);
    }
}

/******************************************************************************/
