/*******************************************************************************

    FILE: juice/analyse.c
    AUTHOR: R. Cooper
    CREATION DATE: Fri Jul 30 14:03:24 2004
    CONTENTS: Analysis routine for the juice task

*******************************************************************************/

#include "../cs.h"
#include "../analyse.h"

/******************************************************************************/

typedef enum juice_object {
    OBJECT_NONE, OBJECT_ORANGE, OBJECT_ORANGE_1, OBJECT_ORANGE_2, OBJECT_TABLE,
    OBJECT_KNIFE_HANDLE, OBJECT_KNIFE_BLADE, OBJECT_GLASS, OBJECT_JUICER, OBJECT_FILTER,
    OBJECT_ORANGE_JUICE, OBJECT_UNKNOWN
} JuiceObject;

typedef enum juice_location {
    LOC_NONE, LOC_HELD, LOC_ON_TABLE, LOC_ON_JUICER
} JuiceLocation;

typedef struct juice_analysis_object_list {
    JuiceObject                        id;        /* Enumerated id */
    JuiceLocation                      location;
    int                                state;
    struct juice_analysis_object_list *next;      /* Next object in the list  */
} JuiceAnalysisObjectList;

typedef struct juice_state {
    // Information on all possible objects in the world
    JuiceAnalysisObjectList *items; 

    //Fixated item:
    JuiceObject              fixated;

    // Other relevant state information:
    Boolean                  orange_squeezed;
    Boolean                  glass_filled;

} JuiceState;

#define JUICE_STATE_SQUEEZED          1
#define JUICE_STATE_CONTAINS_JUICE    2

/******************************************************************************/

static JuiceObject object_id_from_string(char *arg)
{
    if (arg == NULL) {
        return(OBJECT_NONE);
    }
    else if (strcmp(arg, "nothing") == 0) {
        return(OBJECT_NONE);
    }
    else if (strcmp(arg, "glass") == 0) {
        return(OBJECT_GLASS);
    }
    else if (strcmp(arg, "knife_handle") == 0) {
        return(OBJECT_KNIFE_HANDLE);
    }
    else if (strcmp(arg, "knife_blade") == 0) {
        return(OBJECT_KNIFE_BLADE);
    }
    else if (strcmp(arg, "juicer") == 0) {
        return(OBJECT_JUICER);
    }
    else if (strcmp(arg, "filter") == 0) {
        return(OBJECT_FILTER);
    }
    else if (strcmp(arg, "table") == 0) {
        return(OBJECT_TABLE);
    }
    else if (strcmp(arg, "orange") == 0) {
        return(OBJECT_ORANGE);
    }
    else if (strcmp(arg, "orange_juice") == 0) {
        return(OBJECT_ORANGE_JUICE);
    }
    else if (strcmp(arg, "orange_1") == 0) {
        return(OBJECT_ORANGE_1);
    }
    else if (strcmp(arg, "orange_2") == 0) {
        return(OBJECT_ORANGE_2);
    }
    /* Final catch-all case: */

    else {
        report_unrecognised_object(stdout, arg);
        return(OBJECT_UNKNOWN);
    }
}

static JuiceAnalysisObjectList *object_ptr_from_id(JuiceState *state, JuiceObject id)
{
    JuiceAnalysisObjectList *tmp = state->items;

    while ((tmp != NULL) && (tmp->id != id)) {
        tmp = tmp->next;
    }
    if ((tmp == NULL) || (tmp->id != id)) {
        return(NULL);
    }
    else {
        return(tmp);
    }
}

static JuiceAnalysisObjectList *object_ptr_from_string(JuiceState *state, char *arg)
{
    return(object_ptr_from_id(state, object_id_from_string(arg)));
}

static Boolean action_compare(ActionList *this, char *action, JuiceObject arg1)
{
    if (this == NULL) {
        return(FALSE);
    }
    else if (strcmp(this->name, action) != 0) {
        return(FALSE);
    }
    else if (object_id_from_string(this->argument1) != arg1) {
        return(FALSE);
    }
    else {
        return(TRUE);
    }
}

/******************************************************************************/

static void state_add_item(JuiceState *state, JuiceObject id, JuiceLocation loc)
{
    JuiceAnalysisObjectList *new = (JuiceAnalysisObjectList *)malloc(sizeof(JuiceAnalysisObjectList));

    if (new != NULL) {
        new->id = id;
        new->location = loc;
        new->state = 0;
        new->next = state->items;
        state->items = new;
    }
}

static void state_replace_item(JuiceState *state, JuiceObject id1, JuiceObject id2)
{
    JuiceAnalysisObjectList *item = object_ptr_from_id(state, id1);
    item->id = id2;
}

static void state_initialise(JuiceState *state)
{
    state->items = NULL;

    /* All objects must be initialised to match juice/object.c */

    state_add_item(state, OBJECT_GLASS, LOC_ON_TABLE);
    state_add_item(state, OBJECT_KNIFE_HANDLE, LOC_ON_TABLE);
    state_add_item(state, OBJECT_KNIFE_BLADE, LOC_ON_TABLE);
    state_add_item(state, OBJECT_ORANGE, LOC_ON_TABLE);
    state_add_item(state, OBJECT_FILTER, LOC_ON_JUICER);
    state_add_item(state, OBJECT_JUICER, LOC_ON_TABLE);
    state_add_item(state, OBJECT_TABLE, LOC_NONE);

    state->fixated = OBJECT_NONE;

    state->orange_squeezed = FALSE;
    state->glass_filled = FALSE;
}

static void state_destroy(JuiceState *state)
{
    while (state->items != NULL) {
        JuiceAnalysisObjectList *tmp = state->items;
        state->items = tmp->next;
        free(tmp);
    }
}

/******************************************************************************/

static void check_consecutive_toying_actions(ActionList *this, JuiceObject arg1)
{
    if (action_compare(this->next, "PickUp", arg1)) {
        if (action_compare(this->next->next, "PutDown", arg1)) {
            this->next->checked = TRUE;
            this->next->next->checked = TRUE;
            check_consecutive_toying_actions(this->next->next, arg1);
        }
        else if (action_compare(this->next->next, "Discard", arg1)) {
            this->next->checked = TRUE;
            this->next->next->checked = TRUE;
            check_consecutive_toying_actions(this->next->next, arg1);
        }
    }
}

/******************************************************************************/

static void score_action_fixate(JuiceState *state, DataAnalysis *la, ActionList *action, char *arg1)
{
    JuiceAnalysisObjectList *item1;

    if (object_id_from_string(arg1) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "Fixate", arg1, NULL);
        ignore_consecutive_actions(action);
    }
    else if ((item1 = object_ptr_from_string(state, arg1)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else {
        state->fixated = item1->id;
    }
}

/*----------------------------------------------------------------------------*/

static void score_action_pickup(JuiceState *state, DataAnalysis *la, ActionList *action, char *arg1)
{
    JuiceAnalysisObjectList *item1;

    if (object_id_from_string(arg1) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "PickUp", arg1, NULL);
        ignore_consecutive_actions(action);
    }
    else if ((item1 = object_ptr_from_string(state, arg1)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if (item1->location == LOC_HELD) {
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "PickUp", arg1, NULL);
        ignore_consecutive_actions(action);
    }
    else if (item1->id == OBJECT_TABLE) {
        // Score PERPLEXITY: Object cannot be picked up
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "PickUp", arg1, NULL);
        ignore_consecutive_actions(action);
    }
    else {
        item1->location = LOC_HELD;
        if (item1->id == OBJECT_KNIFE_BLADE) {
            // Score MISLOCATION_2 error: picked up knife by blade
            analyseErrorLogAppend(la, ERROR_MISLOCATION_2, "PickUp", arg1, NULL);
        }
        if (action_compare(action->next, "PutDown", item1->id)) {
            // Score TOYING error and ignore any consecutive toying
            analyseErrorLogAppend(la, ERROR_TOYING, "PickUp/PutDown", arg1, NULL);
            check_consecutive_toying_actions(action->next, item1->id);
        }
        else if (action_compare(action->next, "Discard", item1->id)) {
            // Score TOYING error and ignore any consecutive toying
            analyseErrorLogAppend(la, ERROR_TOYING, "PickUp/Discard", arg1, NULL);
            check_consecutive_toying_actions(action->next, item1->id);
        }
    }
}

/*----------------------------------------------------------------------------*/

static void score_action_putdown(JuiceState *state, DataAnalysis *la, ActionList *action, char *arg1)
{
    JuiceAnalysisObjectList *item1;

    if (object_id_from_string(arg1) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "PutDown", arg1, NULL);
        ignore_consecutive_actions(action);
    }
    else if ((item1 = object_ptr_from_string(state, arg1)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if (item1->location != LOC_HELD) {
        // Score PERPLEXITY: Object is not held
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "PutDown", arg1, NULL);
        ignore_consecutive_actions(action);
    }
    else {
        item1->location = LOC_ON_TABLE;
    }
}

/*----------------------------------------------------------------------------*/

static void score_action_discard(JuiceState *state, DataAnalysis *la, ActionList *action, char *arg1)
{
    JuiceAnalysisObjectList *item1;

    if (object_id_from_string(arg1) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "Discard", arg1, NULL);
        ignore_consecutive_actions(action);
    }
    else if ((item1 = object_ptr_from_string(state, arg1)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if (item1->location != LOC_HELD) {
        // Score PERPLEXITY: Object is not held
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "Discard", arg1, NULL);
        ignore_consecutive_actions(action);
    }
    else {
        item1->location = LOC_ON_TABLE;
    }
}

/*----------------------------------------------------------------------------*/

static void score_action_pour(JuiceState *state, DataAnalysis *la, ActionList *action, char *arg1, char *arg2)
{
    JuiceAnalysisObjectList *item1; // Source
    JuiceAnalysisObjectList *item2; // Target

    if (object_id_from_string(arg1) == OBJECT_NONE) {
        // Score PANTOMIME: trying to pour without a source
        analyseErrorLogAppend(la, ERROR_PANTOMIME, "Pour01", arg1, arg2);
    }
    else if (object_id_from_string(arg2) == OBJECT_NONE) {
        // Score PANTOMIME: trying to pour without a target
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "Pour02", arg1, arg2);
    }
    else if ((item1 = object_ptr_from_string(state, arg1)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if ((item2 = object_ptr_from_string(state, arg2)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if (item1 == item2) {
        // Score PERPLEXITY: trying to pour container into itself
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "Pour04", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if (item1->location != LOC_HELD) {
        // Score PERPLEXITY: source container isn't held
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "Pour05", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if ((item1->state & JUICE_STATE_CONTAINS_JUICE) == 0) {
        if ((item1->id == OBJECT_JUICER) || (item1->id == OBJECT_GLASS)) {
            // Score ANTICIPATION error: sensible source poured before being filled with juice
            analyseErrorLogAppend(la, ERROR_ACTION_ANTICIPATION, "Pour06", arg1, arg2);
        }
        else {
            // Score MISUSE_1 error:  Pouring bizarre/inappropriate object 
            analyseErrorLogAppend(la, ERROR_MISUSE_1, "Pour07", arg1, arg2);
        }
    }
    else {
        if (item2->id != OBJECT_GLASS) {
            // Score MISLOCATION1 error: target isn't the glass
            analyseErrorLogAppend(la, ERROR_MISLOCATION_1, "Pour08", arg1, arg2);
        }
        else {
            state->glass_filled = TRUE;
        }
        item1->state = item1->state & ~JUICE_STATE_CONTAINS_JUICE;
        item2->state = item2->state | JUICE_STATE_CONTAINS_JUICE;
    }
}

/*----------------------------------------------------------------------------*/

static void score_action_cut_by_sawing(JuiceState *state, DataAnalysis *la, ActionList *action, char *arg1, char *arg2, char *arg3)
{
    JuiceAnalysisObjectList *item1; // Target
    JuiceAnalysisObjectList *item2; // Tool

    if (object_id_from_string(arg2) == OBJECT_NONE) {
        // Score PANTOMIME: trying to saw without a saw!
        analyseErrorLogAppend(la, ERROR_PANTOMIME, "CutBySawing01", arg1, arg2);
    }
    else if (object_id_from_string(arg1) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "CutBySawing02", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if ((item1 = object_ptr_from_string(state, arg1)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if ((item2 = object_ptr_from_string(state, arg2)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if (item1 == item2) {
        // Score PERPLEXITY: trying to cut object with itself
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "CutBySawing03", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if (item2->id == OBJECT_NONE) {
        if (strcmp(arg3, "nothing") == 0) {
            analyseErrorLogAppend(la, ERROR_PERPLEXITY, "CutBySawing04", arg1, arg2);
            ignore_consecutive_actions(action);
        }
        else {
            analyseErrorLogAppend(la, ERROR_PANTOMIME, "CutBySawing05", arg1, arg2);
        }
    }
    else if (item2->location != LOC_HELD) {
        // Score PERPLEXITY: Tool is not held
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "CutBySawing06", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if (item2->id == OBJECT_KNIFE_BLADE) {
        // Score MISLOCATION_2 error: holding knife by blade
        analyseErrorLogAppend(la, ERROR_MISLOCATION_2, "CutBySawing07", arg1, arg2);
    }
    else if (item2->id != OBJECT_KNIFE_HANDLE) {
        // Score MISUSE_1 error: cutting with something other than a knife
        analyseErrorLogAppend(la, ERROR_MISUSE_1, "CutBySawing08", arg1, arg2);
    }
    else if ((item1->id == OBJECT_ORANGE_1) || (item1->id == OBJECT_ORANGE_2)) {
        // Score PERSEVERATION error: orange is already cut
        analyseErrorLogAppend(la, ERROR_PERSEVERATION, "CutBySawing09", arg1, arg2);
    }
    else if (item1->id != OBJECT_ORANGE) {
        // Score MISLOCATION_1 error: cutting something other than the orange!
        analyseErrorLogAppend(la, ERROR_MISLOCATION_1, "CutBySawing10", arg1, arg2);
    }
    else {
        state_replace_item(state, OBJECT_ORANGE, OBJECT_ORANGE_1);
        state_add_item(state, OBJECT_ORANGE_2, LOC_ON_TABLE);
    }
}

/*----------------------------------------------------------------------------*/

static void score_action_cut_by_pushing(JuiceState *state, DataAnalysis *la, ActionList *action, char *arg1, char *arg2, char *arg3)
{
    JuiceAnalysisObjectList *item1; // Target
    JuiceAnalysisObjectList *item2; // Tool

    if (object_id_from_string(arg2) == OBJECT_NONE) {
        // Score PANTOMIME: trying to cut without a tool!
        analyseErrorLogAppend(la, ERROR_PANTOMIME, "CutByPushing01", arg1, arg2);
    }
    else if (object_id_from_string(arg1) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "CutByPushing02", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if ((item1 = object_ptr_from_string(state, arg1)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if ((item2 = object_ptr_from_string(state, arg2)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if (item2->id == OBJECT_NONE) {
        if (strcmp(arg3, "nothing") == 0) {
            analyseErrorLogAppend(la, ERROR_PERPLEXITY, "CutByPushing03", arg1, arg2);
            ignore_consecutive_actions(action);
        }
        else {
            analyseErrorLogAppend(la, ERROR_PANTOMIME, "CutByPushing04", arg1, arg2);
        }
    }
    else if (item2->location != LOC_HELD) {
        // Score PERPLEXITY: Tool is not held
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "CutByPushing05", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if (item1 == item2) {
        // Score PERPLEXITY: trying to cut object with itself
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "CutByPushing06", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if (item2->id == OBJECT_KNIFE_BLADE) {
        // Score MISLOCATION_2 error: holding knife by blade
        analyseErrorLogAppend(la, ERROR_MISLOCATION_2, "CutByPushing07", arg1, arg2);
    }
    else if (item2->id != OBJECT_KNIFE_HANDLE) {
        // Score MISUSE_1 error: cutting with something other than a knife
        analyseErrorLogAppend(la, ERROR_MISUSE_1, "CutByPushing08", arg1, arg2);
    }
    else if ((item1->id == OBJECT_ORANGE_1) || (item1->id == OBJECT_ORANGE_2)) {
        // Score PERSEVERATION error: orange is already cut
        analyseErrorLogAppend(la, ERROR_PERSEVERATION, "CutByPushing09", arg1, arg2);
    }
    else if (item1->id != OBJECT_ORANGE) {
        // Score MISLOCATION_1 error: cutting something other than the orange!
        analyseErrorLogAppend(la, ERROR_MISLOCATION_1, "CutByPushing10", arg1, arg2);
    }
    else {
        // Score MISUSE_2 error: Should not cut the orange by pushing
        analyseErrorLogAppend(la, ERROR_MISUSE_2, "CutByPushing11", arg1, arg2);
    }
}

/*----------------------------------------------------------------------------*/

static void score_action_squeeze_by_twisting(JuiceState *state, DataAnalysis *la, ActionList *action, char *arg1, char *arg2)
{
    JuiceAnalysisObjectList *item1; // Thing being squeezed
    JuiceAnalysisObjectList *item2; // Tool on which it is twisted

    if (object_id_from_string(arg2) == OBJECT_NONE) {
        // Score PANTOMIME: trying to squeeze without a tool!
        analyseErrorLogAppend(la, ERROR_PANTOMIME, "SqueezeByTwisting", arg1, arg2);
    }
    else if (object_id_from_string(arg1) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "SqueezeByTwisting", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if ((item1 = object_ptr_from_string(state, arg1)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if ((item2 = object_ptr_from_string(state, arg2)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if (item1->location != LOC_HELD) {
        // Score PERPLEXITY: Theme is not held
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "SqueezeByTwisting", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if (item1 == item2) {
        // Score PERPLEXITY: trying to squeeze object with itself
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "SqueezeByTwisting", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if (item1->id == OBJECT_ORANGE) {
        // Score OMISSION error: attempting to squeeze the whole orange!
        analyseErrorLogAppend(la, ERROR_STEP_OMISSION, "SqueezeByTwisting", arg1, arg2);
    }
    else if (item2->id != OBJECT_JUICER) {
        // Score MISLOCATION_1 error: twisting something but not on the juicer
        analyseErrorLogAppend(la, ERROR_MISLOCATION_1, "SqueezeByTwisting", arg1, arg2);
    }
    else if ((item1->id != OBJECT_ORANGE_1) && (item1->id != OBJECT_ORANGE_2)) {
        // Score MISUSE_1 error:  Misuse of item1
        analyseErrorLogAppend(la, ERROR_MISUSE_1, "SqueezeByTwisting", arg1, arg2);
    }
    else {
        item1->state = item1->state | JUICE_STATE_SQUEEZED;
        item2->state = item2->state | JUICE_STATE_CONTAINS_JUICE;
        state->orange_squeezed = TRUE;
    }
}

/*----------------------------------------------------------------------------*/

static void score_action_squeeze_by_pushing(JuiceState *state, DataAnalysis *la, ActionList *action, char *arg1, char *arg2)
{
    JuiceAnalysisObjectList *item1; // Thing being squeezed
    JuiceAnalysisObjectList *item2; // Tool on which it is pushed

    if (object_id_from_string(arg2) == OBJECT_NONE) {
        // Score PANTOMIME: trying to squeeze without a tool!
        analyseErrorLogAppend(la, ERROR_PANTOMIME, "SqueezeByPushing", arg1, arg2);
    }
    else if (object_id_from_string(arg1) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "SqueezeByPushing", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if ((item1 = object_ptr_from_string(state, arg1)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if ((item2 = object_ptr_from_string(state, arg2)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if (item1->location != LOC_HELD) {
        // Score PERPLEXITY: Theme is not held
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "SqueezeByPushing", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if (item1 == item2) {
        // Score PERPLEXITY: trying to squeeze object with itself
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "SqueezeByPushing", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if (item1->id == OBJECT_ORANGE) {
        // Score OMISSION error: attempting to squeeze the whole orange!
        analyseErrorLogAppend(la, ERROR_STEP_OMISSION, "SqueezeByPushing", arg1, arg2);
    }
    else if (item2->id != OBJECT_JUICER) {
        // Score MISLOCATION_1 error: pushing but not on the juicer
        analyseErrorLogAppend(la, ERROR_MISLOCATION_1, "SqueezeByPushing", arg1, arg2);
    }
    else if ((item1->id != OBJECT_ORANGE_1) && (item1->id != OBJECT_ORANGE_2)) {
        // Score MISUSE_1 error:  Misuse of item1
        analyseErrorLogAppend(la, ERROR_MISUSE_1, "SqueezeByPushing", arg1, arg2);
    }
    else {
        item1->state = item1->state | JUICE_STATE_SQUEEZED;
        item2->state = item2->state | JUICE_STATE_CONTAINS_JUICE;
        state->orange_squeezed = TRUE;
        // Score MISUSE_2 error: Should not squeeze the orange by pushing
        analyseErrorLogAppend(la, ERROR_MISUSE_2, "SqueezeByPushing", arg1, arg2);
    }
}

/*----------------------------------------------------------------------------*/

static void score_action_squeeze_by_hand(JuiceState *state, DataAnalysis *la, ActionList *action, char *arg1, char *arg2)
{
    JuiceAnalysisObjectList *item1; // Orange piece
    JuiceAnalysisObjectList *item2; // Target (glass of perhaps juicer)

    if (object_id_from_string(arg1) == OBJECT_NONE) {
        // Score PANTOMIME: trying to squeeze without an orange!
        analyseErrorLogAppend(la, ERROR_PANTOMIME, "SqueezeByHand", arg1, arg2);
    }
    else if (object_id_from_string(arg2) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "SqueezeByHand", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if ((item1 = object_ptr_from_string(state, arg1)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if ((item2 = object_ptr_from_string(state, arg2)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if (item1->location != LOC_HELD) {
        // Score PERPLEXITY: Theme is not held
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "SqueezeByHand", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if (item1 == item2) {
        // Score PERPLEXITY: trying to squeeze object into itself
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "SqueezeByHand", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if (item1->id == OBJECT_ORANGE) {
        // Score OMISSION error: attempting to squeeze the whole (uncat) orange!
        analyseErrorLogAppend(la, ERROR_STEP_OMISSION, "SqueezeByHand", arg1, arg2);
    }
    else if ((item2->id != OBJECT_GLASS) && (item2->id != OBJECT_JUICER)) {
        // Score MISLOCATION_1 error: squeezing into a non-container
        analyseErrorLogAppend(la, ERROR_MISLOCATION_1, "SqueezeByHand", arg1, arg2);
    }
    else {
        // Score TOOL_OMISSION error:  Orange squeezed by hand!
        analyseErrorLogAppend(la, ERROR_TOOL_OMISSION, "SqueezeByHand", arg1, arg2);
        item1->state = item1->state | JUICE_STATE_SQUEEZED;
        item2->state = item2->state | JUICE_STATE_CONTAINS_JUICE;
        state->orange_squeezed = TRUE;
        if (item2->id == OBJECT_GLASS) {
            state->glass_filled = TRUE;
        }
    }
}

/*----------------------------------------------------------------------------*/

static void score_action_bounce(JuiceState *state, DataAnalysis *la, ActionList *action, char *arg1, char *arg2)
{
    JuiceAnalysisObjectList *item1; // Thing being bounced
    JuiceAnalysisObjectList *item2; // Target of the bounce

    if (object_id_from_string(arg1) == OBJECT_NONE) {
        // Score PANTOMIME: trying to bounce nothing!
        analyseErrorLogAppend(la, ERROR_PANTOMIME, "Bounce", arg1, arg2);
    }
    else if ((item1 = object_ptr_from_string(state, arg1)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if (object_id_from_string(arg2) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "Bounce", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if ((item2 = object_ptr_from_string(state, arg2)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if (item1->location != LOC_HELD) {
        // Score PERPLEXITY: Theme is not held
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "Bounce", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if (item1 == item2) {
        // Score PERPLEXITY: trying to bounce object on itself
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "Bounce", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else {
        // Score ADDITION error: Bouncing something!
        analyseErrorLogAppend(la, ERROR_ACTION_ADDITION, "Bounce", arg1, arg2);

        if (item1->id != OBJECT_ORANGE) {
            // Score MISUSE_1 error: Attempting to bounce a non-spherical object
            analyseErrorLogAppend(la, ERROR_MISUSE_1, "Bounce", arg1, arg2);
        }
        else if (item2->id != OBJECT_TABLE) {
            // Score MISLOCATION_1 error: Attempting to bounce on something other than the table
            analyseErrorLogAppend(la, ERROR_MISLOCATION_1, "Bounce", arg1, arg2);
        }
    }
}

static void score_action_drink(JuiceState *state, DataAnalysis *la, ActionList *action, char *arg1, char *arg2)
{
    // Drinking arg1 (the fluid) from arg2 (the container)

    JuiceAnalysisObjectList *item2;

    if (object_id_from_string(arg2) == OBJECT_NONE) {
        // Score PANTOMIME: trying to drink without a container!
        analyseErrorLogAppend(la, ERROR_PANTOMIME, "Drink", arg1, arg2);
    }
    else if ((item2 = object_ptr_from_string(state, arg2)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if (item2->location != LOC_HELD) {
        // Score PERPLEXITY: Container is not held
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "Drink", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if ((item2->state & JUICE_STATE_CONTAINS_JUICE) == 0) {
        if ((item2->id == OBJECT_GLASS) || (item2->id == OBJECT_JUICER)) {
            // Score ANTICIPATION error: Drinking from sensible but empty container!
            analyseErrorLogAppend(la, ERROR_ACTION_ANTICIPATION, "Drink", arg1, arg2);
        }
        else {
            // Score MISUSE_1 error: Attempting to drink from a non-container!
            analyseErrorLogAppend(la, ERROR_MISUSE_1, "Drink", arg1, arg2);
        }
    }
    else {
        // Score ADDITION error: Drinking!
        analyseErrorLogAppend(la, ERROR_ACTION_ADDITION, "Drink", arg1, arg2);
    }
}

/******************************************************************************/

static void score_action(JuiceState *state, DataAnalysis *la, ActionList *action)
{
    if (action == NULL) {
        fprintf(stderr, "INTERNAL ERROR: NULL action in score_action/3 (file is juice/analyse.c)\n");
    }
    else if (strcmp(action->name, "Fixate") == 0) {
        score_action_fixate(state, la, action, action->argument1);
    }
    else if (strcmp(action->name, "PickUp") == 0) {
        score_action_pickup(state, la, action, action->argument1);
    }
    else if (strcmp(action->name, "PutDown") == 0) {
        score_action_putdown(state, la, action, action->argument1);
    }
    else if (strcmp(action->name, "Discard") == 0) {
        score_action_discard(state, la, action, action->argument1);
    }
    else if (strcmp(action->name, "Drink") == 0) {
        score_action_drink(state, la, action, action->argument1, action->argument2);
    }
    else if (strcmp(action->name, "CutBySawing") == 0) {
        score_action_cut_by_sawing(state, la, action, action->argument1, action->argument2, action->argument3);
    }
    else if (strcmp(action->name, "CutByPushing") == 0) {
        score_action_cut_by_pushing(state, la, action, action->argument1, action->argument2, action->argument3);
    }
    else if (strcmp(action->name, "SqueezeByTwisting") == 0) {
        score_action_squeeze_by_twisting(state, la, action, action->argument1, action->argument2);
    }
    else if (strcmp(action->name, "SqueezeByPushing") == 0) {
        score_action_squeeze_by_pushing(state, la, action, action->argument1, action->argument2);
    }
    else if (strcmp(action->name, "SqueezeByHand") == 0) {
        score_action_squeeze_by_hand(state, la, action, action->argument1, action->argument2);
    }
    else if (strcmp(action->name, "Pour") == 0) {
        score_action_pour(state, la, action, action->argument1, action->argument2);
    }
    else if (strcmp(action->name, "Bounce") == 0) {
        score_action_bounce(state, la, action, action->argument1, action->argument2);
    }
    else {
        report_unrecognised_action(stdout, action);
    }
}

/******************************************************************************/

static void score_omissions(JuiceState *state, DataAnalysis *la)
{
    if (!state->orange_squeezed) {
        analyseErrorLogAppend(la, ERROR_STEP_OMISSION, "SqueezeOrange", "orange_piece", NULL);
    }
    if (!state->glass_filled) {
        analyseErrorLogAppend(la, ERROR_STEP_OMISSION, "FillGlass", "glass", "juice");
    }
}

/******************************************************************************/

void analyse_juice_task(DataAnalysis *la)
{
    ActionList *as;
    JuiceState state;

    state_initialise(&state);

    for (as = la->actionSequence; as; as = as->next) {
        as->checked = FALSE;
    }

    for (as = la->actionSequence; as; as = as->next) {
        if (!as->checked) {
            score_action(&state, la, as);
            as->checked = TRUE;
        }
    }

    score_omissions(&state, la);

    state_destroy(&state);
}

/******************************************************************************/
