/*******************************************************************************

    FILE: letter_action.c
    AUTHOR: R. Cooper
    CREATION DATE: Thu Nov 28 15:24:45 2002
    CONTENTS: Actions for the letter task

*******************************************************************************/

#include "../cs.h"
#include "task.h"

/******************************************************************************/

/* A local buffer used for actionMessage:                                     */
static char buffer[128]; 

/******************************************************************************/

void place_stamp(ContentionScheduler *cs, Effector *hand)
{
    Object *envelope = cs->fixated_object;
    Object *stamp = effectorGetContents(hand);

    char *envelope_name = (envelope == NULL ? "nothing" : envelope->name);
    char *hand_name = (hand == NULL ? "nothing" : hand->name);
    char *stamp_name = (stamp == NULL ? "nothing" : stamp->name);

    g_snprintf(buffer, 128, "Applying %s (held by %s) to %s", stamp_name, hand_name, envelope_name);
    actionMessage(cs, "ApplyStamp", buffer);
    record_action(cs, "ApplyStamp", stamp_name, envelope_name, hand_name);

    if (hand == NULL) {
        errorEffectorNull(cs, "PlaceStamp");
    }
    else if (stamp == NULL) {
        errorEffectorState(cs, "PlaceStamp: hand (%s) is empty", hand->name);
    }
    else if ((stamp == envelope) || (stamp->base == envelope) || (envelope->base == stamp)) {
        errorObjectIncorrect(cs, "%s: stamp (%s) and envelope (%s) are incompatible", "PlaceStamp", stamp->name, envelope_name);
    }
    else if (isStamp(stamp) < 0.9) {
        errorObjectIncorrect(cs, "%s: hand (%s) holds incorrect object (%s)", "PlaceStamp", hand->name, stamp->name);
    }
    else if (!isFlatBoolean(envelope)) {
        errorObjectIncorrect(cs, "%s: target (%s) is incorrect", "PlaceStamp", envelope->name, NULL);
    }
    else if (isMoist(stamp) < 0.9) {
        errorObjectState(cs, "%s: stamp (%s) has not been moistened", "PlaceStamp", stamp->name, NULL);
    }
    else {
        traceObjectPreAction(cs, stamp);
        hand->itemsHeld = objectListDeleteElement(hand->itemsHeld, stamp);
        stamp->holder = NULL;
        stamp->state = stamp->state & ~STATE_MOIST;
        traceObjectPostAction(cs, stamp);

        if (isStampBack(stamp) > 0.9) {
            // If it sticks to the envelope (i.e., the sticky back is
            // applied to the envelope) then we remove the stamp...
            cs->objects = objectListDeleteElement(cs->objects, stamp);
            cs->objects = objectListDeleteElement(cs->objects, stamp->base);
            objectDestroy(cs, stamp->base);

            traceObjectPreAction(cs, envelope);
            envelope->state = envelope->state | STATE_STAMPED;
            traceObjectPostAction(cs, envelope);
        }
    }
}

/*----------------------------------------------------------------------------*/

void fold_letter(ContentionScheduler *cs, Effector *full_hand)
{
    Object *letter = effectorGetContents(full_hand);

    char *letter_name = (letter == NULL ? "nothing" : letter->name);
    char *full_hand_name = (full_hand == NULL ? "nothing" : full_hand->name);

    g_snprintf(buffer, 128, "Folding %s (with %s)", letter_name, full_hand_name);
    actionMessage(cs, "FoldLetter", buffer);
    record_action(cs, "FoldLetter", letter_name, full_hand_name, NULL);

    if (full_hand == NULL) {
        errorEffectorNull(cs, "FoldLetter");
    }
    else if (letter == NULL) {
        errorEffectorState(cs, "FoldLetter: hand (%s) is empty", full_hand->name);
    }
    else if (!isFlatBoolean(letter)) {
        errorObjectIncorrect(cs, "%s: hand (%s) holds incorrect object %s", "FoldLetter", full_hand->name, letter->name);
    }
    else {
        Effector *other_hand = effectorGetOtherHand(cs, full_hand);

        if (isFullEffectorBoolean(other_hand)) {
            effectorDiscardContents(cs, other_hand);
        }

        traceObjectPreAction(cs, letter);
        letter->state = letter->state | STATE_FOLDED;
        traceObjectPostAction(cs, letter);
    }
}

/*----------------------------------------------------------------------------*/

void write_address(ContentionScheduler *cs, Effector *hand)
{
    Object *envelope = cs->fixated_object;
    Object *pen = effectorGetContents(hand);

    char *envelope_name = (envelope == NULL ? "nothing" : envelope->name);
    char *pen_name = (pen == NULL ? "nothing" : pen->name);
    char *hand_name = (hand == NULL ? "nothing" : hand->name);

    g_snprintf(buffer, 128, "Writing address on %s with %s (held by %s)", envelope_name, pen_name, hand_name);
    actionMessage(cs, "WriteAddress", buffer);
    record_action(cs, "WriteAddress", envelope_name, pen_name, hand_name);

    if (hand == NULL) {
        errorEffectorNull(cs, "WriteAddress");
    }
    else if (pen == NULL) {
        errorEffectorState(cs, "WriteAddress: hand (%s) is empty", hand->name);
    }
    else if (!isToolBoolean(pen)) {
        errorObjectIncorrect(cs, "%s: hand (%s) holds incorrect object %s", "WriteAddress", hand->name, pen->name);
    }
    else {
        traceObjectPreAction(cs, envelope);
        envelope->state = envelope->state | STATE_ADDRESSED;
        traceObjectPostAction(cs, envelope);
    }
}

/*----------------------------------------------------------------------------*/

void pack_envelope(ContentionScheduler *cs, Effector *envelope_hand, Effector *letter_hand)
{
    Object *envelope = effectorGetContents(envelope_hand);
    Object *letter = effectorGetContents(letter_hand);

    char *envelope_name = (envelope == NULL ? "nothing" : envelope->name);
    char *letter_name = (letter == NULL ? "nothing" : letter->name);

    g_snprintf(buffer, 128, "Packing %s into %s", letter_name, envelope_name);
    actionMessage(cs, "PackEnvelope", buffer);
    record_action(cs, "PackEnvelope", letter_name, envelope_name, NULL);

    if ((envelope_hand == NULL) || (letter_hand == NULL)) {
        errorEffectorNull(cs, "PackEnvelope");
    }
    else if (envelope == NULL) {
        errorEffectorState(cs, "PackEnvelope: hand (%s) is empty", envelope_hand->name);
    }
    else if (letter == NULL) {
        errorEffectorState(cs, "PackEnvelope: hand (%s) is empty", letter_hand->name);
    }
    else if (!isFlatBoolean(letter)) {
        errorObjectIncorrect(cs, "%s: hand (%s) holds incorrect object %s", "PackEnvelope", letter_hand->name, letter->name);
    }
    else if (!isContainerBoolean(envelope)) {
        errorObjectIncorrect(cs, "%s: hand (%s) holds incorrect object %s", "PackEnvelope", letter_hand->name, envelope->name);
    }
    else {
        traceObjectPreAction(cs, letter);
        letter_hand->itemsHeld = objectListDeleteElement(letter_hand->itemsHeld, letter);
        letter->holder = NULL;
        traceObjectPostAction(cs, letter);

        traceObjectPreAction(cs, envelope);
        objectInstallContents(envelope, letter);
        traceObjectPostAction(cs, envelope);

        cs->objects = objectListDeleteElement(cs->objects, letter);
    }
}

/*----------------------------------------------------------------------------*/

void seal_envelope(ContentionScheduler *cs, Effector *hand)
{
    Object *theme = effectorGetContents(hand);

    char *hand_name = (hand == NULL ? "nothing" : hand->name);
    char *theme_name = (theme == NULL ? "nothing" : theme->name);

    g_snprintf(buffer, 128, "Sealing %s", theme_name);
    actionMessage(cs, "SealEnvelope", buffer);
    record_action(cs, "SealEnvelope", theme_name, hand_name, NULL);

    if (hand == NULL) {
        errorEffectorNull(cs, "SealEnvelope");
    }
    else if (theme == NULL) {
        errorEffectorState(cs, "SealEnvelope: hand (%s) is empty", hand->name);
    }
    else if (isEnvelope(theme) < 0.9) {
        errorObjectIncorrect(cs, "%s: hand (%s) holds incorrect object %s", "SealEnvelope", hand->name, theme->name);
    }
    else {
        traceObjectPreAction(cs, theme);
        theme->state = (theme->state | STATE_CLOSED) & ~STATE_OPEN & ~STATE_MOIST;
        traceObjectPostAction(cs, theme);
    }
}

/*----------------------------------------------------------------------------*/

void moisten_by_licking(ContentionScheduler *cs, Effector *hand)
{
    Object *theme = effectorGetContents(hand);

    char *hand_name = (hand == NULL ? "nothing" : hand->name);
    char *theme_name = (theme == NULL ? "nothing" : theme->name);

    g_snprintf(buffer, 128, "MoistenByLicking %s", theme_name);
    actionMessage(cs, "MoistenByLicking", buffer);
    record_action(cs, "MoistenByLicking", theme_name, hand_name, NULL);

    if (hand == NULL) {
        errorEffectorNull(cs, "MoistenByLicking");
    }
    else if (theme == NULL) {
        errorEffectorState(cs, "MoistenByLicking: hand (%s) is empty", hand->name);
    }
    else if ((isStamp(theme) < 0.9) && (isEnvelope(theme) < 0.9)) {
        errorObjectIncorrect(cs, "%s: hand (%s) holds incorrect object %s", "MoistenByLicking", hand->name, theme->name);
    }
    else {
        traceObjectPreAction(cs, theme);
        theme->state = theme->state | STATE_MOIST;
        traceObjectPostAction(cs, theme);
    }
}

/*----------------------------------------------------------------------------*/

void tuck_flap(ContentionScheduler *cs, Effector *hand)
{
    Object *theme = effectorGetContents(hand);

    char *hand_name = (hand == NULL ? "nothing" : hand->name);
    char *theme_name = (theme == NULL ? "nothing" : theme->name);

    g_snprintf(buffer, 128, "Sealing %s by tucking", theme_name);
    actionMessage(cs, "TuckEnvelopeFlap", buffer);
    record_action(cs, "TuckEnvelopeFlap", theme_name, hand_name, NULL);

    if (hand == NULL) {
        errorEffectorNull(cs, "TuckEnvelopeFlap");
    }
    else if (theme == NULL) {
        errorEffectorState(cs, "TuckEnvelopeFlap: hand (%s) is empty", hand->name);
    }
    else if (isEnvelope(theme) < 0.9) {
        errorObjectIncorrect(cs, "%s: hand (%s) holds incorrect object %s", "TuckEnvelopeFlap", hand->name, theme->name);
    }
    else {
        traceObjectPreAction(cs, theme);
        theme->state = (theme->state | STATE_CLOSED) & ~STATE_OPEN & ~STATE_MOIST;
        traceObjectPostAction(cs, theme);
    }
}

/******************************************************************************/
