/*******************************************************************************

    FILE: letter/analyse.c
    AUTHOR: R. Cooper
    CREATION DATE: Thu Aug 12 16:20:38 2004
    CONTENTS: Analysis routine for the letter task

*******************************************************************************/

#include "../cs.h"
#include "../analyse.h"

/******************************************************************************/

typedef enum letter_object {
    OBJECT_NONE, OBJECT_PAPER, OBJECT_PEN_NIB, OBJECT_PEN_STALK,
    OBJECT_STAMP_FRONT, OBJECT_ENVELOPE_FRONT,
    OBJECT_STAMP_BACK, OBJECT_ENVELOPE_BACK,
    OBJECT_UNKNOWN
} LetterObject;

typedef enum letter_location {
    LOC_NONE, LOC_HELD, LOC_ON_TABLE, LOC_IN_ENVELOPE, LOC_ON_ENVELOPE
} LetterLocation;

typedef struct letter_analysis_object_list {
    LetterObject                       id;        /* Enumerated id */
    LetterLocation                     location;
    int                                state;
    struct letter_analysis_object_list *next;      /* Next object in the list  */
} LetterAnalysisObjectList;

typedef struct letter_state {
    // Information on all possible objects in the world
    LetterAnalysisObjectList   *items; 

    // Fixated item:
    LetterObject                fixated;

    // Other relevant state information:
    Boolean                     envelope_stamped;
    Boolean                     envelope_addressed;
    Boolean                     envelope_stuffed;
    Boolean                     envelope_sealed;

} LetterState;

#define LETTER_STATE_LICKED         1
#define LETTER_STATE_SEALED         2
#define LETTER_STATE_FOLDED         4
#define LETTER_STATE_ADDRESSED      8

/******************************************************************************/

static LetterObject object_id_from_string(char *arg)
{
    if (arg == NULL) {
        return(OBJECT_NONE);
    }
    else if (strcmp(arg, "nothing") == 0) {
        return(OBJECT_NONE);
    }
    else if (strcmp(arg, "notePaper") == 0) {
        return(OBJECT_PAPER);
    }
    else if (strcmp(arg, "pen_nib") == 0) {
        return(OBJECT_PEN_NIB);
    }
    else if (strcmp(arg, "pen_stalk") == 0) {
        return(OBJECT_PEN_STALK);
    }
    else if (strcmp(arg, "envelope_front") == 0) {
        return(OBJECT_ENVELOPE_FRONT);
    }
    else if (strcmp(arg, "envelope_back") == 0) {
        return(OBJECT_ENVELOPE_BACK);
    }
    else if (strcmp(arg, "stamp_front") == 0) {
        return(OBJECT_STAMP_FRONT);
    }
    else if (strcmp(arg, "stamp_back") == 0) {
        return(OBJECT_STAMP_BACK);
    }

    /* Final catch-all case: */

    else {
        report_unrecognised_object(stdout, arg);
        return(OBJECT_UNKNOWN);
    }
}

static LetterAnalysisObjectList *object_ptr_from_id(LetterState *state, LetterObject id)
{
    LetterAnalysisObjectList *tmp = state->items;

    while ((tmp != NULL) && (tmp->id != id)) {
        tmp = tmp->next;
    }
    if ((tmp == NULL) || (tmp->id != id)) {
        return(NULL);
    }
    else {
        return(tmp);
    }
}

static LetterAnalysisObjectList *object_ptr_from_string(LetterState *state, char *arg)
{
    return(object_ptr_from_id(state, object_id_from_string(arg)));
}

static Boolean action_compare(ActionList *this, char *action, LetterObject arg1)
{
    if (this == NULL) {
        return(FALSE);
    }
    else if (strcmp(this->name, action) != 0) {
        return(FALSE);
    }
    else if (object_id_from_string(this->argument1) != arg1) {
        return(FALSE);
    }
    else {
        return(TRUE);
    }
}

/******************************************************************************/

static void state_add_item(LetterState *state, LetterObject id, LetterLocation loc)
{
    LetterAnalysisObjectList *new = (LetterAnalysisObjectList *)malloc(sizeof(LetterAnalysisObjectList));

    if (new != NULL) {
        new->id = id;
        new->location = loc;
        new->state = 0;
        new->next = state->items;
        state->items = new;
    }
}

static void state_initialise(LetterState *state)
{
    state->items = NULL;

    /* All objects must be initialised to match letter/object.c */

    state_add_item(state, OBJECT_PEN_NIB, LOC_ON_TABLE);
    state_add_item(state, OBJECT_PEN_STALK, LOC_ON_TABLE);
    state_add_item(state, OBJECT_PAPER, LOC_ON_TABLE);
    state_add_item(state, OBJECT_STAMP_FRONT, LOC_ON_TABLE);
    state_add_item(state, OBJECT_STAMP_BACK, LOC_NONE);
    state_add_item(state, OBJECT_ENVELOPE_FRONT, LOC_ON_TABLE);
    state_add_item(state, OBJECT_ENVELOPE_BACK, LOC_NONE);

    state->fixated = OBJECT_NONE;

    state->envelope_addressed = FALSE;
    state->envelope_stamped = FALSE;
    state->envelope_stuffed = FALSE;
    state->envelope_sealed = FALSE;
}

static void state_destroy(LetterState *state)
{
    while (state->items != NULL) {
        LetterAnalysisObjectList *tmp = state->items;
        state->items = tmp->next;
        free(tmp);
    }
}

/******************************************************************************/

static void check_consecutive_toying_actions(ActionList *this, LetterObject arg1)
{
    if (action_compare(this->next, "PickUp", arg1)) {
        if (action_compare(this->next->next, "PutDown", arg1)) {
            this->next->checked = TRUE;
            this->next->next->checked = TRUE;
            check_consecutive_toying_actions(this->next->next, arg1);
        }
        else if (action_compare(this->next->next, "Discard", arg1)) {
            this->next->checked = TRUE;
            this->next->next->checked = TRUE;
            check_consecutive_toying_actions(this->next->next, arg1);
        }
    }
}

static Boolean pen_is_held(LetterState *state)
{
    LetterAnalysisObjectList *ol;

    for (ol = state->items; ol != NULL; ol = ol->next) {
        if ((ol->id == OBJECT_PEN_NIB) && (ol->location == LOC_HELD)) {
            return(TRUE);
        }
        else if ((ol->id == OBJECT_PEN_STALK) && (ol->location == LOC_HELD)) {
            return(TRUE);
        }
    }
    return(FALSE);
}

static Boolean pen_nib_is_held(LetterState *state)
{
    LetterAnalysisObjectList *ol;

    for (ol = state->items; ol != NULL; ol = ol->next) {
        if ((ol->id == OBJECT_PEN_NIB) && (ol->location == LOC_HELD)) {
            return(TRUE);
        }
    }
    return(FALSE);
}

static Boolean stamp_front_licked(LetterState *state)
{
    LetterAnalysisObjectList *item = object_ptr_from_id(state, OBJECT_STAMP_FRONT);

    return((item != NULL) && (item->state & LETTER_STATE_LICKED));
}

/******************************************************************************/

static void score_action_fixate(LetterState *state, DataAnalysis *la, ActionList *action, char *arg1)
{
    LetterAnalysisObjectList *item1;

    if (object_id_from_string(arg1) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "Fixate", arg1, NULL);
        ignore_consecutive_actions(action);
    }
    else if ((item1 = object_ptr_from_string(state, arg1)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else {
        state->fixated = item1->id;
    }
}

/*----------------------------------------------------------------------------*/

static void score_action_pickup(LetterState *state, DataAnalysis *la, ActionList *action, char *arg1)
{
    LetterAnalysisObjectList *item1;

    if (object_id_from_string(arg1) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "PickUp", arg1, NULL);
        ignore_consecutive_actions(action);
    }
    else if ((item1 = object_ptr_from_string(state, arg1)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if (item1->location == LOC_HELD) {
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "PickUp", arg1, NULL);
        ignore_consecutive_actions(action);
    }
    else {
        item1->location = LOC_HELD;
        if (action_compare(action->next, "PutDown", item1->id)) {
            // Score TOYING error and ignore any consecutive toying
            analyseErrorLogAppend(la, ERROR_TOYING, "PickUp/PutDown", arg1, NULL);
            check_consecutive_toying_actions(action->next, item1->id);
        }
        else if (action_compare(action->next, "Discard", item1->id)) {
            // Score TOYING error and ignore any consecutive toying
            analyseErrorLogAppend(la, ERROR_TOYING, "PickUp/Discard", arg1, NULL);
            check_consecutive_toying_actions(action->next, item1->id);
        }
    }
}

/*----------------------------------------------------------------------------*/

static void score_action_putdown(LetterState *state, DataAnalysis *la, ActionList *action, char *arg1)
{
    LetterAnalysisObjectList *item1;

    if (object_id_from_string(arg1) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "PutDown", arg1, NULL);
        ignore_consecutive_actions(action);
    }
    else if ((item1 = object_ptr_from_string(state, arg1)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if (item1->location != LOC_HELD) {
        // Score PERPLEXITY: Object is not held
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "PutDown", arg1, NULL);
        ignore_consecutive_actions(action);
    }
    else {
        item1->location = LOC_ON_TABLE;
    }
}

/*----------------------------------------------------------------------------*/

static void score_action_discard(LetterState *state, DataAnalysis *la, ActionList *action, char *arg1)
{
    LetterAnalysisObjectList *item1;

    if (object_id_from_string(arg1) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "Discard", arg1, NULL);
        ignore_consecutive_actions(action);
    }
    else if ((item1 = object_ptr_from_string(state, arg1)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if (item1->location != LOC_HELD) {
        // Score PERPLEXITY: Object is not held
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "Discard", arg1, NULL);
        ignore_consecutive_actions(action);
    }
    else {
        item1->location = LOC_ON_TABLE;
    }
}

/*----------------------------------------------------------------------------*/

static void score_action_apply_stamp(LetterState *state, DataAnalysis *la, ActionList *action, char *arg1, char *arg2)
{
    LetterAnalysisObjectList *item1; // The back of the stamp
    LetterAnalysisObjectList *item2; // The front of the envelope

    if (object_id_from_string(arg1) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PANTOMIME, "ApplyStamp", arg1, arg2);
    }
    else if ((item1 = object_ptr_from_string(state, arg1)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if (object_id_from_string(arg2) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "ApplyStamp", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if ((item2 = object_ptr_from_string(state, arg2)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if (item1 == item2) {
        // Score PERPLEXITY: cannot apply stamp to itself
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "ApplyStamp", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if ((item1->id == OBJECT_STAMP_FRONT) && (item2->id == OBJECT_STAMP_BACK)) {
        // Score PERPLEXITY: cannot apply stamp to self!
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "ApplyStamp", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if ((item1->id == OBJECT_STAMP_BACK) && (item2->id == OBJECT_STAMP_FRONT)) {
        // Score PERPLEXITY: cannot apply stamp to self!
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "ApplyStamp", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if ((item1->id == OBJECT_ENVELOPE_FRONT) && (item2->id == OBJECT_ENVELOPE_BACK)) {
        // Score PERPLEXITY: cannot apply envelope to self!
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "ApplyStamp", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if ((item1->id == OBJECT_ENVELOPE_BACK) && (item2->id == OBJECT_ENVELOPE_FRONT)) {
        // Score PERPLEXITY: cannot apply envelope to self!
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "ApplyStamp", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if ((item1->id != OBJECT_STAMP_FRONT) && (item1->id != OBJECT_STAMP_BACK)) {
        // Score MISUSE_1 error:  This isn't a stamp!
        analyseErrorLogAppend(la, ERROR_MISUSE_1, "ApplyStamp", arg1, arg2);
    }
    else if (item2->id == OBJECT_ENVELOPE_BACK) {
        // Score MISLOCATION_2 error: stamping on wrong side of envelope
        analyseErrorLogAppend(la, ERROR_MISLOCATION_2, "ApplyStamp", arg1, arg2);
        if (item1->id == OBJECT_STAMP_BACK) {
            if (item1->state & LETTER_STATE_LICKED) {
                item1->location = LOC_ON_ENVELOPE;
                state->envelope_stamped = TRUE;
            }
            else if (!stamp_front_licked(state)) {
                // Score STEP_OMISSION error: Forgot to lick the stamp!
                analyseErrorLogAppend(la, ERROR_STEP_OMISSION, "ApplyStampWithoutLicking", arg1, arg2);
            }
	}
    }
    else if (item2->id != OBJECT_ENVELOPE_FRONT) {
        // Score MISLOCATION_1 error: stamping on wrong thing
        analyseErrorLogAppend(la, ERROR_MISLOCATION_1, "ApplyStamp", arg1, arg2);
    }
    else if ((item1->state & LETTER_STATE_LICKED) == 0) {
        // Score STEP_OMISSION error: Forgot to lick the stamp!
        analyseErrorLogAppend(la, ERROR_STEP_OMISSION, "ApplyStampWithoutLicking", arg1, arg2);
    }
    else if (item1->id == OBJECT_STAMP_FRONT) {
        // Score MISLOCATION_2 error: stamping on wrong side of stamp
        analyseErrorLogAppend(la, ERROR_MISLOCATION_2, "ApplyStamp", arg1, arg2);
    }
    else {
        // Ok: Envelope stamped
        item1->location = LOC_ON_ENVELOPE;
        state->envelope_stamped = TRUE;
    }
}

/*----------------------------------------------------------------------------*/

static void score_action_fold_letter(LetterState *state, DataAnalysis *la, ActionList *action, char *arg1)
{
    LetterAnalysisObjectList *item1; // The note paper

    if (object_id_from_string(arg1) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PANTOMIME, "FoldLetter", arg1, NULL);
    }
    else if ((item1 = object_ptr_from_string(state, arg1)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if ((item1->state & LETTER_STATE_FOLDED) != 0) {
        // Score PERSEVERATION error: object is already folded
        analyseErrorLogAppend(la, ERROR_PERSEVERATION, "FoldLetter", arg1, NULL);
    }
    else if ((item1->id == OBJECT_PEN_NIB) || (item1->id == OBJECT_PEN_STALK)) {
        // Score PERPLEXITY: cannot fold pen!
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "FoldLetter", arg1, NULL);
        ignore_consecutive_actions(action);
    }
    else if (item1->id != OBJECT_PAPER) {
        // Score ACTION ADDITION error: folding non-paper
        analyseErrorLogAppend(la, ERROR_ACTION_ADDITION, "FoldLetter", arg1, NULL);
        item1->state = item1->state | LETTER_STATE_FOLDED;
    }
    else {
        // Ok: Folding the note paper
        item1->state = item1->state | LETTER_STATE_FOLDED;
    }
}

/*----------------------------------------------------------------------------*/

static void score_action_write_address(LetterState *state, DataAnalysis *la, ActionList *action, char *arg1, char *arg2)
{
    LetterAnalysisObjectList *item1; // The front of the envelope

    if (object_id_from_string(arg1) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "WriteAddress", arg1, NULL);
        ignore_consecutive_actions(action);
    }
    else if ((item1 = object_ptr_from_string(state, arg1)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if (!pen_is_held(state)) {
        // Score PANTOMIME: pen is not held!
        analyseErrorLogAppend(la, ERROR_PANTOMIME, "WriteAddress", arg1, NULL);
    }
    else if ((item1->state & LETTER_STATE_ADDRESSED) != 0) {
        // Score PERSEVERATION error: object is already addressed
        analyseErrorLogAppend(la, ERROR_PERSEVERATION, "WriteAddress", arg1, NULL);
    }
    else if (item1->id == OBJECT_ENVELOPE_BACK) {
        // Score MISLOCATION_2 error: addressing on wrong side of envelope
        analyseErrorLogAppend(la, ERROR_MISLOCATION_2, "WriteAddress", arg1, NULL);
        item1->state = item1->state | LETTER_STATE_ADDRESSED;
        state->envelope_addressed = TRUE;
    }
    else if (!pen_nib_is_held(state)) {
        // Score MISLOCATION_2 error: holding wrong end of pen
        analyseErrorLogAppend(la, ERROR_MISLOCATION_2, "WriteAddress", arg1, NULL);
        item1->state = item1->state | LETTER_STATE_ADDRESSED;
        state->envelope_addressed = TRUE;
    }
    else if (item1->id != OBJECT_ENVELOPE_FRONT) {
        // Score MISLOCATION_1 error: addressing on non-envelope
        analyseErrorLogAppend(la, ERROR_MISLOCATION_1, "WriteAddress", arg1, NULL);
        item1->state = item1->state | LETTER_STATE_ADDRESSED;
        state->envelope_addressed = TRUE;
    }
    else {
        // Ok: Addressing front of envelope
        item1->state = item1->state | LETTER_STATE_ADDRESSED;
        state->envelope_addressed = TRUE;
    }
}

/*----------------------------------------------------------------------------*/

static void score_action_pack_envelope(LetterState *state, DataAnalysis *la, ActionList *action, char *arg1, char *arg2)
{
    LetterAnalysisObjectList *item1; // The note paper
    LetterAnalysisObjectList *item2; // The envelope

    if (object_id_from_string(arg1) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "PackEnvelope", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if ((item1 = object_ptr_from_string(state, arg1)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if (object_id_from_string(arg2) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PANTOMIME, "PackEnvelope", arg1, arg2);
    }
    else if ((item2 = object_ptr_from_string(state, arg2)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if (item1 == item2) {
        // Score PERPLEXITY: cannot pack item into self!
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "PackEnvelope", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if ((item1->id == OBJECT_ENVELOPE_BACK) && (item2->id == OBJECT_ENVELOPE_FRONT)) {
        // Score PERPLEXITY: cannot pack item into self!
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "PackEnvelope", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if ((item1->id == OBJECT_ENVELOPE_FRONT) && (item2->id == OBJECT_ENVELOPE_BACK)) {
        // Score PERPLEXITY: cannot pack item into self!
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "PackEnvelope", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if ((item2->id != OBJECT_ENVELOPE_BACK) && (item2->id != OBJECT_ENVELOPE_FRONT)) {
        // Score PERPLEXITY: target isn't an envelope!
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "PackEnvelope", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if ((item2->state & LETTER_STATE_SEALED) != 0) {
        // Score PERPLEXITY error: envelope is already sealed
        analyseErrorLogAppend(la, ERROR_PERPLEXITY, "PackEnvelope", arg1, arg2);
        ignore_consecutive_actions(action);
    }
    else if (item1->id != OBJECT_PAPER) {
        // Score ACTION ADDITION error: packing non-letter into envelope
        analyseErrorLogAppend(la, ERROR_ACTION_ADDITION, "PackEnvelope", arg1, arg2);
        item1->location = LOC_IN_ENVELOPE;
    }
    else if ((item1->state & LETTER_STATE_FOLDED) == 0) {
        // Score ACTION_ANTICIPATION error: failed to fold letter
        analyseErrorLogAppend(la, ERROR_ACTION_ANTICIPATION, "PackEnvelopeWithoutFolding", arg1, arg2);
        item1->location = LOC_IN_ENVELOPE;
        state->envelope_stuffed = TRUE;
    }
    else {
        // Ok: Letter packed
        item1->location = LOC_IN_ENVELOPE;
        state->envelope_stuffed = TRUE;
    }
}

/*----------------------------------------------------------------------------*/

static void score_action_seal_envelope(LetterState *state, DataAnalysis *la, ActionList *action, char *arg1)
{
    LetterAnalysisObjectList *item1;

    if (object_id_from_string(arg1) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PANTOMIME, "SealEnvelope", arg1, NULL);
    }
    else if ((item1 = object_ptr_from_string(state, arg1)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if ((item1->state & LETTER_STATE_SEALED) != 0) {
        // Score PERSEVERATION error: object is already sealed
        analyseErrorLogAppend(la, ERROR_PERSEVERATION, "SealEnvelope", arg1, NULL);
    }
    else if (item1->id == OBJECT_ENVELOPE_FRONT) {
        // Score MISLOCATION_2 error: sealing on wrong side of envelope
        analyseErrorLogAppend(la, ERROR_MISLOCATION_2, "SealEnvelope", arg1, NULL);
        item1->state = item1->state | LETTER_STATE_SEALED;
        state->envelope_sealed = TRUE;
    }
    else if (item1->id != OBJECT_ENVELOPE_BACK) {
        // Score MISLOCATION_1 error: sealing non-envelope
        analyseErrorLogAppend(la, ERROR_MISLOCATION_1, "SealEnvelope", arg1, NULL);
        item1->state = item1->state | LETTER_STATE_SEALED;
    }
    else {
        // Ok: Sealing back of envelope
        item1->state = item1->state | LETTER_STATE_SEALED;
        state->envelope_sealed = TRUE;
    }
}

/*----------------------------------------------------------------------------*/

static void score_action_moisten_by_licking(LetterState *state, DataAnalysis *la, ActionList *action, char *arg1)
{
    LetterAnalysisObjectList *item1; // The back of the stamp or envelope 

    if (object_id_from_string(arg1) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PANTOMIME, "MoistenByLicking01", arg1, NULL);
    }
    else if ((item1 = object_ptr_from_string(state, arg1)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if ((item1->state & LETTER_STATE_LICKED) != 0) {
        // Score PERSEVERATION error: object is already licked
        analyseErrorLogAppend(la, ERROR_PERSEVERATION, "MoistenByLicking03", arg1, NULL);
    }
    else if ((item1->id == OBJECT_ENVELOPE_FRONT) || (item1->id == OBJECT_STAMP_FRONT)) {
        // Score MISLOCATION_2 error: licking wrong side
        item1->state = item1->state | LETTER_STATE_LICKED;
        analyseErrorLogAppend(la, ERROR_MISLOCATION_2, "MoistenByLicking04", arg1, NULL);
    }
    else if (item1->id == OBJECT_STAMP_BACK) {
        // Everything ok!
        item1->state = item1->state | LETTER_STATE_LICKED;
    }
    else if (item1->id == OBJECT_ENVELOPE_BACK) {
        // Probably ok ...
        item1->state = item1->state | LETTER_STATE_LICKED;
        // Possibly the envelope has been sealed. If so, we'll only reach
        // here if it wasn't licked before it was sealed, in which case we
        // would already have recorded an anticipation error. So nothing
        // special is required here.
    }
    else {
        // Score OBJECT_MISLOCATION_1 error:  Licking the note paper or pen!
        analyseErrorLogAppend(la, ERROR_MISLOCATION_1, "MoistenByLicking05", arg1, NULL);
    }
}

/*----------------------------------------------------------------------------*/

static void score_action_tuck_envelope_flap(LetterState *state, DataAnalysis *la, ActionList *action, char *arg1)
{
    LetterAnalysisObjectList *item1;

    if (object_id_from_string(arg1) == OBJECT_NONE) {
        analyseErrorLogAppend(la, ERROR_PANTOMIME, "TuckEnvelopeFlap", arg1, NULL);
    }
    else if ((item1 = object_ptr_from_string(state, arg1)) == NULL) {
        report_unrecognised_action(stdout, action);
    }
    else if ((item1->state & LETTER_STATE_SEALED) != 0) {
        // Score PERSEVERATION error: object is already sealed
        analyseErrorLogAppend(la, ERROR_PERSEVERATION, "TuckEnvelopeFlap", arg1, NULL);
    }
    else if (item1->id == OBJECT_ENVELOPE_FRONT) {
        // Score MISLOCATION_2 error: sealing on wrong side of envelope
        analyseErrorLogAppend(la, ERROR_MISLOCATION_2, "TuckEnvelopeFlap", arg1, NULL);
        item1->state = item1->state | LETTER_STATE_SEALED;
        state->envelope_sealed = TRUE;
    }
    else if (item1->id != OBJECT_ENVELOPE_BACK) {
        // Score MISLOCATION_1 error: sealing non-envelope
        analyseErrorLogAppend(la, ERROR_MISLOCATION_1, "TuckEnvelopeFlap", arg1, NULL);
        item1->state = item1->state | LETTER_STATE_SEALED;
    }
    else {
        // Ok: Sealing back of envelope, but we should be sealing by licking!
        // Score MISUSE_2 error: sealing non-envelope
        analyseErrorLogAppend(la, ERROR_MISUSE_2, "TuckEnvelopeFlap", arg1, NULL);

        item1->state = item1->state | LETTER_STATE_SEALED;
        state->envelope_sealed = TRUE;
    }
}

/******************************************************************************/

static void score_action(LetterState *state, DataAnalysis *la, ActionList *action)
{
    if (action == NULL) {
        fprintf(stderr, "INTERNAL ERROR: NULL action in score_action/3 (file is letter/analyse.c)\n");
    }
    else if (strcmp(action->name, "Fixate") == 0) {
        score_action_fixate(state, la, action, action->argument1);
    }
    else if (strcmp(action->name, "PickUp") == 0) {
        score_action_pickup(state, la, action, action->argument1);
    }
    else if (strcmp(action->name, "PutDown") == 0) {
        score_action_putdown(state, la, action, action->argument1);
    }
    else if (strcmp(action->name, "Discard") == 0) {
        score_action_discard(state, la, action, action->argument1);
    }
    else if (strcmp(action->name, "ApplyStamp") == 0) {
        score_action_apply_stamp(state, la, action, action->argument1, action->argument2);
    }
    else if (strcmp(action->name, "FoldLetter") == 0) {
        score_action_fold_letter(state, la, action, action->argument1);
    }
    else if (strcmp(action->name, "WriteAddress") == 0) {
        score_action_write_address(state, la, action, action->argument1, action->argument2);
    }
    else if (strcmp(action->name, "PackEnvelope") == 0) {
        score_action_pack_envelope(state, la, action, action->argument1, action->argument2);
    }
    else if (strcmp(action->name, "SealEnvelope") == 0) {
        score_action_seal_envelope(state, la, action, action->argument1);
    }
    else if (strcmp(action->name, "MoistenByLicking") == 0) {
        score_action_moisten_by_licking(state, la, action, action->argument1);
    }
    else if (strcmp(action->name, "TuckEnvelopeFlap") == 0) {
        score_action_tuck_envelope_flap(state, la, action, action->argument1);
    }
    else {
        report_unrecognised_action(stdout, action);
    }
}

/******************************************************************************/

static void score_omissions(LetterState *state, DataAnalysis *la)
{
    if (!state->envelope_stamped) {
        analyseErrorLogAppend(la, ERROR_STEP_OMISSION, "StampEnvelope", NULL, NULL);
    }
    if (!state->envelope_addressed) {
        analyseErrorLogAppend(la, ERROR_STEP_OMISSION, "AddressEnvelope", NULL, NULL);
    }
    if (!state->envelope_stuffed) {
        analyseErrorLogAppend(la, ERROR_STEP_OMISSION, "StuffEnvelope", NULL, NULL);
    }
    if (!state->envelope_sealed) {
        analyseErrorLogAppend(la, ERROR_STEP_OMISSION, "SealEnvelope", NULL, NULL);
    }
}

/******************************************************************************/

void analyse_letter_task(DataAnalysis *la)
{
    ActionList *as;
    LetterState state;

    state_initialise(&state);

    for (as = la->actionSequence; as; as = as->next) {
        as->checked = FALSE;
    }

    for (as = la->actionSequence; as; as = as->next) {
        if (!as->checked) {
            score_action(&state, la, as);
            as->checked = TRUE;
        }
    }

    score_omissions(&state, la);

    state_destroy(&state);
}

/******************************************************************************/
