/*******************************************************************************

    FILE: 
    AUTHOR: R. Cooper (translated from C++ written by SVW)
    CREATION DATE: 
    CONTENTS: 

*******************************************************************************/

#include "cs.h"
#include <math.h>

/* Lateral inhibition is relative to lateralBaseline, which might be either   */
/* restAct or minAct (RPC: Thu Jun 24 10:44:23 1999).                         */

#define lateralBaseline minAct

extern Schema *selected_parent(ContentionScheduler *cs, Schema *this);
extern Boolean goalPostConditionAchieved(ContentionScheduler *cs, Goal *goal);
extern Boolean goalPreConditionAchieved(ContentionScheduler *cs, Goal *goal);

/******************************************************************************/
/* Random number generator (from SICStus 2.1 distribution) ********************/

static short RAND1 = 27134;
static short RAND2 = 9213;
static short RAND3 = 17773;

double cs_random()
{
    int X, Y, Z;
    double T;

    X = (RAND1*171) % 30269;
    Y = (RAND2*172) % 30307;
    Z = (RAND3*170) % 30323;
    T = X/30269.0 + Y/30307.0 + Z/30323.0;

    RAND1 = X;
    RAND2 = Y;
    RAND3 = Z;

    return(T - (int) T);
}

/*----------------------------------------------------------------------------*/

char *random_seed_get()
{
    static char buffer[30];

    g_snprintf(buffer, 30, "%d:%d:%d", RAND1, RAND2, RAND3);
    return(buffer);
}

/*----------------------------------------------------------------------------*/

void random_seed_save(char *filename, ContentionScheduler *cs)
{
    FILE *fp = fopen(filename, "w");

    if (fp != NULL) {
        fprintf(fp, "%d %d %d \n",  RAND1, RAND2, RAND3);
        fclose(fp);
    }
    else {
        char buffer[128];

        g_snprintf(buffer, 128, "Cannot write random seed to \"%s\"", filename);
        message(WARNING, cs, buffer);
    }
}

/*----------------------------------------------------------------------------*/

void random_seed_load(char *filename, ContentionScheduler *cs)
{
    FILE *fp = fopen(filename, "r");

    if (fp != NULL) {
        fscanf(fp, "%hd%hd%hd", &RAND1, &RAND2, &RAND3);
        fclose(fp);
    }
    else {
        char buffer[128];

        g_snprintf(buffer, 128, "Cannot read random seed from \"%s\"", filename);
        message(WARNING, cs, buffer);
    }
}

/******************************************************************************/
/******************************************************************************/

double squash(double x)
{
   double y = exp(x);
   return((y - 1.0) / (y + 1.0));
}

// static double sigmoid(double x)
// {
//    double y = exp(x);
//    return(y / (y + 1.0));
// }

/*............................................................................*/

static double bounding_function(MathsType mathsSystem, double x)
{
    double r = (restAct - ((maxAct+minAct) / 2.0)) * (2.0 / (maxAct - minAct));
    double fx, squash, a;

    a = log ((1.0 + r) / (1.0 - r));

    switch (mathsSystem)  {
	case SIGMOID:
	    fx = x + a;
	    break;

	case TANH:
	    fx = (2.0 * x) + a;
	    break;

	case HALFX:
	    fx = (x / 2.0) + a;
	    break;

	default:
	    fprintf(stderr, "Invalid MathsType in bounding_function\n");
	    return 0.0;
	    break;
    }

    squash = (exp(fx) - 1.0) / (exp(fx) + 1.0);

    return((squash * (maxAct - minAct) / 2.0) + ((maxAct + minAct) / 2.0));
}

/*............................................................................*/

static double inverse_bounding_function(MathsType mathsSystem, double x)
{
    double r = (restAct - ((maxAct + minAct) / 2.0)) * (2.0 / (maxAct - minAct));
    double t = (x -       ((maxAct + minAct) / 2.0)) * (2.0 / (maxAct - minAct));
    double a = log ((1.0 + r) / (1.0 - r));
    double b = log ((1.0 + t) / (1.0 - t));
    double c = b - a;
    double fx = 0.0;

    switch (mathsSystem) {
	case SIGMOID:
	    fx = c;
	    break;

	case TANH:
	    fx = c / 2.0;
	    break;

	case HALFX:
	    fx = c * 2.0;
	    break;

	default:
	    fprintf(stderr, "Invalid MathsType in inverse_bounding_function\n");
	    break;
    }

    return(fx);
}

/*............................................................................*/

static double epsilon_function(double previous, double persistence, double input, double x)
{
    double epsilon = 0.0;

    if ((input > 0.0) && (previous < maxAct)) {
	epsilon = x * (maxAct - previous);
    }
    else if ((input < 0.0) && (previous > minAct)) {
	epsilon = x * (previous - minAct);
    }

    return((persistence * previous) + ((1 - persistence) * restAct) + epsilon);
}

/*............................................................................*/

static double apply_equations(double previous, double persistence, double input, MathsType mathsSystem)
{
    double x;
    double fx = 0.0;

    switch (mathsSystem) {

    case SIGMOID:
	x = inverse_bounding_function(SIGMOID,previous)*persistence+input;
	fx = bounding_function(SIGMOID, x);
	break;

    case TANH:
	x = inverse_bounding_function(TANH,previous)*persistence+input;
	fx = bounding_function(TANH, x);
	break;

    case HALFX:
	x = inverse_bounding_function(HALFX,previous)*persistence+input;
	fx = bounding_function(HALFX, x);
	break;

    case M_R:
	x = input;
	fx = epsilon_function(previous, persistence, input, x);
	break;

    case HOUGHTON:
	x = squash(input);
	fx = epsilon_function(previous, persistence, input, x);
	break;

    case NORMALISED:
	fx = (previous * persistence) + input;
	break;

    default:
	fprintf(stderr, "Invalid MathsType (%d) in apply_equations\n", (int) mathsSystem);
	break;
    }

    return(fx);
}

/*............................................................................*/

double normal_noise(double mean, double sd)
{
    return(mean + sd * sqrt(-2 * log(cs_random())) * cos(2 * M_PI * cs_random()));
}

/******************************************************************************/
/******************************************************************************/

double initial_activation(ComponentType role)
{
    double input = normal_noise(0.0, noiseSD[role]);
    return(apply_equations(restAct, persistence, input, maths));
}

/******************************************************************************/
/******************************************************************************/

void normalise_schema_activations(ContentionScheduler *cs)
{
    SchemaList *tmp;
    double sum = 0;

    for (tmp = cs->schemas; tmp != NULL; tmp = tmp->tail) {
        sum += (tmp->head->activation[cs->cycle] - minAct) / (maxAct - minAct);
    }

    for (tmp = cs->schemas; tmp != NULL; tmp = tmp->tail) {
        tmp->head->activation[cs->cycle] = ((tmp->head->activation[cs->cycle] - minAct) / sum) - minAct;
    }
}

/*............................................................................*/

void normalise_object_activations(ContentionScheduler *cs)
{
    ObjectList *tmp;
    ObjectRole role; 

    for (role = 0; role < (MaxRoles-1) ; role++) {
        double sum = 0;

        for (tmp = cs->objects; tmp != NULL; tmp = tmp->tail) {
            sum += (tmp->head->activation[role][cs->cycle] - minAct) / (maxAct - minAct);
        }

        for (tmp = cs->objects; tmp != NULL; tmp = tmp->tail) {
            tmp->head->activation[role][cs->cycle] = ((tmp->head->activation[role][cs->cycle] - minAct) / sum) - minAct;
        }
    }
}

/*............................................................................*/

void normalise_effector_activations(ContentionScheduler *cs)
{
    EffectorList *tmp;
    double sum = 0;

    for (tmp = cs->effectors; tmp != NULL; tmp = tmp->tail) {
        sum += (tmp->head->activation[cs->cycle] - minAct) / (maxAct - minAct);
    }

    for (tmp = cs->effectors; tmp != NULL; tmp = tmp->tail) {
        tmp->head->activation[cs->cycle] = ((tmp->head->activation[cs->cycle] - minAct) / sum) - minAct;
    }
}

/******************************************************************************/

extern double situationTriggerFromObjectTrigger(ContentionScheduler *cs, Schema *this);

static double schema_environmental_influence(ContentionScheduler *cs, Schema *this)
{
//    fprintf(stdout, "%d: Unsquash env input to %s: %f\n", cs->cycle, this->name, situationTriggerFromObjectTrigger(cs, this));

    return(squash(situationTriggerFromObjectTrigger(cs, this)));
}

/*............................................................................*/

static int length(GoalList *subgoals)
{
    int l = 0;
    while (subgoals != NULL) {
        l++; subgoals = subgoals->tail;
    }
    return(l);
}

static double schema_parental_influence(ContentionScheduler *cs, Schema *this)
{
    Goal     *subgoal = NULL;
    GoalList *gs;
    double    ip;
    double    sf; /* A scale factor... */
    Schema *parent;

    if (this == cs->intended_schema) {
	return(1.0);
    }
    else if ((parent = selected_parent(cs, this)) != NULL) {

/* Activate any schema whose parent is selected:                              */

        ip = ((parent->activation[cs->cycle-1]  - minAct) / (maxAct - minAct)) / length(parent->subgoals);

/* Now find the subgoal of the selected parent corresponding to this schema.  */
/* Note: There may be more than one, so loop over all matching subgoals:      */

        sf = 0.0;

        for (gs = parent->subgoals; gs != NULL; gs = gs->tail) {
	    if (strcmp(gs->head->name, this->goal) == 0) {
		subgoal = gs->head;

/* So subgoal points to a goal structure that is achieved by this schema:     */

/* sf inhibits those corresponding to unnecessary or inappropriate subgoals: */

                if (subgoal->explicitly_achieved) {
                    sf = MAX(sf, 0.0);
                }
                else if (!goalPreConditionAchieved(cs, subgoal)) {
                    sf = MAX(sf, 0.0);
                }
                else if (goalPostConditionAchieved(cs, subgoal)) {
                    sf = MAX(sf, 0.0);
                }
//              else {
//                  sf = MAX(sf, sigmoid(goalPrecision * (MIN(subgoal->precondition(cs), 1.0 - subgoal->postcondition(cs)) - goalCutoff)));
//              }
                else {
                    sf = MAX(sf, 1.0);
		}
	    }
	}

        return(ip * sf);
    }
    else {
        return(0.0);
    }
}

/*............................................................................*/

static Boolean schema_self_influence_is_excitatory(ContentionScheduler *cs, Schema *this)
{
    if ((this->selected) && schema_goal_achieved(cs, this)) {
        return(FALSE);
    }
    else {
        return(TRUE);
    }
}

/*............................................................................*/

static double schema_self_influence(ContentionScheduler *cs, Schema *this)
{
    double d = (this->activation[cs->cycle-1] - minAct) / (maxAct - minAct);

    /* This used to return -d for the inhibitory case. Returning -1 makes */
    /* self inhibition stronger, especially for weakly active schemas.    */

    return(schema_self_influence_is_excitatory(cs, this) ? d : -1.0);
}

/*............................................................................*/

static double schema_lateral_influence(ContentionScheduler *cs, Schema *this)
{
    SchemaList *ss;
    double max_comp = 0.0;
    int n = 0;

    for (ss = this->competitors; ss != NULL; ss = ss->tail) {
        if (ss->head->activation[cs->cycle-1] > max_comp) {
            max_comp = ss->head->activation[cs->cycle-1];
        }
        n++;
    }

    if (n > 0) {
        return((lateralBaseline - max_comp) / (maxAct - minAct));
    }
    else {
        return(lateralBaseline - restAct);
    }
}

/*............................................................................*/

void schema_calculate_activation(ContentionScheduler *cs, Schema *this)
{
    double p	= schema_parental_influence(cs, this);
    double e	= schema_environmental_influence(cs, this);
    double s	= schema_self_influence(cs, this);
    double l	= schema_lateral_influence(cs, this);
    double net_input	= parental[SCHEMA] * p + environmental[SCHEMA] * e + self[SCHEMA] * s + lateral[SCHEMA] * l;
    double gross_input	= normal_noise(net_input, noiseSD[SCHEMA]);

    this->activation[cs->cycle] = apply_equations(this->activation[cs->cycle-1], persistence, gross_input, maths);
}

/******************************************************************************/
/******************************************************************************/

double object_self_influence_factor(ContentionScheduler *cs, Object *this, ObjectRole role)
{
//    return( this->inhibit[role] ? -1.0 : 1.0 );


    if (this->inhibit[role]) {
        return(-1.0);
    }
    else if (object_can_serve_role(this, role)) {
        return(1.0);
    }
    else {
        return(0.0);
    }
}

/*............................................................................*/

double object_self_influence(ContentionScheduler *cs, Object *this, ObjectRole role)
{
    double d = (this->activation[role][cs->cycle-1] - minAct) / (maxAct - minAct);
    return(object_self_influence_factor(cs, this, role) * d);
}

/*............................................................................*/

double object_environmental_influence(Object *this, ObjectRole r)
{
    return(squash(0.0));
}

/*............................................................................*/

static double schema_triggers_object(ContentionScheduler *cs, Schema *schema, Object *object, ObjectRole role)
{
    ObjectTriggerFunctionList *tf;
    double result = 0.0;

    for (tf = schema->objectTriggers; tf != NULL; tf = tf->next) {
        if (tf->role == role) {
            result += tf->trigger(cs, object);
        }
    }
    return(result);
}

double object_intentional_influence(ContentionScheduler *cs, Object *this, ObjectRole role)
{
    double total = 0.0;
    double contribution;
    double credibility;

    SchemaList *ss;

    if (object_can_serve_role(this, role)) {
	for (ss = cs->schemas; ss != NULL; ss = ss->tail) {
 	    if ((credibility = schema_triggers_object(cs, ss->head, this, role)) > 0) {
		contribution = (ss->head->activation[cs->cycle-1] - restAct) / (maxAct - minAct);
		total += (contribution * credibility);
	    }
	}
    }
    return(squash(total));
}

/*............................................................................*/

double object_lateral_influence(ContentionScheduler *cs, Object *this, ObjectRole role)
{
    ObjectList * os;
    double max_comp = 0.0;
    double lat;
    int n = 0;

    for (os = cs->objects; os; os = os->tail) {
	if (os->head != this) {
            if (os->head->activation[role][cs->cycle-1] > max_comp) {
                max_comp = os->head->activation[role][cs->cycle-1];
            }
            n++;
	}
    }
    if (n > 0) {
        lat = (lateralBaseline - max_comp) / (maxAct - minAct);
    }
    else {
        lat = lateralBaseline - restAct;
    }

    return lat;
}

/*............................................................................*/

void object_calculate_activation(ContentionScheduler *cs, Object *this)
{
    ObjectRole role;

    for (role = 0; role < (MaxRoles-1); role++)    {

	double previous			= this->activation[role][cs->cycle-1];
	double parentInput		= parental[OBJECT] * object_intentional_influence(cs, this, role);
	double environmentInput		= environmental[OBJECT] * object_environmental_influence(this, role);
	double selfInput		= self[OBJECT] * object_self_influence(cs, this, role);
	double lateralInput		= lateral[OBJECT] * object_lateral_influence(cs, this, role);
	double systemInput		= lateralInput + selfInput;
	double externalInput		= parentInput + environmentInput;
	double rawInput			= systemInput + externalInput;
	double input = normal_noise(rawInput, noiseSD[OBJECT]);

        /* Inhibit used objects until their activation is exceeded by that of a competitor */
        /* RPC: Wed Jul 21 15:40:53 1999                                                   */
        /* Minor adjustment: do this before calculating the current activation             */
        /* RPC: Fri Dec  7 14:04:19 2001                                                   */

        if (this->inhibit[role]) {
            ObjectList *os;

            for (os = cs->objects; os; os = os->tail) {
                if ((os->head != this) && (os->head->activation[role][cs->cycle-1] > this->activation[role][cs->cycle-1])) {
                    this->inhibit[role] = FALSE; /* Turn off inhibition */
                }
            }
        }

	this->activation[role][cs->cycle] = apply_equations(previous, persistence, input, maths);
    }
}

/******************************************************************************/
/******************************************************************************/

double effector_intentional_influence(ContentionScheduler *cs, Effector *this)
{
    double contribution, credibility;
    double total = 0.0;

    SchemaList *ss;

    for (ss = cs->schemas; ss; ss = ss->tail) {
	if ((credibility = ss->head->effectorTrigger(cs, this)) > 0) {
	    contribution = (ss->head->activation[cs->cycle-1] - restAct) / (maxAct - minAct);
	    contribution *= credibility;
	    total += contribution;
	}
    }
    return(squash(total));
}

/*----------------------------------------------------------------------------*/

double effector_self_influence(ContentionScheduler *cs, Effector *this)
{
    return((this->activation[cs->cycle-1] - minAct) / (maxAct - minAct));
}

/*----------------------------------------------------------------------------*/

double effector_lateral_influence(ContentionScheduler *cs, Effector *this)
{
  /* Effectors have no competitors. Lateral inhibition should be the average  */
  /* excitation of all competitors. With zero competitors, this average works */
  /* out to zero / zero. Therefore, set the mean lateral inhibition to        */
  /* what it would be if it had competitors at rest.                          */

    return(lateralBaseline - restAct);
}

/*----------------------------------------------------------------------------*/

void effector_calculate_activation(ContentionScheduler *cs, Effector *this)
{

/*  NB environmental influence in the effector domain is not implemented      */

    double previous		= this->activation[cs->cycle-1];
    double parentInput		= parental[EFFECTOR] * effector_intentional_influence(cs, this);
    double environmentInput	= environmental[EFFECTOR] * 0.0;
    double selfInput		= self[EFFECTOR] * effector_self_influence(cs, this);
    double lateralInput		= lateral[EFFECTOR] * effector_lateral_influence(cs, this);
    double systemInput		= lateralInput + selfInput;
    double externalInput	= parentInput + environmentInput;
    double rawInput		= systemInput + externalInput;
    double input 		= normal_noise(rawInput, noiseSD[EFFECTOR]);

    this->activation[cs->cycle] = apply_equations(previous, persistence, input, maths);
}

/******************************************************************************/
/* Display functions for debugging pruposes: **********************************/

static char *getGoalString(Schema *this)
{
    GoalList *gs;
    static char buffer[32];
    int j = 0;

    buffer[j++] = '[';

    for (gs = this->subgoals; gs && (j < 31); gs = gs->tail) {
        buffer[j++] = (gs->head->explicitly_achieved ? '+' : '-');
	buffer[j++] = ',';
    }
    buffer[((j > 1) ? j-1 : j++)] = ']';
    buffer[j] = '\0';

    return(buffer);
}

/*----------------------------------------------------------------------------*/

void list_schema_activation(ContentionScheduler *cs, Schema *this)
{
    extern void record_schema_activity(ContentionScheduler *cs, char *name, Boolean selected, double before, double s, double l, double p, double e, double in, double after, char *note);

    double p	= schema_parental_influence(cs, this);
    double e	= schema_environmental_influence(cs, this);
    double s	= schema_self_influence(cs, this);
    double l	= schema_lateral_influence(cs, this);
    double in	= parental[SCHEMA] * p + environmental[SCHEMA] * e + self[SCHEMA] * s + lateral[SCHEMA] * l;
    char *string = (this->selected ? getGoalString(this) : "");

    record_schema_activity(cs, this->name, this->selected, this->activation[cs->cycle-1], s, l, p, e, in, this->activation[cs->cycle], string);
}

/*::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::*/

void list_object_activation(ContentionScheduler *cs, Object *this, ObjectRole role)
{
    extern void record_object_activity(ContentionScheduler *cs, char *name, int r, Boolean selected, double before, double s, double l, double p, double e, double in, double after, char *note);
    extern void list_object_trace_attributes(char *trace, int l, Object *this);

    double p	= object_intentional_influence(cs, this, role);
    double e	= object_environmental_influence(this, role);
    double s	= object_self_influence(cs, this, role);
    double l	= object_lateral_influence(cs, this, role);
    double in	= parental[OBJECT] * p + environmental[OBJECT] * e + self[OBJECT] * s + lateral[OBJECT] * l;
    char   trace[80];

    list_object_trace_attributes(trace, 80, this);
    record_object_activity(cs, this->name, role, (cs->current_object[role] == this), this->activation[role][cs->cycle-1], s, l, p, e, in, this->activation[role][cs->cycle], trace);
}

/*::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::*/

static void list_effector_attributes(char *trace, int l, Effector *effector)
{
    ObjectList *os;

    if (effector->itemsHeld) {
        int m;

        g_snprintf(trace, l, "holds");
	for (os = effector->itemsHeld; os; os = os->tail) {
            m = strlen(trace);
	    if ((m < l) && (os->tail)) {
                g_snprintf(&trace[m], l - m, " %s,", os->head->name);
            }
	    else if (m < l) {
                g_snprintf(&trace[m], l - m, " %s", os->head->name);
	    }
        }
    }
    else {
	g_snprintf(trace, l, "empty");
    }
}

/*----------------------------------------------------------------------------*/

void list_effector_activation(ContentionScheduler *cs, Effector *this)
{
    extern void record_effector_activity(ContentionScheduler *cs, char *name, Boolean selected, double before, double s, double l, double p, double e, double in, double after, char *note);

    double p	= effector_intentional_influence(cs, this);
    double e	= 0.0;
    double s	= effector_self_influence(cs, this);
    double l	= effector_lateral_influence(cs, this);
    double in	= parental[EFFECTOR] * p + environmental[EFFECTOR] * e + self[EFFECTOR] * s + lateral[EFFECTOR] * l;
    char   trace[80];

    list_effector_attributes(trace, 80, this);
    record_effector_activity(cs, this->name, FALSE, this->activation[cs->cycle-1], s, l, p, e, in, this->activation[cs->cycle], trace);
}

/******************************************************************************/

void print_excitation_of_schema_by_objects(FILE *fp, ContentionScheduler *cs, Schema *s)
{
}

void print_excitation_of_object_by_schemas(FILE *fp, ContentionScheduler *cs, Object *this)
{
    double total = 0.0;
    double contribution;
    double credibility;
    ObjectRole role;

    SchemaList *ss;

    for (role = SOURCE; role < EMPTY; role++) {
        fprintf(fp, "Excitation of %s as %s by schemas:\n", this->name, roleName[role]);
	for (ss = cs->schemas; ss != NULL; ss = ss->tail) {
 	    if ((credibility = schema_triggers_object(cs, ss->head, this, role)) > 0) {
		contribution = (ss->head->activation[cs->cycle-1] - restAct) / (maxAct - minAct);
                fprintf(fp, "    %s: %f x %f\n", ss->head->name, contribution, credibility);
		total += (contribution * credibility);
	    }
	}
        fprintf(fp, "    Sum: %f; Squashed sum: %f\n", total, squash(total));
    }
}

void print_excitation_of_effector_by_schemas(FILE *fp, ContentionScheduler *cs, Effector *this)
{
    double contribution, credibility;
    double total = 0.0;

    SchemaList *ss;

    fprintf(fp, "Excitation of %s by schemas:\n", this->name);
    for (ss = cs->schemas; ss; ss = ss->tail) {
	if ((credibility = ss->head->effectorTrigger(cs, this)) > 0) {
	    contribution = (ss->head->activation[cs->cycle-1] - restAct) / (maxAct - minAct);
            fprintf(fp, "    %s: %f x %f\n", ss->head->name, contribution, credibility);
	    contribution *= credibility;
	    total += contribution;
	}
    }
    fprintf(fp, "    Sum: %f; Squashed sum: %f\n", total, squash(total));
}
