/*******************************************************************************

    FILE: object.c
    AUTHOR: R. Cooper (translated from C++ written by SVW)
    CREATION DATE: 
    CONTENTS: 

*******************************************************************************/

#include "cs.h"
#include <math.h>

/* Forward declarations: ******************************************************/

void objectDestroy(ContentionScheduler *cs, Object *this);
ObjectList *objectListDeleteElement(ObjectList *ol, Object *theme);
ObjectList *objectListAppendElement(ObjectList *ol, Object *theme);

/******************************************************************************/

#undef DEBUGGING

#ifdef DEBUGGING
static void print_sublist(FILE *fp, char *title, ObjectList *list)
{
    ObjectList *ol;

    if (list != NULL) {
        fprintf(fp, "  %s: ", title);
        for (ol = list; ol != NULL; ol = ol->tail) {
            fprintf(fp, "%p ", list->head);
	}
        fprintf(fp, " ||\n");
        for (ol = list; ol != NULL; ol = ol->tail) {
            fprintf(fp, "    %p = %s: ", ol->head, ol->head->name);
            fprintf(fp, "Container: %p; Carrier: %p; Base: %p\n", ol->head->container, ol->head->carrier, ol->head->base);
	}

    }
}

static void dump_object_pointers(FILE *fp, ObjectList *objects)
{
    ObjectList *ol;
    int i = 1;

    for (ol = objects; ol != NULL; ol = ol->tail) {
        fprintf(fp, "  Object %2d: %s: %p\n", i++, ol->head->name, ol->head);
        fprintf(fp, "    Container: %p; Carrier: %p; Base: %p\n", ol->head->container, ol->head->carrier, ol->head->base);
        print_sublist(fp, "  Contents", ol->head->itemsContained);
        print_sublist(fp, "  Carried", ol->head->itemsCarried);
        print_sublist(fp, "  Attachments", ol->head->itemsAttached);
    }
}
#endif

/******************************************************************************/

void objectDeleteContained(Object *container, Object *contained)
{
    container->itemsContained = objectListDeleteElement(container->itemsContained, contained);
    contained->container = NULL;
}

void objectDeleteCarried(Object *carrier, Object *carried)
{
    carrier->itemsCarried = objectListDeleteElement(carrier->itemsCarried, carried);
    carried->carrier = NULL;
}

void objectDeleteAttached(Object *base, Object *attached)
{
    base->itemsAttached = objectListDeleteElement(base->itemsAttached, attached);
    attached->base = NULL;
}

/******************************************************************************/
/* Action-related object functions: *******************************************/

void objectRemoveCover(Object *this, Object *cover)
{
    if (cover != NULL) {
        this->itemsCarried = objectListDeleteElement(this->itemsCarried, cover);
//        objectSetPosition(cover, 0, 0);
        cover->carrier = NULL;
        if (isContainerBoolean(this)) {
            this->state = (this->state | STATE_OPEN) & ~STATE_CLOSED;
        }
    }
}

/******************************************************************************/

static void initialise_object_activations(ContentionScheduler *cs)
{
    ObjectList *ol;
    ObjectRole role;

    for (ol = cs->objects; ol != NULL; ol = ol->tail) {
        for (role = 0; role < (MaxRoles-1) ; role++) {
            ol->head->activation[role][cs->cycle] = initial_activation(OBJECT);
        }
    }

    if (maths == NORMALISED) {
        normalise_object_activations(cs);
    }
}

static void objectListDestroyItems(ContentionScheduler *cs, Object *this, ObjectList *list, char *arg1, char *arg2)
{
#ifdef DEBUGGING
    Boolean echo = (list != NULL);

    if (echo) {
        fprintf(stderr, "->objectListDestroyItems while destroying %s items of %s\n", arg1, arg2);
    }
#endif
    while (list != NULL) {
        ObjectList *tmp = list->tail;

        if (list->head->container != NULL) {
            objectDeleteContained(list->head->container, list->head);
        }
        if (list->head->carrier != NULL) {
            objectDeleteCarried(list->head->carrier, list->head);
        }
        if (list->head->base != NULL) {
            objectDeleteAttached(list->head->base, list->head);
	}

        /* Try deleting the object from cs->objects, in case it exists in its own right */
        cs->objects = objectListDeleteElement(cs->objects, list->head);

        /* And make sure the item isn't also somewhere else  */
        this->itemsContained = objectListDeleteElement(this->itemsContained, list->head);
        this->itemsCarried = objectListDeleteElement(this->itemsCarried, list->head);
        this->itemsAttached = objectListDeleteElement(this->itemsAttached, list->head);

        objectDestroy(cs, list->head);
        free(list);

        list = tmp;
    }

#ifdef DEBUGGING
    if (echo) {
        fprintf(stderr, "<-objectListDestroyItems while destroying %s items of %s\n", arg1, arg2);
    }
#endif
}

void objectDestroy(ContentionScheduler *cs, Object *this)
{
    if (this != NULL) {

        ObjectList *tmp;
        ObjectRole role;

#ifdef DEBUGGING
        fprintf(stderr, "->objectDestroy/2 on %s ...\n", this->name);
        fprintf(stderr, "  Current objects before destruction are:\n");
        dump_object_pointers(stderr, cs->objects);
        fprintf(stderr, "  End of current objects\n\n");
#endif

        /* If the object is bound to a role, unbind it */
        for (role = 0; role < MaxRoles; role++) {
            if (cs->current_object[role] == this) {
                cs->current_object[role] = NULL;
            }
        }

        /* If X is in something, then remove X from that thing's contained list: */
        if (this->container != NULL) {
            objectDeleteContained(this->container, this);
        }

        /* If X is carried by something, then remove X from that thing's carried list: */
        if (this->carrier != NULL) {
            objectDeleteCarried(this->carrier, this);
        }

        /* If X is attached to something, then remove X from that thing's attached list: */
        if (this->base != NULL) {
            objectDeleteAttached(this->base, this);
        }

        if (this->itemsContained != NULL) {
            /* Destroy any things contained in X: */
            tmp = this->itemsContained; this->itemsContained = NULL;
            objectListDestroyItems(cs, this, tmp, "contained", this->name);
        }

        if (this->itemsCarried != NULL) {
            /* Destroy any things carried by X: */
            tmp = this->itemsCarried; this->itemsCarried = NULL;
            objectListDestroyItems(cs, this, tmp, "carried", this->name);

        }

        if (this->itemsAttached != NULL) {
            /* Destroy any things attached to X: */
            tmp = this->itemsAttached; this->itemsAttached = NULL;
            objectListDestroyItems(cs, this, tmp, "attached", this->name);
	}

        /* If X is held by something, then remove X from that thing's attached list: */
        if (this->holder != NULL) {
            this->holder->itemsHeld = objectListDeleteElement(this->holder->itemsHeld, this);
            objectSetPosition(this, 0.0, 0.0);
        }

        if (cs->fixated_object == this) {
            cs->fixated_object = NULL;
        }
        free(this->name);
        free(this);

#ifdef DEBUGGING
        fprintf(stderr, "  Current objects after destruction are:\n");
        dump_object_pointers(stderr, cs->objects);
        fprintf(stderr, "  End of current objects\n\n");
        fprintf(stderr, "<-objectDestroy/2\n");
#endif
    }
}

void destroy_objects(ContentionScheduler *cs)
{
#ifdef DEBUGGING
    fprintf(stderr, "Destroying all objects in the object list...\n");
#endif
    while (cs->objects != NULL) {
        ObjectList *tmp = cs->objects;
        cs->objects = cs->objects->tail;
#ifdef DEBUGGING
        fprintf(stderr, "Destroying %p (%s)\n", tmp->head, tmp->head->name);
#endif
        objectDestroy(cs, tmp->head);
#ifdef DEBUGGING
        fprintf(stderr, "-------------------------------------------------------------------------------\n");
#endif
        free(tmp);
    }
}

void initialise_objects(ContentionScheduler *cs)
{
    cs->task.install_objects(cs);
    initialise_object_activations(cs);
}

void reinitialise_objects(ContentionScheduler *cs)
{
    destroy_objects(cs);
    initialise_objects(cs);
}

/******************************************************************************/

char *string_copy(char *buffer)
{
    char *x = (char *)malloc(sizeof(char) * (strlen(buffer) + 1));
    strcpy(x, buffer);
    return(x);
}

Object *objectCreate(ContentionScheduler *cs, char *object_name, unsigned int features, int state)
{
    Object *new;
    ObjectRole r;

    if ((new = (Object *)malloc(sizeof(Object))) == NULL) {
        fprintf(stderr, "MEMORY ALLOCATION FAILURE: objectCreate.1\n");
    }
    else {
        new->name = string_copy(object_name);
        new->features = features;
        new->state = state;
        new->highlighted = FALSE;
        new->holder = NULL;
        new->itemsContained = NULL;
        new->itemsCarried = NULL;
        new->itemsAttached = NULL;
        new->container = NULL;
        new->carrier = NULL;
        new->base = NULL;
        new->holder = NULL;
        new->time0 = cs->cycle;
        for (r = 0; r < MaxRoles; r++) {
            new->activation[r][cs->cycle] = initial_activation(OBJECT);
            new->inhibit[r] = FALSE;
        }
    }
    return(new);
}

Object *objectInstallItem(ContentionScheduler *cs, char *name, unsigned int features)
{
    Object *item = objectCreate(cs, name, features, 0);
    return(item);
}

Object *objectInstall(ContentionScheduler *cs, char *object_name, unsigned int features, int state)
{
    ObjectList *ol;

    if ((ol = (ObjectList *)malloc(sizeof(ObjectList))) == NULL) {
        fprintf(stderr, "MEMORY ALLOCATION FAILURE: install_object.1\n");
        return(NULL);
    }
    else if ((ol->head = objectCreate(cs, object_name, features, state)) == NULL) {
        free(ol);
        return(NULL);
    }
    else {
        ol->tail = cs->objects;
        cs->objects = ol;
        return(ol->head);
    }
}

void objectSetPosition(Object *object, double x, double y)
{
    /* This could be more sophistocated: see Stephanie's C++ code */

    object->position[0] = x;
    object->position[1] = y;
    object->holder = NULL;
}

/*----------------------------------------------------------------------------*/

void objectInstallContents(Object *object, Object *content)
{
    if (content != NULL) {
        object->itemsContained = objectListAppendElement(object->itemsContained, content);
        objectSetPosition(content, object->position[0], object->position[1]);
        content->container = object;
    }
}

void objectInstallAttachment(Object *object, Object *item)
{
    ObjectList *ol;

    if (item != NULL) {
        if ((ol = (ObjectList *)malloc(sizeof(ObjectList))) == NULL) {
            fprintf(stderr, "MEMORY ALLOCATION FAILURE: objectInstallAttachment\n");
        }
        else {
            objectSetPosition(item, object->position[0], object->position[1]);
            ol->head = item;
            ol->tail = object->itemsAttached;
            item->base = object;
            object->itemsAttached = ol;
        }
    }
}

void objectInstallCover(Object *object, Object *cover)
{
    if (cover != NULL) {
        object->itemsCarried = objectListAppendElement(object->itemsCarried, cover);
        objectSetPosition(cover, object->position[0], object->position[1]);
        cover->carrier = object;
        if (isContainerBoolean(object)) {
            object->state = (object->state | STATE_CLOSED) & ~STATE_OPEN;
        }
    }
}

/******************************************************************************/

Object *objectGetAttachment(Object *this)
{
    if ((this == NULL) || (this->itemsAttached == NULL)) {
        return(NULL);
    }
    else {
        return(this->itemsAttached->head);
    }
}

Object *objectGetCover(Object *this)
{
    if ((this == NULL) || (this->itemsCarried == NULL)) {
        return(NULL);
    }
    else {
        return(this->itemsCarried->head);
    }
}

Object *objectGetContent(Object *this)
{
    if ((this == NULL) || (this->itemsContained == NULL)) {
        return(NULL);
    }
    else {
        return(this->itemsContained->head);
    }
}

/******************************************************************************/

ObjectList *objectListAppendElement(ObjectList *ol, Object *theme)
{
    ObjectList *tmp;

    if ((tmp = (ObjectList *)malloc(sizeof(ObjectList))) == NULL) {
        return(ol);
    }
    else {
        tmp->head = theme;
        tmp->tail = ol;
        return(tmp);
    }
}

/*----------------------------------------------------------------------------*/

ObjectList *objectListDeleteElement(ObjectList *ol, Object *this)
{
#ifdef DEBUGGING
//    ObjectList *tmp;
//    fprintf(stderr, "->objectListDeleteElement/2: Deleting %s from [", this->name);
//    for (tmp = ol; tmp != NULL; tmp = tmp->tail) {
//        fprintf(stderr, " <<%s>>", tmp->head->name);
//    }
//    fprintf(stderr, " ]\n");
#endif
    if (ol == NULL) {
        return(NULL);
    }
    else if (ol->head == this) {
        ObjectList *tmp = ol->tail;
        free(ol);
        return(tmp);
    }
    else if (ol->tail != NULL) {
        ObjectList *tmp = ol, *tmp2;
        while (tmp->tail != NULL) {
            if (tmp->tail->head == this) {
                tmp2 = tmp->tail;
                tmp->tail = tmp->tail->tail;
                free(tmp2);
                return(ol);
            }
            else {
                tmp = tmp->tail;
            }
        }
        return(ol);
    }
    else {
        return(ol);
    }
}

/******************************************************************************/
/* This is a recursive function which finds out whether this object is        */
/* directly or indirectly in contact with the theme object.                   */

Boolean isInSubordinateContactWith(Object *this, Object *theme)
{
    ObjectList *os;
    Boolean     found = FALSE;

    if (!theme) {
	return(FALSE);
    }

    if (theme->itemsCarried) {
	for (os=theme->itemsCarried; os && !found; os=os->tail) {
	    if (os->head == this) {
		found = TRUE;
	    }
            else {
		found = isInSubordinateContactWith(this, os->head);
	    }
	}
    }

    if (theme->itemsContained && !found) {
	for (os=theme->itemsContained; os && !found; os=os->tail) {
	    if (os->head == this) {
		found = TRUE;
	    }
	    else {
		found = isInSubordinateContactWith(this, os->head);
	    }
	}
    }

    if (theme->itemsAttached && !found) {
	for (os=theme->itemsAttached; os && !found; os=os->tail) {
	    if (os->head == this) {
		found = TRUE;
	    }
	    else {
		found = isInSubordinateContactWith(this, os->head);
	    }
	}
    }
    return(found);
}

/******************************************************************************/

Object *objectGetMostActive(ContentionScheduler *cs, ObjectRole r)
{
    ObjectList *os;
    Object *this = NULL;
    double tmp = 0.0;

    /* Note: This should only return NULL if there are no objects, */
    /* but there might be no objects if we eat them all!           */

    for (os = cs->objects; os != NULL; os = os->tail) {
        if (os->head->activation[r][cs->cycle-1] > tmp) {
            this = os->head;
            tmp = this->activation[r][cs->cycle-1];
        }
    }
    return(this);
}

/******************************************************************************/

double feature_overlap(int af, int bf)
{
    int num = 0, denom = 0;

    while ((af > 0) || (bf > 0)) {
        if ((af & 1) || (bf & 1)) {
            denom++;
        }
        if ((af & 1) && (bf & 1)) {
            num++;
        }
        af = af >> 1;
        bf = bf >> 1;
    }
    return((double) num / (double) denom);
}

double object_has_feature(Object *this, int feature)
{
    return((this->features & feature) ? 1.0 : 0.0);
}

/******************************************************************************/

static Boolean objectListElementIsHeld(ObjectList *ol)
{
    while (ol != NULL) {
        if (ol->head->holder != NULL) {
            return(TRUE);
        }
        else {
            ol = ol->tail;
        }
    }
    return(FALSE);
}

Boolean isOnTable(Object *this)
{
    if (this->holder != NULL) {
        return(FALSE);
    }
    else if ((this->base != NULL) && (this->base->holder != NULL)) {
        return(FALSE);
    }
    else if ((this->carrier != NULL) && (this->carrier->holder != NULL)) {
        return(FALSE);
    }
    else if ((this->container != NULL) && (this->container->holder != NULL)) {
        return(FALSE);
    }
    else if (objectListElementIsHeld(this->itemsAttached)) {
        return(FALSE);
    }
    else if (objectListElementIsHeld(this->itemsCarried)) {
        return(FALSE);
    }
    else if (objectListElementIsHeld(this->itemsContained)) {
        return(FALSE);
    }
    else {
        return(TRUE);
    }
}

GoalType isHeld(Object *this)
{
    return((this->holder != NULL) ? GOAL_TRUE : GOAL_FALSE);
}
 
GoalType hasAttachment(Object *this)
{
    return((this->itemsAttached != NULL) ? GOAL_TRUE : GOAL_FALSE);
}
 
GoalType isAttached(Object *this)
{
    return((this->base != NULL) ? GOAL_TRUE : GOAL_FALSE);
}

GoalType isCarried(Object *this)
{
    return((this->carrier != NULL) ? GOAL_TRUE : GOAL_FALSE);
}
 
GoalType isContained(Object *this)
{
    return((this->container != NULL) ? GOAL_TRUE : GOAL_FALSE);
}
 
GoalType isCovered(Object *this)
{
    return((this->itemsCarried != NULL) ? GOAL_TRUE : GOAL_FALSE);
}
 
GoalType isEmptyObject(Object *this)
{
    return((this->itemsContained == NULL) ? GOAL_TRUE : GOAL_FALSE);
}

GoalType isFullObject(Object *this)
{
    return((this->itemsContained != NULL) ? GOAL_TRUE : GOAL_FALSE);
}

GoalType isTool(Object *this)
{
    return(object_has_feature(this, IS_TOOL));
}

GoalType isContainer(Object *this)
{
    return(object_has_feature(this, IS_CONTAINER));
}
 
GoalType isSharp(Object *this)
{
    return(object_has_feature(this, IS_SHARP));
}

GoalType isSphere(Object *this)
{
    return(object_has_feature(this, IS_SPHERE));
}

GoalType wasBounced(Object *this)
{
    return((this->state & STATE_BOUNCED) ? GOAL_TRUE : GOAL_FALSE);
}

GoalType isClosed(Object *this)
{
    return((this->state & STATE_CLOSED) ? GOAL_TRUE : GOAL_FALSE);
}

GoalType isOpen(Object *this)
{
    return((this->state & STATE_OPEN) ? GOAL_TRUE : GOAL_FALSE);
}

GoalType isUsed(Object *this)
{
    return((this->state & STATE_USED) ? GOAL_TRUE : GOAL_FALSE);
}

GoalType isTable(Object *this)
{
    return((this->features == 0) ? GOAL_TRUE : GOAL_FALSE);
}

GoalType isEmptyContainer(Object *this)
{
    return(FZAND(isContainer(this), isEmptyObject(this)));
}

/*----------------------------------------------------------------------------*/

Boolean isEmptyObjectBoolean(Object *this)
{
    return(this->itemsContained == NULL);
}

Boolean isFullObjectBoolean(Object *this)
{
    return(this->itemsContained != NULL);
}

Boolean isOpenBoolean(Object *this)
{
    return((this->state & STATE_OPEN) ? TRUE : FALSE);
}

Boolean isClosedBoolean(Object *this)
{
    return((this->state & STATE_CLOSED) ? TRUE : FALSE);
}

Boolean isAttachedBoolean(Object *this)
{
    return(this->base != NULL);
}

Boolean isContainedBoolean(Object *this)
{
    return(this->container != NULL);
}
 
Boolean isHeldBoolean(Object *this)
{
    return(this->holder != NULL);
}

Boolean isCarriedBoolean(Object *this)
{
    return(this->carrier != NULL);
}

Boolean isCoveredBoolean(Object *this)
{
    return(this->itemsCarried != NULL);
}

Boolean isContainerBoolean(Object *this)
{
    return((this->features & IS_CONTAINER) ? TRUE : FALSE);
}

Boolean isFlatBoolean(Object *this)
{
    return((this->features & IS_FLAT) ? TRUE : FALSE);
}

Boolean isToolBoolean(Object *this)
{
    return((this->features & IS_TOOL) ? TRUE : FALSE);
}

Boolean isFluidBoolean(Object *this)
{
    return((this->features & IS_FLUID) ? TRUE : FALSE);
}

/******************************************************************************/
/* Display functions: *********************************************************/
/******************************************************************************/

void list_object_trace_attributes(char *trace, int l, Object *this)
{
    int m = 0;

    trace[m] = '\0';
    if (isContainerBoolean(this)) {
        m = strlen(trace);
        g_snprintf(&trace[m], l - m, "%s ", (isOpenBoolean(this) ? "open" : "closed"));
        m = strlen(trace);
        g_snprintf(&trace[m], l - m, "%s ", (isFullObjectBoolean(this) ? "full" : "empty"));
    }
    if (this->itemsCarried != NULL) {
        m = strlen(trace);
        g_snprintf(&trace[m], l - m, "covered by %s ", this->itemsCarried->head->name);
    }
    if (this->carrier != NULL) {
        m = strlen(trace);
        g_snprintf(&trace[m], l - m, "carried by %s ", this->carrier->name);
    }
    if (this->holder != NULL) {
        m = strlen(trace);
        g_snprintf(&trace[m], l - m, "held by %s", this->holder->name);
    }
    else if ((this->base != NULL) && (this->base->holder != NULL)) {
        m = strlen(trace);
        g_snprintf(&trace[m], l - m, "base held by %s", this->base->holder->name);
    }
}

static void write_generic_object_attributes(FILE *fp, Object *this)
{
    ObjectList *os, *os2;

    if (this->itemsAttached) {
        for (os=this->itemsAttached; os; os = os->tail) {
            fprintf(fp, "has %s attached, ",os->head->name);
        }
    }

    if (this->itemsContained != NULL) {
        fprintf(fp, "holds ");

        for (os=this->itemsContained; os; os = os->tail) {
            if (os->head->itemsContained) {
                fprintf(fp, "%s (containing ",os->head->name);

                for (os2=os->head->itemsContained; os2; os2 = os2->tail) {
                    if (os2->tail) {
                        fprintf(fp, "%s, ", os2->head->name);
                    }
                    else {
                        fprintf(fp, "%s", os2->head->name);
                    }
                }
                fprintf(fp, "), ");
            }
            else {
                fprintf(fp, "%s, ",os->head->name);
            }
        }
    }

    if (this->itemsCarried) {
        fprintf(fp, "covered by ");

        for (os=this->itemsCarried; os; os = os->tail) {
            fprintf(fp, "%s, ", os->head->name);
        }
    }

    if (this->holder != NULL) {
        fprintf(fp, "held by %s", this->holder->name);
    }
    else if (this->container != NULL) {
        fprintf(fp, "inside %s", this->container->name);
    }
    else if (this->carrier != NULL) {
        fprintf(fp, "on %s", this->carrier->name);
    }
    else if ((this->base != NULL) && (this->base->holder != NULL)) {
        fprintf(fp, "base held by %s", this->base->holder->name);
    }
    else {
        fprintf(fp, "on table");
    }
}

void traceObjectPreAction(ContentionScheduler *cs, Object *this)
{
    if (messages[WORLD] != NULL) {
        fprintf(messages[WORLD], "%5d\t%s: [", cs->cycle, this->name);
        cs->task.write_object_attributes(messages[WORLD], this);
        write_generic_object_attributes(messages[WORLD], this);
        if (this == cs->fixated_object) {
            fprintf(messages[WORLD], ", fixated");
        }
	fprintf(messages[WORLD], "]");
	fflush(messages[WORLD]);
    }
}

/*----------------------------------------------------------------------------*/

void traceObjectPostAction(ContentionScheduler *cs, Object *this)
{
    if (messages[WORLD] != NULL) {
        fprintf(messages[WORLD]," --> [");
        cs->task.write_object_attributes(messages[WORLD], this);
        write_generic_object_attributes(messages[WORLD], this);
        if (this == cs->fixated_object) {
            fprintf(messages[WORLD], ", fixated");
        }
        fprintf(messages[WORLD], "]\n");
    }
}

/*----------------------------------------------------------------------------*/

void traceObjectPostCreate(ContentionScheduler *cs, Object *this)
{
    if (messages[WORLD] != NULL) {
        fprintf(messages[WORLD], "%5d\t%s: [] --> [", cs->cycle, this->name);
        cs->task.write_object_attributes(messages[WORLD], this);
        write_generic_object_attributes(messages[WORLD], this);
        if (this == cs->fixated_object) {
            fprintf(messages[WORLD], ", fixated");
        }
        fprintf(messages[WORLD], "]\n");
    }
}

/*----------------------------------------------------------------------------*/

void traceObjectPostNameChange(ContentionScheduler *cs, Object *this)
{
    if (messages[WORLD] != NULL) {
        fprintf(messages[WORLD], " --> %s: [", this->name);
        cs->task.write_object_attributes(messages[WORLD], this);
        write_generic_object_attributes(messages[WORLD], this);
        if (this == cs->fixated_object) {
            fprintf(messages[WORLD], ", fixated");
        }
        fprintf(messages[WORLD], "]\n");
    }
}

/******************************************************************************/

Boolean object_can_serve_role(Object *this, ObjectRole role)
{
    // This is counter-productive in simple MOTs -- RPC: Tue Aug  3 15:44:46 2004

    return(TRUE);
}

/******************************************************************************/

void install_null_task_objects(ContentionScheduler *cs)
{
}

void write_null_task_object_attributes(FILE *fp, Object *this)
{
}

/******************************************************************************/
