/*******************************************************************************

    FILE: 
    AUTHOR: R. Cooper (translated from C++ written by SVW)
    CREATION DATE: 
    CONTENTS: 

*******************************************************************************/

#include "cs.h"

/******************************************************************************/

static Actions *action_list_free(Actions *actions)
{
    while (actions != NULL) {
        Actions *tmp = actions;
        actions = tmp->next;
        free(tmp);
    }
    return(actions);
}

/*----------------------------------------------------------------------------*/

ContentionScheduler *create_contention_scheduler()
{
    roleName[SOURCE] = "Source";
    roleName[TARGET] = "Target";
    roleName[THEME] = "Theme";
    roleName[IMPLEMENT] = "Implement";

    return((ContentionScheduler *)malloc(sizeof(ContentionScheduler)));
}

/*----------------------------------------------------------------------------*/

void initialise_contention_scheduler(ContentionScheduler *cs)
{
    ObjectRole role;

    cs->cycle = 0;
    cs->logfile = NULL;
    cs->actions = NULL;
    cs->cl_actions = NULL;
    cs->cl_analysis = NULL;
    cs->cl_schemas = NULL;
    for (role = 0; role < MaxRoles; role++) {
        cs->cl_objects[role] = NULL;
    }
    cs->cl_effectors = NULL;

    cs->schemas = NULL;
    cs->objects = NULL;
    cs->effectors = NULL;

    cs->update_goal_queue = NULL;
    cs->selection_queue = NULL;
    cs->deselection_queue = NULL;
    cs->execution_queue = NULL;

    random_seed_load("random_seed", cs);

    fprintf(stderr, "%5d\tInitialising... [Task: %s] ", cs->cycle, task_names[cs->task.id]); fflush(stderr);

    initialise_schemas(cs);
    initialise_objects(cs);
    initialise_effectors(cs);

    cs->fixated_object = NULL;
    for (role = 0; role < MaxRoles; role++) {
        cs->current_object[role] = NULL;
    }

    fprintf(stderr, "Done\n"); fflush(stderr);
}

/*----------------------------------------------------------------------------*/

void reinitialise_contention_scheduler(ContentionScheduler *cs, char *logfile)
{
    ObjectRole role;

    cs->actions = action_list_free(cs->actions);
    cs->cycle = 0;

    if (cs->logfile != NULL) {
        free(cs->logfile);
    }
    cs->logfile = string_copy(logfile);

    fprintf(stderr, "%5d\tReinitialising... [Task: %s; Log: %s] ", cs->cycle, task_names[cs->task.id], cs->logfile); fflush(stderr);
    fclose(fopen(cs->logfile, "w"));

    random_seed_load("5072/tmp.44.0.seed", cs);

    reinitialise_schemas(cs);
    reinitialise_objects(cs);
    reinitialise_effectors(cs);

    cs->fixated_object = NULL;
    for (role = 0; role < MaxRoles; role++) {
        cs->current_object[role] = NULL;
    }

    fprintf(stderr, "Done\n"); fflush(stderr);
}

/*----------------------------------------------------------------------------*/

void destroy_contention_scheduler(ContentionScheduler *cs)
{
    extern void destroy_schemas(ContentionScheduler *cs);
    extern void destroy_objects(ContentionScheduler *cs);
    extern void destroy_effectors(ContentionScheduler *cs);

    destroy_schemas(cs);
    destroy_objects(cs);
    destroy_effectors(cs);

    if (cs->logfile != NULL) {
        free(cs->logfile);
    }

    free(cs);
}

/******************************************************************************/

void message(MessageType m, ContentionScheduler *cs, char *string)
{
    if (messages[m] != NULL) {
        fprintf(messages[m], "%5d\t%s\n", cs->cycle, string);
    }
}

/******************************************************************************/

static void save_activations(ContentionScheduler *cs) {}

/*----------------------------------------------------------------------------*/

static void step_activations(ContentionScheduler *cs)
{
    SchemaList *sl;
    ObjectList *ol;
    EffectorList *el;

    for (sl = cs->schemas; sl != NULL; sl = sl->tail) {
        schema_calculate_activation(cs, sl->head);
    }

    for (ol = cs->objects; ol != NULL; ol = ol->tail) {
        object_calculate_activation(cs, ol->head);
    }

    for (el = cs->effectors; el != NULL; el = el->tail) {
        effector_calculate_activation(cs, el->head);
    }

    if (maths == NORMALISED) {
        normalise_schema_activations(cs);
        normalise_object_activations(cs);
        normalise_effector_activations(cs);
    }

    if (cs->data_logging) {
	save_activations(cs);
    }
}

/******************************************************************************/
/* Run one cycle, logging progress if necessary:                              */

static void run_one_cycle(ContentionScheduler *cs, Boolean evolve)
{
    extern Boolean all_intended_schemas_complete(ContentionScheduler *cs);
    extern void post_cycle_trace(ContentionScheduler *cs);
    extern void step_selections(ContentionScheduler *cs);

    cs->cycle++;

    step_activations(cs);
    step_selections(cs);

/*  update_cycle_display(cs); */

    if (evolve) {
        attach_data(cs);
    }

    if (all_intended_schemas_complete(cs)) {
	if (messages[WARNING] != NULL) {
	    fprintf(messages[WARNING], "*"); fflush(messages[WARNING]);
	}
    }

    post_cycle_trace(cs);
}

/*----------------------------------------------------------------------------*/
/* Run exactly n cycles:                                                      */

static void multi_run(ContentionScheduler *cs, int n, Boolean evolve)
{
    while (n-- > 0) {
	run_one_cycle(cs, evolve);
    }
}

/*----------------------------------------------------------------------------*/
/* Close the log file (done at the end of each run).                          */

static void terminate_logfile(ContentionScheduler *cs, Boolean ok)
{
    FILE *fp;

    if ((cs->logfile != NULL) && ((fp = fopen(cs->logfile, "a")) != NULL)) {
        fprintf(fp, "%s %d\n", (ok ? "complete" : "incomplete"), cs->cycle);
        fclose(fp);
    }
}

/*----------------------------------------------------------------------------*/
/* Return true if more cycles should be executed to complete the task:        */

static Boolean task_is_incomplete(ContentionScheduler *cs, Boolean evolve)
{
    extern Boolean all_intended_schemas_complete(ContentionScheduler *cs);

    if (cs->cycle >= (MaxCycles-1)) {
	terminate_logfile(cs, FALSE);
	message(WARNING, cs, "Too many cycles: terminating immediately.");
	return(FALSE);
    }
    else if (all_intended_schemas_complete(cs)) {
	terminate_logfile(cs, TRUE);
	message(WARNING, cs, "Trial complete.");
	message(WARNING, cs, "Running 10 more cycles before halting.");
	multi_run(cs, (cs->cycle > (MaxCycles-11)) ? (MaxCycles-cs->cycle-1) : 10, evolve);
	return(FALSE);
    }
    else {
        return(TRUE);
    }
}

/*----------------------------------------------------------------------------*/
/* Step up to n trials:                                                       */

void step(ContentionScheduler *cs, int n, Boolean evolve)
{
    extern void process_gtk_events();

    while ((n-- > 0) && task_is_incomplete(cs, evolve)) {
        process_gtk_events();
	run_one_cycle(cs, evolve);
    } 
}

/*----------------------------------------------------------------------------*/
/* Run the task until completion:                                             */

void run(ContentionScheduler *cs, Boolean evolve)
{
    extern void process_gtk_events();

    while (task_is_incomplete(cs, evolve)) {
        process_gtk_events();
	run_one_cycle(cs, evolve);
    } 
}

/******************************************************************************/
