/*******************************************************************************

    FILE: 
    AUTHOR: R. Cooper (translated from C++ written by SVW)
    CREATION DATE: 
    CONTENTS: 

*******************************************************************************/

#include "cs.h"

/******************************************************************************/
/* Find competitors: **********************************************************/

static Boolean is_parent(Schema *this, Schema *that)
{
    GoalList *sg;

    for (sg = this->subgoals; sg != NULL; sg = sg->tail) {
        if (strcmp(sg->head->name, that->goal) == 0) { 
            return(TRUE);
        }
    }
    return(FALSE);
}

/*----------------------------------------------------------------------------*/

static Boolean ancestor(ContentionScheduler *cs, Schema *this, Schema *that)
{
    if (this->subgoals == NULL) {
        return(FALSE);
    }
    else if (is_parent(this, that)) {
        return(TRUE);
    }
    else {
        SchemaList *ss;
        for (ss = cs->schemas; ss != NULL; ss = ss->tail) {
            if (is_parent(ss->head, that) && ancestor(cs, this, ss->head)) {
                return(TRUE);
            }
        }
        return(FALSE);
    }
}

static Boolean is_competitor(ContentionScheduler *cs, Schema *this, Schema *that)
{
    return((this != that) && !ancestor(cs, this, that) && !ancestor(cs, that, this) && (this->level == that->level));
}

static void compute_competitors(ContentionScheduler *cs)
{
    SchemaList *ss1, *ss2, *new;
    for (ss1 = cs->schemas; ss1 != NULL; ss1 = ss1->tail) {
        for (ss2 = cs->schemas; ss2 != NULL; ss2 = ss2->tail) {
            if (is_competitor(cs, ss1->head, ss2->head)) {
                if ((new = (SchemaList *)malloc(sizeof(SchemaList))) != NULL) {
                    new->head = ss2->head;
                    new->tail = ss1->head->competitors;
                    ss1->head->competitors = new;
                }
            }
	}
    }
}

/******************************************************************************/
/* Schema initialisation/destruction: *****************************************/

static void initialise_schema_activations(ContentionScheduler *cs)
{
    SchemaList *sl;
    GoalList *gl;

    for (sl = cs->schemas; sl != NULL; sl = sl->tail) {
        sl->head->activation[cs->cycle] = initial_activation(SCHEMA);
        sl->head->selected = FALSE;
        for (gl = sl->head->subgoals; gl != NULL; gl = gl->tail) {
            gl->head->explicitly_achieved = FALSE;
        }
    }

    if (maths == NORMALISED) {
        normalise_schema_activations(cs);
    }
}

static void free_subgoal_list(GoalList *gl)
{
    while (gl != NULL) {
        GoalList *tmp = gl->tail;
        free(gl->head);
        free(gl);
        gl = tmp;
    }
}

static void free_competitor_list_glue(SchemaList *sl)
{
    while (sl != NULL) {
        SchemaList *tmp = sl->tail;
        free(sl);
        sl = tmp;
    }
}

static void destroy_schema(Schema *s)
{
    free_subgoal_list(s->subgoals);
    free_competitor_list_glue(s->competitors);
    free(s);
};

void destroy_schemas(ContentionScheduler *cs)
{
    while (cs->schemas != NULL) {
        SchemaList *tmp = cs->schemas;
        cs->schemas = cs->schemas->tail;
        destroy_schema(tmp->head);
        free(tmp);
    }
}

void initialise_schemas(ContentionScheduler *cs)
{
    cs->task.install_schemas(cs);
    compute_competitors(cs);    
    initialise_schema_activations(cs);
}

void reinitialise_schemas(ContentionScheduler *cs)
{
    destroy_schemas(cs);
    initialise_schemas(cs);
}

/******************************************************************************/

Schema *install_schema(ContentionScheduler *cs, char *schema_name, char *goal_name, ActionType act, SchemaLevel level)
{
    SchemaList *sl;
    int r;

    if ((sl = (SchemaList *)malloc(sizeof(SchemaList))) == NULL) {
	fprintf(stderr, "MEMORY ALLOCATION FAILURE: install_schema.1\n");
    }
    else if ((sl->head = (Schema *)malloc(sizeof(Schema))) == NULL) {
	fprintf(stderr, "MEMORY ALLOCATION FAILURE: install_schema.2\n");
        free(sl);
    }
    else {
        sl->head->name = schema_name;
        sl->head->goal = goal_name;
        sl->head->competitors = NULL;
        sl->head->highlighted = FALSE;
        sl->head->selected = FALSE;
        sl->head->executed = FALSE;
        sl->head->corresponding_action = act;
        sl->head->level = level;
        sl->head->objectTriggers = NULL;
        sl->head->effectorTrigger = NULL;
        sl->head->competitors = NULL;
        sl->head->subgoals = NULL;
        for (r = 0; r < MaxRoles-1; r++) {
            sl->head->scope[r] = FALSE;
        }
        for (r = 0; r < RESOURCES_MAX; r++) {
            sl->head->resource[r] = FALSE;
        }
        sl->tail = cs->schemas;
        cs->schemas = sl;
    }
    return(sl->head);
}

void install_schema_object_trigger(Schema *s, ObjectRole role, double object_trigger())
{
    ObjectTriggerFunctionList *new;

    if ((new = (ObjectTriggerFunctionList *)malloc(sizeof(ObjectTriggerFunctionList))) != NULL) {
        new->next = s->objectTriggers;
        new->role = role;
        new->trigger = object_trigger;
        s->objectTriggers = new;
    }
}

void install_schema_effector_trigger(Schema *s, double effector_trigger())
{
    s->effectorTrigger = effector_trigger;
}

/******************************************************************************/

void install_schema_subgoal(Schema *s, char *goal_name, GoalType precondition(), GoalType postcondition())
{
    GoalList *gl;

    if ((gl = (GoalList *)malloc(sizeof(GoalList))) == NULL) {
	fprintf(stderr, "MEMORY ALLOCATION FAILURE: install_schema_subgoal.1\n");
    }
    else if ((gl->head = (Goal *)malloc(sizeof(Goal))) == NULL) {
	fprintf(stderr, "MEMORY ALLOCATION FAILURE: install_schema_subgoal.2\n");
        free(gl);
    }
    else {
        gl->head->name = goal_name;
        gl->head->explicitly_achieved = FALSE;
        gl->head->precondition = precondition;
        gl->head->postcondition = postcondition;

        gl->tail = s->subgoals;
        s->subgoals = gl;
    }
}

void install_schema_scope(Schema *s, ObjectRole role)
{
    s->scope[role] = TRUE;
}

void install_schema_resource(Schema *s, Resource res)
{
    s->resource[res] = TRUE;
}

/******************************************************************************/

#ifdef FUZZY_GOALS

Boolean goalPreConditionAchieved(ContentionScheduler *cs, Goal *goal)
{
    return(goal->precondition(cs) > 0.99);
}

Boolean goalPostConditionAchieved(ContentionScheduler *cs, Goal *goal)
{
    return(goal->postcondition(cs) > 0.99);
}

#else

Boolean goalPreConditionAchieved(ContentionScheduler *cs, Goal *goal)
{
    return(goal->precondition(cs));
}

Boolean goalPostConditionAchieved(ContentionScheduler *cs, Goal *goal)
{
    return(goal->postcondition(cs));
}

#endif

/******************************************************************************/

void update_goals(ContentionScheduler *cs, Schema *this)
{
    /* This function marks any subgoal achieved on the current cycle as */
    /* achieved. It is called for each selected schema on every cycle.  */

    GoalList *gs;

//    fprintf(stdout, "%d: Updating goals of %s\n", cs->cycle, this->name);

    for (gs = this->subgoals; gs != NULL; gs = gs->tail) {
        if (!gs->head->explicitly_achieved && goalPostConditionAchieved(cs, gs->head)) {
            gs->head->explicitly_achieved = TRUE;
//            fprintf(stdout, "    %s is now achieved\n", gs->head->name);
	}
    }
}

/******************************************************************************/

static Boolean schemaSubtasksComplete(Schema *this)
{
    GoalList *gs;

    for (gs = this->subgoals; gs != NULL; gs = gs->tail)  {
	if (!gs->head->explicitly_achieved) {
	    return(FALSE);
	}
    }
    return(TRUE);
}

/*----------------------------------------------------------------------------*/

static Boolean schemaGoalStateAchieved(ContentionScheduler *cs, Schema *this)
{
    extern Schema *selected_parent(ContentionScheduler *cs, Schema *this);

    SchemaList *ss;
    GoalList *goals;
    Schema *parent;

    if ((parent = selected_parent(cs, this)) != NULL) {
        for (goals = parent->subgoals; goals != NULL; goals = goals->tail) {
            if ((strcmp(this->goal, goals->head->name)) == 0) {
                return(goalPostConditionAchieved(cs, goals->head));
            }
        }
    }
    else {
        /* I'm not sure if this is right. The problem is that rogue schemas */
        /* don't really have goals (postconditions), so it is unclear if it */
        /* makes sense to talk about their goal state being achieved.       */

        for (ss = cs->schemas; ss != NULL; ss = ss->tail) {
            for (goals = ss->head->subgoals; goals != NULL; goals = goals->tail) {
                if ((strcmp(this->goal, goals->head->name)) == 0) {
                    return(goalPostConditionAchieved(cs, goals->head));
                }
            }
        }
    }
    /* We should never get to here: */
    return(0.0);
}

/*----------------------------------------------------------------------------*/

Boolean schema_goal_achieved(ContentionScheduler *cs, Schema *this)
{
    /* This function is only called for selected schemas.                   */
    /* A selected schema's goal is achieved in two cases:                   */
    /*  1) If the schema doesn't correspond to a primitive action, then it  */
    /*  is achieved if all its subgoals are marked as being achieved;       */
    /*  2) If the schema does correspond to a primitive action, then it is  */
    /*  achieved if its corresponding action has been executed.             */

    if (this->corresponding_action == ACTION_NONE) {
        return(schemaGoalStateAchieved(cs, this) || schemaSubtasksComplete(this));
    }
    else {
        return(this->executed);
    }
}

/******************************************************************************/

static Boolean goal_has_been_achieved(ContentionScheduler *cs, Goal *head)
{
    return(head->explicitly_achieved || goalPostConditionAchieved(cs, head));
}

Boolean all_intended_schemas_complete(ContentionScheduler *cs)
{
    Schema    *this = cs->intended_schema;
    Boolean    complete = TRUE;
    GoalList  *gs;

    if (this != NULL) {
        /* TRUE if every subgoal of the specified schema is either:           */
        /*   explicitly marked as achieved, has true post conditions, or is   */
        /*   achieved in virtue of corresponding to a schema that is achieved */

        for (gs = this->subgoals; gs && complete; gs = gs->tail) {
            complete = goal_has_been_achieved(cs, gs->head);
        }
    }
    return(complete);
}

/******************************************************************************/

void install_null_task_schemas(ContentionScheduler *cs)
{
    cs->intended_schema = NULL;
}

/******************************************************************************/
