/*******************************************************************************

    FILE: 
    AUTHOR: R. Cooper (translated from C++ written by SVW)
    CREATION DATE: 
    CONTENTS: 

*******************************************************************************/

#include "cs.h"

/*----------------------------------------------------------------------------*/

extern void action_log_add_warning(ContentionScheduler *cs, char *type, char *warning);
extern void execute_action(ContentionScheduler *cs, Schema *schema);
extern void update_goals(ContentionScheduler *cs, Schema *this);

/******************************************************************************/
/* If this has a parent schema that is selected, return a pointer to that     */
/* parent schema.                                                             */

static Boolean schema_is_child(Schema *parent, Schema *child)
{
    GoalList *subgoals = parent->subgoals;

    while (subgoals != NULL) {
        if (strcmp(subgoals->head->name, child->goal) == 0) {
            return(TRUE);
        }
        else {
            subgoals = subgoals->tail;
        }
    }
    return(FALSE);
}

Schema *selected_parent(ContentionScheduler *cs, Schema *this)
{
    SchemaList *ss;

    for (ss = cs->schemas; ss != NULL; ss = ss->tail) {
        if ((ss->head->selected) && schema_is_child(ss->head, this)) {
            return(ss->head);
        }
    }
    return(NULL);
}

/******************************************************************************/
/* An unselected schema whose activation is above threshold and which is      */
/* the winner amongst its competitors should be selected.                     */

static Boolean schema_is_selectable(ContentionScheduler *cs, Schema *this)
{
    return((!this->selected) && (this->activation[cs->cycle-1] > selectionThreshold));
}

/*----------------------------------------------------------------------------*/
/* A schema is the winner if no competitor or sibling has greater activation. */

static Boolean schema_list_member(SchemaList *list, Schema *element)
{
    while (list != NULL) {
        if (list->head == element) {
            return(TRUE);
        }
        else {
            list = list->tail;
        }
    }
    return(FALSE);
}

static Boolean schema_is_winning_schema(ContentionScheduler *cs, Schema *this)
{
    SchemaList *ss;
    Boolean winner = TRUE;

    for (ss = cs->schemas; ss && winner; ss = ss->tail) {
	if (this->activation[cs->cycle-1] < ss->head->activation[cs->cycle-1]) {
	    if (schema_list_member(this->competitors, ss->head)) {
		winner = FALSE;
	    }
	}
    }
    return(winner);
}

/*----------------------------------------------------------------------------*/
/* An unselected schema whose activation is above threshold and which is      */
/* the winner amongst its competitors should be selected.                     */

static Boolean schema_should_be_selected(ContentionScheduler *cs, Schema *this)
{
    return(schema_is_selectable(cs, this) && schema_is_winning_schema(cs, this));
}

/*----------------------------------------------------------------------------*/

Boolean schema_should_be_deselected(ContentionScheduler *cs, Schema *this)
{
    Schema *parent;

    if (!this->selected) {
	return FALSE;
    }
    else if (this == cs->intended_schema) {
        /* The highest level schema should not be deselected: */
	return(FALSE);
    }
    else if (!schema_is_winning_schema(cs, this)) {
        /* A selected schema that is no longer the winner amongst its         */
        /* competitors and siblings should be deselected.                     */
	return(TRUE);
    }
    else if ((parent = selected_parent(cs, this)) != NULL) {
        /* A schema that wins amongst its competitors and siblings should be  */
        /* deselected if its parent is deselected. This solves a              */
        /* synchronisation problem that arose when the activation of an       */
        /* achieved parent fell before its last child could be deselected.    */
        return(schema_should_be_deselected(cs, parent));
    }
    else {
        /* A schema that wins amongst its competitors and siblings should be  */
        /* deselected if it has a parent that should be selected.             */

	Boolean parent_should_be_selected = FALSE;
	SchemaList *ss;

	for (ss = cs->schemas; ((ss != NULL) && !parent_should_be_selected); ss = ss->tail) {
            if (schema_is_child(ss->head, this) && schema_should_be_selected(cs, ss->head)) {
		parent_should_be_selected = TRUE;
	    }
	}
	return(parent_should_be_selected);
    }
}

/******************************************************************************/

extern Boolean goalPreConditionAchieved(ContentionScheduler *cs, Goal *goal);
extern Boolean goalPostConditionAchieved(ContentionScheduler *cs, Goal *goal);
extern void record_intervention(ContentionScheduler *cs, char *action);

/******************************************************************************/
/* This function is called whenever a schema is deslected. All the bound      */
/* objects in the schema's scope are nullified so that negative influence     */
/* can no longer be applied to the objects which have been suppressed. This   */
/* function is also responsible for switching on self-inhibition of utilised  */
/* objects in scope.                                                          */

static void unbind_arguments(ContentionScheduler *cs, Schema *this)
{
    char buffer[256];
    ObjectRole r;

    for (r = 0; r < MaxRoles-1; r++) {
        if (this->scope[r]) {
            if (cs->current_object[r] != NULL) {
                g_snprintf(buffer, 256, "%s: Unbinding %s as %s", this->name, cs->current_object[r]->name, roleName[r]);
                message(BIND, cs, buffer);
                cs->current_object[r]->inhibit[r] = TRUE;
                cs->current_object[r] = NULL;
            }
            else {
                g_snprintf(buffer, 256, "UNBIND ERROR: %s: Cannot unbind NULL %s", this->name, roleName[r]);
                message(BIND, cs, buffer);
            }
        }
    }
}

static void unbind_all_roles(ContentionScheduler *cs, char *agent)
{
    char buffer[256];
    ObjectRole r;

    for (r = 0; r < MaxRoles-1; r++) {
        if (cs->current_object[r] != NULL) {
            g_snprintf(buffer, 256, "%s: Unbinding %s as %s", agent, cs->current_object[r]->name, roleName[r]);
            message(BIND, cs, buffer);
            cs->current_object[r]->inhibit[r] = TRUE;
            cs->current_object[r] = NULL;
        }
    }
}

/*----------------------------------------------------------------------------*/

static void do_selection_bookkeeping(ContentionScheduler *cs, Schema *schema)
{
    /* Intervention is only relevant when deselecting, not when selecting.  */
    /* Even so, it is helpful to report any precondition failures.          */

    if (interventions) {
        Schema *parent = selected_parent(cs, schema);
        char string[128];

	if (parent != NULL) {
            Boolean preconditions_ok = FALSE;
            GoalList *gs;

            /* This schema could be relevant for several subgoals. Check    */
            /* that for at least one the preconditions are true and the     */
            /* postconditions are false.                                    */

	    for (gs = parent->subgoals; gs != NULL; gs = gs->tail)  {
		if (strcmp(schema->goal, gs->head->name) == 0) {
                    if (goalPreConditionAchieved(cs, gs->head) && !goalPostConditionAchieved(cs, gs->head)) {
                        preconditions_ok = TRUE;
                    }
                }
            }
            if (!preconditions_ok) {
                action_log_add_warning(cs, "PRECONDITION FAILURE", schema->goal);
                g_snprintf(string, 128, "Failed precondition of %s within %s", schema->goal, parent->name);
                message(MONITOR, cs, string);
                record_intervention(cs, "false selection");
            }
        }
    }
}

static void do_deselection_bookkeeping(ContentionScheduler *cs, Schema *schema)
{
    Boolean found = FALSE;
    char buffer[256];
    Schema *parent;
    GoalList *gs;

    /* Unbind its arguments: */

    if (schema->corresponding_action == ACTION_NONE) {
        unbind_arguments(cs, schema);
    }

    /* Set the explicitly_achieved flag of the selected parent for */
    /* the child schema's goal:                                    */

    if ((parent = selected_parent(cs, schema)) != NULL) {
        for (gs = parent->subgoals; gs != NULL; gs = gs->tail)  {
            if (strcmp(schema->goal, gs->head->name) == 0) {
                if (goalPostConditionAchieved(cs, gs->head)) {
                    gs->head->explicitly_achieved = TRUE;
                    found = TRUE;
                }
            }
        }
        if (!found && interventions) {
            unbind_all_roles(cs, "SAS");

            /* The following loop ensures that the failed goals are retried: */
            for (gs = parent->subgoals; gs != NULL; gs = gs->tail)  {
                if (strcmp(schema->goal, gs->head->name) == 0) {
                    gs->head->explicitly_achieved = FALSE;
                }
            }

            action_log_add_warning(cs, "POSTCONDITION FAILURE", schema->goal);
            g_snprintf(buffer, 256, "Failed postcondition of %s within %s", schema->goal, parent->name);
            message(MONITOR, cs, buffer);
            record_intervention(cs, "false deselection");
        }
    }
    else {
        unbind_all_roles(cs, "Monitoring");
        action_log_add_warning(cs, "Deselection with unselected parent", schema->name);
        g_snprintf(buffer, 256, "Null parent detected when deselecting %s", schema->name);
        message(MONITOR, cs, buffer);
        record_intervention(cs, "null selected parent");
    }
}

static void reinitialise_subgoal_list(Schema *this)
{
    /* Whether selected or deselected, reinitialise the subgoal list: */

    GoalList *gs;

    for (gs = this->subgoals; gs != NULL; gs = gs->tail)  {
	gs->head->explicitly_achieved = FALSE;
    }
}

/*----------------------------------------------------------------------------*/

static void report_selection(ContentionScheduler *cs, Schema *schema, Boolean sign)
{
    Schema *parent;
    char buffer[80];

    if ((parent = selected_parent(cs, schema)) != NULL) {
	g_snprintf(buffer, 80, "%*s%c%s under %s", 2, "", (sign?'+':'-'), schema->name, parent->name);
    }
    else {
	g_snprintf(buffer, 80, "%*s%c%s with no selected parent", 2, "", (sign?'+':'-'), schema->name);
    }

    if (messages[SELECTION] != NULL) {
        fprintf(messages[SELECTION], "%s\n", buffer);
    }

    if (messages[RUN] != NULL) {
        Schema *tmp = schema;

	fprintf(messages[RUN], "%5d", cs->cycle);
        while (tmp != NULL) {
	    fprintf(messages[RUN], "   ");
            tmp = selected_parent(cs, tmp);
        }
	fprintf(messages[RUN], "%c%s,\n", (sign?'+':'-'), schema->name);
    }
}

/*----------------------------------------------------------------------------*/

static void adjust_selection(ContentionScheduler *cs, Schema *schema, Boolean sign)
{
    extern void action_log_add_selection(ContentionScheduler *cs, char *schema_name, Boolean on);

    schema->selected = sign;
    if (sign) {
        do_selection_bookkeeping(cs, schema);
    }
    else {
        do_deselection_bookkeeping(cs, schema);
    }
    reinitialise_subgoal_list(schema);
    action_log_add_selection(cs, schema->name, sign);
    report_selection(cs, schema, sign);
}

/******************************************************************************/

static void queueForGoalUpdate(ContentionScheduler *cs, Schema *schema)
{
    SchemaList *ss;

    if ((ss = (SchemaList *)malloc(sizeof(SchemaList))) == NULL) {
	fprintf(stderr, "MEMORY ALLOCATION FAILURE: queueForGoalUpdate\n");
    }
    else {
        ss->head = schema;
	ss->tail = cs->update_goal_queue;
	cs->update_goal_queue = ss;
    }
}

/*----------------------------------------------------------------------------*/

static void queueForDeselection(ContentionScheduler *cs, Schema *schema)
{
    SchemaList *ss;

    if ((ss = (SchemaList *)malloc(sizeof(SchemaList))) == NULL) {
	fprintf(stderr, "MEMORY ALLOCATION FAILURE: queueForDeselection\n");
    }
    else {
        ss->head = schema;
	ss->tail = cs->deselection_queue;
	cs->deselection_queue = ss;

        /* Other coding is simpler if the deselection cue is ordered, with    */
        /* children before their parents ... so shuffle the new element into  */
        /* a sensible place.                                                  */

        ss = cs->deselection_queue;

        while ((ss->tail != NULL) && (ss->tail->head != selected_parent(cs, ss->head))) {
            ss->head = ss->tail->head;
            ss->tail->head = schema;
            ss = ss->tail;
        }
    }
}

/*----------------------------------------------------------------------------*/

static void queueForSelection(ContentionScheduler *cs, Schema *schema)
{
    SchemaList *ss;

    if ((ss = (SchemaList *)malloc(sizeof(SchemaList))) == NULL) {
	fprintf(stderr, "MEMORY ALLOCATION FAILURE: queueForSelection\n");
    }
    else {
        ss->head = schema;
	ss->tail = cs->selection_queue;
	cs->selection_queue = ss;
    }
}

/*----------------------------------------------------------------------------*/

static void queueForExecution(ContentionScheduler *cs, SchemaList *listElement)
{
    listElement->tail = cs->execution_queue;
    cs->execution_queue = listElement;
}

/******************************************************************************/

static void create_selection_queues(ContentionScheduler *cs)
{
    SchemaList *ss;

    for (ss = cs->schemas; ss != NULL; ss = ss->tail) {
	if (ss->head->selected) {
	    if (schema_should_be_deselected(cs, ss->head))  {
		queueForDeselection(cs, ss->head);
	    }
	    else {
		queueForGoalUpdate(cs, ss->head);
	    }
	}
	else if (schema_should_be_selected(cs, ss->head)) {
	    queueForSelection(cs, ss->head);
	}
    }
}

/*----------------------------------------------------------------------------*/

static void process_selection_queues(ContentionScheduler *cs)
{
    while (cs->update_goal_queue != NULL) {
	SchemaList *ss = cs->update_goal_queue;
        update_goals(cs, ss->head);
	cs->update_goal_queue = ss->tail;
	free(ss);
    }
    while (cs->selection_queue != NULL) {
	SchemaList *ss = cs->selection_queue;
	adjust_selection(cs, ss->head, TRUE);
	cs->selection_queue = ss->tail;
	if (ss->head->corresponding_action == ACTION_NONE) {
	    free(ss);
	}
	else {
	    queueForExecution(cs, ss);
	}
    }
    while (cs->deselection_queue != NULL) {
	SchemaList *ss = cs->deselection_queue;
	adjust_selection(cs, ss->head, FALSE);
	cs->deselection_queue = ss->tail;
	free(ss);
    }
    while (cs->execution_queue != NULL) {
	SchemaList *ss = cs->execution_queue;
	execute_action(cs, ss->head);
	ss->head->executed = TRUE;
	cs->execution_queue = ss->tail;
	free(ss);
    }
}

/*----------------------------------------------------------------------------*/

void step_selections(ContentionScheduler *cs)
{
    create_selection_queues(cs);
    process_selection_queues(cs);
}

/******************************************************************************/
