/*******************************************************************************

    FILE: xvcs.c
    AUTHOR: R. Cooper (translated from C++ written by SVW)
    CREATION DATE: Sun Oct 15 18:33:51 2000
    CONTENTS: GTK+ routines for X interface to CS

*******************************************************************************/

#include "cs.h"

#include <X11/Xlib.h>
#include <gtk/gtk.h>

/******************************************************************************/

extern void read_default_parameters(GtkWidget *button, ContentionScheduler *cs);
extern char *string_copy(char *buffer);

typedef enum {VARY_NONE, VARY_SELF_LAT, VARY_INT_EXT, VARY_SCHEMA_OBJECT, VARY_REST, VARY_IER, VARY_NOISE} VaryType;

typedef struct parameter_space {
    VaryType vary;
    char *outputDir;
    double increment;
    int incPrecision;
    int maxTrials;
    int startNum1;
    int endNum1;
    int startNum2;
    int endNum2;
} ParameterSpace;

#define INITIAL_TASK CANDLE_BASIC

/******************************************************************************/
/* Global Variables: **********************************************************/

#include "globals.h"

ContentionScheduler *scheduler;
extern GtkWidget *make_widgets();
extern void set_frame_title(TaskTypeId tid);

/*----------------------------------------------------------------------------*/

static void set_logfile(ContentionScheduler *cs, ParameterSpace *ps, double p1, double p2, int replication)
{
    char pathname[128];
    FILE *fp;

    if (cs->logfile != NULL) {
        free(cs->logfile);
    }
    switch (ps->vary) {
        case VARY_INT_EXT: {
            g_snprintf(pathname, 128, "%s/log.par%1.*fext%1.*f.%03d", ps->outputDir, ps->incPrecision, p1, ps->incPrecision, p2, replication);
            break;
        }
        case VARY_SELF_LAT: {
            g_snprintf(pathname, 128, "%s/log.self%1.*flat%1.*f.%03d", ps->outputDir, ps->incPrecision, p1, ps->incPrecision, p2, replication);
            break;
        }
        case VARY_SCHEMA_OBJECT: {
            g_snprintf(pathname, 128, "%s/log.schema%1.*fobject%1.*f.%03d", ps->outputDir, ps->incPrecision, p1, ps->incPrecision, p2, replication);
            break;
        }
        case VARY_NOISE: {
            g_snprintf(pathname, 128, "%s/log.noise%1.*f.%03d", ps->outputDir, ps->incPrecision, p1, replication);
            break;
        }
        case VARY_IER: {
            g_snprintf(pathname, 128, "%s/log.ier%1.*f.%03d", ps->outputDir, ps->incPrecision, p1, replication);
            break;
        }
        case VARY_REST: {
            g_snprintf(pathname, 128, "%s/log.rest%1.*f.%03d", ps->outputDir, ps->incPrecision, p1, replication);
            break;
        }
        default: {
            fprintf(stderr, "Unhandled case in set_logfile\n");
            break;
        }
    }
    cs->logfile = string_copy(pathname);
    if ((fp = fopen(cs->logfile, "w")) != NULL) { fclose(fp); }
}

/*----------------------------------------------------------------------------*/

static void run_one_trial(ContentionScheduler *cs, int k, int i, int j)
{
    fprintf(stderr, "BEGIN TRIAL %d.%d.%d\n", k + 1, i, j);
    reinitialise_contention_scheduler(cs, "tmp.log");
    run(cs, FALSE);
}

/*----------------------------------------------------------------------------*/

static void scan_noise_space(ParameterSpace *space, ContentionScheduler *cs, int k)
{
    int i;

    for (i = space->startNum1; i <= space->endNum1; i++) {
        noiseSD[SCHEMA] = i * space->increment;
        noiseSD[OBJECT] = i * space->increment;
        noiseSD[EFFECTOR] = i * space->increment;
        set_logfile(cs, space, noiseSD[SCHEMA], 0, k);
        run_one_trial(cs, k, i, 1);
    }
}

static void scan_ier_space(ParameterSpace *space, ContentionScheduler *cs, int k)
{
    int i;

/* In the default case: Parent + Environment = 0.6 */

    for (i = space->startNum1; i <= space->endNum1; i++) {
        parental[SCHEMA] = i * space->increment * 0.6;
        environmental[SCHEMA] = 0.6 - parental[SCHEMA];
        set_logfile(cs, space, parental[SCHEMA], environmental[SCHEMA], k);
        run_one_trial(cs, k, i, 1);
    }
}

static void scan_rest_space(ParameterSpace *space, ContentionScheduler *cs, int k)
{
    int i;

    for (i = space->startNum1; i <= space->endNum1; i++) {
        restAct = i * space->increment;
        set_logfile(cs, space, restAct, 0, k);
        run_one_trial(cs, k, i, 1);
    }
}

static void scan_self_lat_space(ParameterSpace *space, ContentionScheduler *cs, int k)
{
    int i, j;

    for (i = space->startNum1; i <= space->endNum1; i++) {
        self[SCHEMA] = i * space->increment;
        self[OBJECT] = i * space->increment;
        self[EFFECTOR] = i * space->increment;
        for (j = space->startNum2; j <= space->endNum2; j++) {
            lateral[SCHEMA] = j * space->increment;
            lateral[OBJECT] = j * space->increment;
            lateral[EFFECTOR] = j * space->increment;
            set_logfile(cs, space, self[SCHEMA], lateral[SCHEMA], k);
            run_one_trial(cs, k, i, j);
        }
    }
}

static void scan_int_ext_space(ParameterSpace *space, ContentionScheduler *cs, int k)
{
    int i, j;

    for (i = space->startNum1; i <= space->endNum1; i++) {
        parental[SCHEMA] = i * space->increment;
        for (j = space->startNum2; j <= space->endNum2; j++) {
            environmental[SCHEMA] = j * space->increment;
            set_logfile(cs, space, parental[SCHEMA], environmental[SCHEMA], k);
            run_one_trial(cs, k, i, j);
        }
    }
}

static void scan_schema_object_space(ParameterSpace *space, ContentionScheduler *cs, int k)
{
    int i, j;

    for (i = space->startNum1; i <= space->endNum1; i++) {
        environmental[SCHEMA] = i * space->increment;
        for (j = space->startNum2; j <= space->endNum2; j++) {
            parental[OBJECT] = j * space->increment;
            set_logfile(cs, space, environmental[SCHEMA], parental[OBJECT], k);
            run_one_trial(cs, k, i, j);
        }
    }
}

/*----------------------------------------------------------------------------*/

static void scan_parameter_space(ParameterSpace *space, ContentionScheduler *cs)
{
    int block;

    for (block = 0; block < space->maxTrials; block++) {
        fprintf(stderr, "******************** START BLOCK %d (of %d) ******************************\n", block+1, space->maxTrials);
        switch (space->vary) {
            case VARY_NOISE: { scan_noise_space(space, cs, block); break; }
            case VARY_IER: { scan_ier_space(space, cs, block); break; }
            case VARY_REST: { scan_rest_space(space, cs, block); break; }
            case VARY_SELF_LAT: { scan_self_lat_space(space, cs, block); break; }
            case VARY_INT_EXT: { scan_int_ext_space(space, cs, block); break; }
            case VARY_SCHEMA_OBJECT: { scan_schema_object_space(space, cs, block); break; }
            default: { fprintf(stderr, "Unhandled case in scan_parameter_space\n"); break; }
        }
        fprintf(stderr, "******************** END BLOCK %d (of %d) ********************************\n\n", block+1, space->maxTrials);
    }
}

/******************************************************************************/

static Boolean parse_command_line_arguments(int argc, char **argv, ParameterSpace *ps)
{
    /* Check for distractor command line switch: */

    if (argc < 1) {
        return(FALSE);
    }
    else if (argc == 1) {
        ps->vary = VARY_NONE;
        return(TRUE);
    }
    else if (argc != 9) {
        return(FALSE);
    }
    else {
        if (strcmp(argv[1], "-intext") == 0) {
            ps->vary = VARY_INT_EXT;
        }
        else if (strcmp(argv[1], "-selflat") == 0) {
            ps->vary = VARY_SELF_LAT;
        }
        else if (strcmp(argv[1], "-noise") == 0) {
            ps->vary = VARY_NOISE;
        }
        else if (strcmp(argv[1], "-ier") == 0) {
            ps->vary = VARY_IER;
        }
        else if (strcmp(argv[1], "-rest") == 0) {
            ps->vary = VARY_REST;
        }
        else if (strcmp(argv[1], "-schobj") == 0) {
            ps->vary = VARY_SCHEMA_OBJECT;
        }
        else {
            return(FALSE);
        }
    }

    /* Note: No type checking, so take care! */

    ps->outputDir    = argv[2];
    ps->increment    = atof(argv[3]);
    ps->incPrecision = strlen(strrchr(argv[3], '.')+1); 
    ps->maxTrials    = (int) atol(argv[4]);
    ps->startNum1    = (int) atol(argv[5]);
    ps->endNum1      = (int) atol(argv[6]);
    ps->startNum2    = (int) atol(argv[7]);
    ps->endNum2      = (int) atol(argv[8]);

    return(TRUE);
}

/******************************************************************************/
/* Main ***********************************************************************/

int main(int argc, char **argv)
{
    ParameterSpace ps;
    GtkWidget *frame;

    if (parse_command_line_arguments(argc, argv, &ps)) {

        gtk_set_locale();
        gtk_init(&argc, &argv);

        /* Create dynamic global variables: */

        scheduler = create_contention_scheduler();

        initialise_task(INITIAL_TASK, &(scheduler->task));

        /* We must initialise BEFORE making form, so as to get correct menus: */

        initialise_contention_scheduler(scheduler);
        reinitialise_contention_scheduler(scheduler, "tmp.log");
        read_default_parameters(NULL, scheduler);

        if (ps.vary != VARY_NONE) {
            scan_parameter_space(&ps, scheduler);
        }
        else if ((frame = make_widgets(scheduler)) != NULL)  {
            attach_data(scheduler);
            set_frame_title(INITIAL_TASK);
            update_cycle_display(scheduler);
            gtk_widget_show(frame);
            gtk_main();
        }

        destroy_contention_scheduler(scheduler);

        exit(0);
    }
    else {
        fprintf(stderr, "USAGE: %s \n", argv[0]);
        fprintf(stderr, "USAGE: %s { -intext | -selflat | -schobj | -noise | -ier | -rest} dir increment maxTrials startN1 endN1 startN2 endN2\n", argv[0]);
        fprintf(stderr, "Options:\n");
        fprintf(stderr, "   -intext:  Vary Schema Parental/Environmental\n");
        fprintf(stderr, "   -selflat: Vary Schema & Object Self/Lateral\n");
        fprintf(stderr, "   -schobj:  Vary Schema/Object excitation\n");
        fprintf(stderr, "   -noise:   Vary noise (ignore last 2 arguments)\n");
        fprintf(stderr, "   -ier:     Vary Schema Parent/Environment with fixed ratio (ignore last 2 arguments)\n");
        fprintf(stderr, "   -rest:    Vary rest activation (ignore last 2 arguments)\n");
        exit(1);
    }
}

/******************************************************************************/
