/*******************************************************************************

    FILE: 
    AUTHOR: R. Cooper (translated from C++ written by SVW)
    CREATION DATE: 
    CONTENTS: 

*******************************************************************************/

#include "cs.h"

#include <math.h>
#include <X11/Xlib.h>
#include <gtk/gtk.h>
#include "analyse.h"

#define canvasHeight 120
#define frameWidth   700

#define xMin 20
#define xMax -7
#define yMin 5
#define yMax -30

#define PARAMS_FILE    "task_parameters"

extern double feature_overlap(int af, int bf);

static void populate_confusability_matrix(ContentionScheduler *cs);
static void analysis_add_entry(ContentionScheduler *cs, char *stat, char *val);
static Boolean update_cl_schemas(GtkWidget *clist, GdkEventExpose *event, ContentionScheduler *cs);
static Boolean update_cl_objects(GtkWidget *clist, GdkEventExpose *event, ContentionScheduler *cs);
static Boolean update_cl_effectors(GtkWidget *clist, GdkEventExpose *event, ContentionScheduler *cs);

/******************************************************************************/
/* Local typedefs: ************************************************************/

typedef enum {TINY_SLIDER, SMALL_SLIDER, BIG_SLIDER, LOG_SLIDER, INTEGER_SLIDER, MATHS_SELECTOR, BOOLEAN_SELECTOR, TASK_SELECTOR} ParameterType;

typedef struct parameter_list {
    ParameterType type;
    void         *adj_widget;
    GtkWidget    *value_widget;
    void         *value;
    struct parameter_list *next;
} ParameterList;

/******************************************************************************/
/* Local variables: ***********************************************************/

/* gc[0,1,2] are black; gc[3,4,5] are blue; gc[6,7,8] are red; all increase   */
/* in sparsity L->R                                                           */
/* font_info[0] = 10 point (small); font_info[1] = 12 point (normal)          */

typedef struct x_widgets {
    ContentionScheduler *cs;
    GtkWidget     *frame;
    GtkWidget     *conf_matrix;
    GtkWidget     *cycle_widget;
    GtkWidget     *canvas[3];
    GdkGC         *gc[9];
    GdkFont       *font_info[2];
    GdkCursor     *busy;
    ParameterList *parameters;
    Boolean        evolve_graphs;
} XGlobals;

static XGlobals xg =   {NULL,
                        NULL,
                        NULL,
                        NULL,
                        {NULL, NULL, NULL},
                        {NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL},
                        {NULL, NULL},
                        NULL,
                        NULL,
                        FALSE};

/******************************************************************************/
/* Timer functions ************************************************************/

#include <sys/stat.h>
#include <sys/resource.h>

int usertime() /* Return total milliseconds of user time */
{
    struct rusage resource_usage;
    struct timeval *user_time = &(resource_usage.ru_utime);

    getrusage(RUSAGE_SELF, &resource_usage);
    return(1000 * user_time->tv_sec + ((int) user_time->tv_usec/1000));
}

int systime() /* Return total milliseconds of system time */
{
    struct rusage resource_usage;
    struct timeval *sys_time = &(resource_usage.ru_stime);

    getrusage(RUSAGE_SELF, &resource_usage);
    return(1000 * sys_time->tv_sec + ((int) sys_time->tv_usec/1000));
}

/******************************************************************************/

static void set_busy_cursor()
{
    gdk_window_set_cursor(xg.frame->window, xg.busy);
    while (gtk_events_pending()) {
        gtk_main_iteration();
    }
}

static void unset_cursor()
{
    gdk_window_set_cursor(xg.frame->window, NULL);
}

/******************************************************************************/
/* Frame, plotter, menu creation functions ************************************/

static void create_gcs(GtkWidget *widget)
{
    static signed char dashed[2] = {2,1};
    static signed char dotted[2] = {1,3};
    GdkColor colour;
    int i, j;

    for (i = 0; i < 3; i++) {
        for (j = 0; j < 3; j++) {
            xg.gc[(i*3)+j] = gdk_gc_new(widget->window);
        }

        colour.pixel = 0;
        colour.red = (i == 1)?65535:0;
        colour.green = 0;
        colour.blue = (i == 2)?65535:0;

        gdk_color_alloc(gdk_window_get_colormap(widget->window), &colour);

        for (j = 0; j < 3; j++) {
            gdk_gc_set_foreground(xg.gc[(i*3)+j], &colour);
        }

        gdk_gc_set_line_attributes(xg.gc[i*3], 0, GDK_LINE_SOLID, GDK_CAP_BUTT, GDK_JOIN_MITER);
        gdk_gc_set_dashes(xg.gc[(i*3)+1], 0, dashed, 2);
        gdk_gc_set_line_attributes(xg.gc[(i*3)+1], 0, GDK_LINE_ON_OFF_DASH, GDK_CAP_BUTT, GDK_JOIN_MITER);
        gdk_gc_set_dashes(xg.gc[(i*3)+2], 0, dotted, 2);
        gdk_gc_set_line_attributes(xg.gc[(i*3)+2], 0, GDK_LINE_ON_OFF_DASH, GDK_CAP_BUTT, GDK_JOIN_MITER);
    }

    if ((xg.font_info[0] = gdk_font_load("lucidasans-10")) == NULL) {
	xg.font_info[0] = gdk_font_load("fixed");
    }

    if ((xg.font_info[1] = gdk_font_load("lucidasans-12")) == NULL) {
	xg.font_info[1] = gdk_font_load("fixed");
    }
}

/******************************************************************************/

static int calculate_range(int count)
{
    return(count + 100);
#if FALSE
    if (count > 0) {
	/* If count < 10^(n+1), round up to nearest 10^n */
	/* svw: For some strange reason, hack required for range 100-999 */

	int n = (int) (log(count) / log(10.0));
	int scale = ((n == 2) ? (int) (pow(10.0, (double) n) + 1) : (int) pow(10.0, (double) n));

	return(scale * (1 + (count / scale)));
    }
    else {
        return(1);
    }
#endif
}

/*----------------------------------------------------------------------------*/

static int calculate_x_step(int range)
{
    return(10 * (int) (range / 100));
}

/*----------------------------------------------------------------------------*/

static void draw_axes(GtkWidget *graph, GdkPixmap *canvas, double x_scale, int x_range, double y_scale, double y_range)
{
    int height = graph->allocation.height;
    int width = graph->allocation.width;
    char buffer[16];
    int i, j;

    gdk_draw_line(canvas, xg.gc[0], xMin, height+yMax+1, xMin, yMin-1);
    gdk_draw_line(canvas, xg.gc[0], xMin-1, height+yMax, width+xMax+1, height+yMax);

    /* Draw x-axis labels and vertical grid lines */

    i = calculate_x_step(x_range);

    for (j = 0; j <= x_range; j += i) {
        short w, x;

	g_snprintf(buffer, 16, "%d", j);
	w = gdk_text_width(xg.font_info[0], buffer, strlen(buffer));
	x = ((short) (j * x_scale) + xMin);
	gdk_draw_text(canvas, xg.font_info[0], xg.gc[0], x - (w/2), height+yMax+12, buffer, strlen(buffer));
	gdk_draw_line(canvas, xg.gc[2], x, yMin-2, x, height+yMax+1);
    }

    /* Draw y-axis labels and horizontal grid lines */

    for (j = 0; j <= 5; j++) {
	short w, y;

	g_snprintf(buffer, 16, "%3.1f", j * 0.2);
	w = gdk_text_width(xg.font_info[0], buffer, strlen(buffer));
	y = (height - (short) (j * 0.2 * y_scale) + yMax);
	gdk_draw_text(canvas, xg.font_info[0], xg.gc[0], xMin-5-w, y+4, buffer, strlen(buffer));
	gdk_draw_line(canvas, xg.gc[2], xMin-2, y, width+xMax+1, y);
    }
}

/******************************************************************************/

static void reset_parameter_widgets(ParameterList *parameters); /* Forward declaration */

void read_default_parameters(GtkWidget *button, ContentionScheduler *cs)
{
    FILE *fp = fopen(PARAMS_FILE, "r");
    
    if (fp != NULL) {
        int i, j;

        message(WARNING, cs, "Reading default parameters");

        fscanf(fp, "%lf %lf %lf %lf", &(parental[SCHEMA]), &(environmental[SCHEMA]), &(self[SCHEMA]), &(lateral[SCHEMA]));
        fscanf(fp, "%lf %lf %lf %lf", &(parental[OBJECT]), &(environmental[OBJECT]), &(self[OBJECT]), &(lateral[OBJECT]));
        fscanf(fp, "%lf %lf %lf %lf", &(parental[EFFECTOR]), &(environmental[EFFECTOR]), &(self[EFFECTOR]), &(lateral[EFFECTOR]));
        fscanf(fp, "%lf %lf %lf %lf %lf %lf", &restAct, &selectionThreshold, &persistence, &noiseSD[SCHEMA], &noiseSD[OBJECT], &noiseSD[EFFECTOR]);
        fscanf(fp, "%d", &i);
        maths = (MathsType) i;
        for (j = 0; j < MaxMessageTypes; j++) {
            fscanf(fp, " %d", &i);
            messages[j] = ((i == 0) ? NULL : stderr);
        }
	fclose(fp);

        reset_parameter_widgets(xg.parameters);
    }
    else {
        char buffer[80];

	g_snprintf(buffer, 80, "Cannot open parameters file (%s): using built-in defaults", PARAMS_FILE);
        message(WARNING, cs, buffer);
    }
}

static void write_default_parameters(GtkWidget *button, ContentionScheduler *cs)
{
    FILE *fp = fopen(PARAMS_FILE, "w");
    
    if (fp != NULL) {
        int j;
        message(WARNING, cs, "Writing default parameters");

        fprintf(fp, "%5.3f %5.3f %5.3f %5.3f\n", parental[SCHEMA], environmental[SCHEMA], self[SCHEMA], lateral[SCHEMA]);
        fprintf(fp, "%5.3f %5.3f %5.3f %5.3f\n", parental[OBJECT], environmental[OBJECT], self[OBJECT], lateral[OBJECT]);
        fprintf(fp, "%5.3f %5.3f %5.3f %5.3f\n", parental[EFFECTOR], environmental[EFFECTOR], self[EFFECTOR], lateral[EFFECTOR]);
        fprintf(fp, "%5.3f %5.3f %5.3f %7.5f %7.5f %7.5f\n", restAct, selectionThreshold, persistence, noiseSD[SCHEMA], noiseSD[OBJECT], noiseSD[EFFECTOR]);
        fprintf(fp, "%1d", (int) maths);
        for (j = 0; j < MaxMessageTypes; j++) {
            fprintf(fp, " %1d", ((messages[j] == 0) ? 0 : 1));
        }
        fprintf(fp, "\n");
        fclose(fp);
    }
    else {
        char buffer[80];

	g_snprintf(buffer, 80, "Cannot open parameters file (%s): write aborted", PARAMS_FILE);
        message(WARNING, cs, buffer);
    }
}

/******************************************************************************/

static char *print_int(char *buffer, int l, int arg)
{
    g_snprintf(buffer, l, " %d", arg);
    return(buffer);
}

static char *print_dbl(char *buffer, int l, double arg)
{
    g_snprintf(buffer, l, " %f", arg);
    return(buffer);
}

static char *print_bool(char *buffer, int l, Boolean arg)
{
    g_snprintf(buffer, l, " %s", (arg ? "true" : "false"));
    return(buffer);
}

static int split_error_string(char *line)
{
    int j = 0;
    while (line[j] != '\0') {
        if (line[j] == ':') {
            line[j+1] = '\0';
            return(j+2);
        }
        else {
            j++;
        }
    }
    return(0);
}

static void do_analysis(ContentionScheduler *cs)
{
    extern LineList *free_line_list(LineList *ll);

    DataAnalysis la;
    char buffer[10];
    LineList *tmp;

    la.actionSequence = NULL;
    la.log = NULL;

    if (cs->logfile == NULL) {
        fprintf(stderr, "Analysis aborted: logfile is NULL\n");
    }
    else if (analyse_results(cs->logfile, &(cs->task), &la)) {
        analysis_add_entry(cs, "  Accomplished:          ", print_int(buffer, 10, la.numAccomplished));
        analysis_add_entry(cs, "  Actions:               ", print_int(buffer, 10, la.numActions));
        analysis_add_entry(cs, "  Interventions:         ", print_int(buffer, 10, la.numInterventions));
        analysis_add_entry(cs, "  Total Errors:          ", print_int(buffer, 10, la.numErrors));
        analysis_add_entry(cs, "  Sequence errors:       ", " ");
        analysis_add_entry(cs, "    Action Additions:      ", print_int(buffer, 10, la.numActionAdditions));
        analysis_add_entry(cs, "    Action Anticipations:  ", print_int(buffer, 10, la.numActionAnticipations));
        analysis_add_entry(cs, "    Step Omissions:        ", print_int(buffer, 10, la.numStepOmissions));
        analysis_add_entry(cs, "    Perseverations:        ", print_int(buffer, 10, la.numPerseverations));
        analysis_add_entry(cs, "  Conceptual errors:       ", " ");
        analysis_add_entry(cs, "    Misuse type 1:         ", print_int(buffer, 10, la.numMisUse1));
        analysis_add_entry(cs, "    Misuse type 2:         ", print_int(buffer, 10, la.numMisUse2));
        analysis_add_entry(cs, "    Mislocation type 1:    ", print_int(buffer, 10, la.numMisLocation1));
        analysis_add_entry(cs, "    Mislocation type 2:    ", print_int(buffer, 10, la.numMisLocation2));
        analysis_add_entry(cs, "    Object Substitution:   ", print_int(buffer, 10, la.numObjectSubstitution));
        analysis_add_entry(cs, "    Tool Omissions:        ", print_int(buffer, 10, la.numToolOmissions));
        analysis_add_entry(cs, "    Toying:                ", print_int(buffer, 10, la.numToying));
        analysis_add_entry(cs, "    Pantomime:             ", print_int(buffer, 10, la.numPantomime));
        analysis_add_entry(cs, "    Perplexity:            ", print_int(buffer, 10, la.numPerplexity));
        analysis_add_entry(cs, "  Comission Proportion:  ", print_dbl(buffer, 10, la.propCommission));
        analysis_add_entry(cs, "  Cycles:                ", print_int(buffer, 10, la.numCycles));
        analysis_add_entry(cs, "  Complete:              ", print_bool(buffer, 10, la.complete));
    }

    if (la.log != NULL) {
         analysis_add_entry(cs, " ", " ");
        analysis_add_entry(cs, "List of Errors:", " ");
        for (tmp = la.log; tmp != NULL; tmp = tmp->next) {
            int j = split_error_string(tmp->line);
            analysis_add_entry(cs, tmp->line, &(tmp->line[j]));
        }
        free_line_list(la.log);
    }
}

static void step_callback(GtkWidget *button, ContentionScheduler *cs)
{
    set_busy_cursor();
    step(cs, (int) gtk_object_get_data(GTK_OBJECT(button), "steps"), xg.evolve_graphs);
    if (!xg.evolve_graphs) {
        attach_data(cs);
        update_cycle_display(cs);
    }
    unset_cursor();
}

/*----------------------------------------------------------------------------*/

static void dump_callback(GtkWidget *button, ContentionScheduler *cs)
{
    FILE *fp = fopen("tmp.agr", "w");
    SchemaList *ss;
    int i, i0 = MAX(0, cs->cycle - 200);

    fprintf(fp, "@with g0\n");
    fprintf(fp, "@g0 on\n");
    for (ss = cs->schemas, i = 0; ss; ss = ss->tail, i++)    {
        fprintf(fp, "@    legend string %d \"%s\"\n", i, ss->head->name);
    }
    fprintf(fp, "@with g0\n");
    fprintf(fp, "@g0 on\n");
    for (ss = cs->schemas; ss; ss = ss->tail)    {
        for (i = i0; i < cs->cycle; i++) {
            fprintf(fp, "%d %5.3f\n", i, ss->head->activation[i]);
        }
        fprintf(fp, "&\n");
    }
    fclose(fp);
}

static void run_callback(GtkWidget *button, ContentionScheduler *cs)
{
    int t0, u0, t1, u1, c0;

    t0 = systime();
    u0 = usertime();
    c0 = cs->cycle;

    set_busy_cursor();
    run(cs, xg.evolve_graphs);
    if (!xg.evolve_graphs) {
        attach_data(cs);
        update_cycle_display(cs);
    }

    t1 = systime();
    u1 = usertime();
    c0 = cs->cycle - c0;
    printf("Run timing: %d cycles: %d | %d ms\n", c0, t1 - t0, u1 - u0);
    printf("Run timing: %f | %f ms per cycle\n", (t1 - t0) / (double) c0, (u1 - u0) / (double) c0);
    do_analysis(cs);
    unset_cursor();
}

/*----------------------------------------------------------------------------*/

static void clear_logs(ContentionScheduler *cs)
{
    if (cs->cl_actions != NULL) {
        gtk_clist_clear(GTK_CLIST(cs->cl_actions));
    }
    if (cs->cl_analysis != NULL) {
        gtk_clist_clear(GTK_CLIST(cs->cl_analysis));
    }
}

static void reinitialise_callback(GtkWidget *button, ContentionScheduler *cs)
{
    reinitialise_contention_scheduler(cs, "tmp.log");
    clear_logs(cs);
    attach_data(cs);
    update_cycle_display(cs);

    update_cl_schemas(NULL, NULL, cs);
    update_cl_objects(NULL, NULL, cs);
    update_cl_effectors(NULL, NULL, cs);
}

/******************************************************************************/

static void slider_callback(GtkAdjustment *adjustment, ParameterList *p)
{
    if (p == NULL) {
	fprintf(stderr, "INTERNAL ERROR: NULL parameter in slider_callback\n");
    }
    else {
	double value = (double) adjustment->value;
	char buffer[10];

	switch (p->type) {
	    case TINY_SLIDER: {
		double real_value = (double) value / 1000.0;
		g_snprintf(buffer, 10, "(%5.3f)", real_value);
		*((double *)p->value) = real_value;
		break;
	    }

	    case SMALL_SLIDER: {
		double real_value = (double) value / 100.0;
		g_snprintf(buffer, 10, "(%4.2f)", real_value);
		*((double *)p->value) = real_value;
		break;
	    }

	    case BIG_SLIDER: {
		double real_value = (double) value / 10.0;
		g_snprintf(buffer, 10, "(%4.2f)", real_value);
		*((double *)p->value) = real_value;
		break;
	    }

	    case LOG_SLIDER: {
		double real_value = pow(10.0, (((double) value - 100.0) / 10.0));
		g_snprintf(buffer, 10, "(%3.2e)", real_value);
		*((double *)p->value) = real_value;
		break;
	    }

	    case INTEGER_SLIDER: {
		int real_value = value * 5.0;
		g_snprintf(buffer, 10, "(%d)", real_value);
		*((int *)p->value) = real_value;
		break;
	    }

	    default: {
		return;
	    }
	}
	gtk_label_set_text(GTK_LABEL(p->value_widget), buffer);
    }
}

/*----------------------------------------------------------------------------*/

static void maths_callback(GtkWidget *widget, ParameterList *p)
{
    GtkWidget *mi = gtk_menu_get_active(p->adj_widget);

    *(MathsType *)p->value = (MathsType) gtk_object_get_data(GTK_OBJECT(mi), "data");
}

/*----------------------------------------------------------------------------*/

void set_frame_title(TaskTypeId tid)
{
    char buffer[128];

    g_snprintf(buffer, 128, "Contention Scheduling: %s", task_names[tid]);
    gtk_window_set_title(GTK_WINDOW(xg.frame), buffer);
}

static void set_task(TaskTypeId tid)
{
    initialise_task(tid, &(xg.cs->task));
    reinitialise_callback(NULL, xg.cs);
    populate_confusability_matrix(xg.cs);
    set_frame_title(tid);
}

static void task_callback(GtkWidget *widget, ParameterList *p)
{
    GtkWidget *mi = gtk_menu_get_active(p->adj_widget);

    set_task((TaskTypeId) gtk_object_get_data(GTK_OBJECT(mi), "data"));
}

/*----------------------------------------------------------------------------*/

static void boolean_callback(GtkWidget *widget, ParameterList *p)
{
    if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(p->adj_widget))) {
        *(FILE **)p->value = stderr;
    }
    else {
        *(FILE **)p->value = NULL;
    }
}

/*----------------------------------------------------------------------------*/

static void create_param_popup_widget(GtkWidget *box, char *label, void *param, ParameterType type)
{
    char *maths_name[NORMALISED+1] = {"Sigmoid", "Tanh", "Half X", "M/R", "Houghton", "Normalised"};
    ParameterList *p = (ParameterList *)malloc(sizeof(ParameterList));
    GtkWidget *hbox, *tmp;

    if (p != NULL) {
	p->value = param;
	p->type = type;

        hbox = gtk_hbox_new(FALSE, 4);
        gtk_box_pack_start(GTK_BOX(box), hbox, FALSE, FALSE, 4);
        gtk_widget_show(hbox);

        tmp = gtk_label_new(label);
        gtk_box_pack_start(GTK_BOX(hbox), tmp, FALSE, FALSE, 0);
        gtk_widget_show(tmp);

	if (type == MATHS_SELECTOR) {
	    GtkWidget *mi;
            int i;

            p->value_widget = gtk_option_menu_new();
            p->adj_widget = gtk_menu_new();

            for (i = SIGMOID; i <= NORMALISED; i++) {
                mi = gtk_menu_item_new_with_label(maths_name[i]);
                gtk_menu_append(GTK_MENU(p->adj_widget), mi);
                gtk_signal_connect(GTK_OBJECT(mi), "activate", GTK_SIGNAL_FUNC(maths_callback), p);
                gtk_object_set_data(GTK_OBJECT(mi), "data", (void *)i);
                gtk_widget_show(mi);
            }
            gtk_option_menu_set_menu(GTK_OPTION_MENU(p->value_widget), p->adj_widget);
            gtk_widget_show(p->value_widget);
            gtk_box_pack_start(GTK_BOX(hbox), GTK_WIDGET(p->value_widget), FALSE, FALSE, 0);
	}
	else if (type == TASK_SELECTOR) {
	    GtkWidget *mi;
            int i;

            p->value_widget = gtk_option_menu_new();
            p->adj_widget = gtk_menu_new();

            for (i = 0; i < TT_MAX; i++) {
                mi = gtk_menu_item_new_with_label(task_names[i]);
                gtk_menu_append(GTK_MENU(p->adj_widget), mi);
                gtk_signal_connect(GTK_OBJECT(mi), "activate", GTK_SIGNAL_FUNC(task_callback), p);
                gtk_object_set_data(GTK_OBJECT(mi), "data", (void *)i);
                gtk_widget_show(mi);
            }
            gtk_option_menu_set_menu(GTK_OPTION_MENU(p->value_widget), p->adj_widget);
            gtk_widget_show(p->value_widget);
            gtk_box_pack_start(GTK_BOX(hbox), GTK_WIDGET(p->value_widget), FALSE, FALSE, 0);
	}
	else if (type == BOOLEAN_SELECTOR) {
	    p->adj_widget = (void *)gtk_check_button_new();
            gtk_signal_connect(GTK_OBJECT(p->adj_widget), "toggled", GTK_SIGNAL_FUNC(boolean_callback), p);
	    gtk_widget_set_usize(p->adj_widget, 200, 0);
            gtk_box_pack_end(GTK_BOX(hbox), GTK_WIDGET(p->adj_widget), FALSE, FALSE, 0);
            gtk_widget_show(GTK_WIDGET(p->adj_widget));
	}
	else { /* Some kind of slider widget ... */
            p->adj_widget = (void *)gtk_adjustment_new(0.0, 0.0, 100.0, 1.0, 1.0, 0.0);
	    p->value_widget = gtk_label_new(" ");
	    gtk_widget_set_usize(p->value_widget, 50, 0);
            gtk_box_pack_end(GTK_BOX(hbox), p->value_widget, FALSE, FALSE, 0);
	    gtk_widget_show(p->value_widget);

	    tmp = gtk_hscale_new(GTK_ADJUSTMENT(p->adj_widget));
            gtk_widget_set_usize(tmp, 190, 5);
            gtk_scale_set_draw_value(GTK_SCALE(tmp), FALSE);
            gtk_box_pack_end(GTK_BOX(hbox), tmp, FALSE, FALSE, 0);
	    gtk_widget_show(tmp);

            gtk_signal_connect(GTK_OBJECT(p->adj_widget), "value_changed", GTK_SIGNAL_FUNC(slider_callback), p);
	}
	p->next = xg.parameters;
	xg.parameters = p;
    }
    else {
	fprintf(stderr, "MEMORY ALLOCATION FAILURE: Parameter allocation failed: Extra parameters ignored\n");
    }
}

/*----------------------------------------------------------------------------*/

static GtkWidget *create_button(GtkWidget *box, char *label, void callback(), void *data)
{
    GtkWidget *button = gtk_button_new_with_label(label);
    gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(callback), data);
    gtk_box_pack_start(GTK_BOX(box), button, TRUE, TRUE, 0);
    gtk_widget_show(button);
    return(button);
}

static void exit_xvcs(GtkWidget *button, void *data)
{
    gtk_exit(0);
}

static void quit_callback(GtkWidget *button, GtkWidget *frame)
{
    gtk_widget_destroy(frame);
}

/*----------------------------------------------------------------------------*/

static void reset_parameter_widgets(ParameterList *parameters)
{
    ParameterList *p;
    char buffer[10];
    double d;

    for (p = parameters; p; p = p->next) {
	switch (p->type) {
	    case TINY_SLIDER:  {
		d = 1000 * *((double *)p->value);
		gtk_adjustment_set_value(GTK_ADJUSTMENT(p->adj_widget), d);
		g_snprintf(buffer, 10, "(%5.3f)", *((double *)p->value));
		gtk_label_set_text(GTK_LABEL(p->value_widget), buffer);
		break;
	    }

	    case SMALL_SLIDER:  {
		d = 100 * *((double *)p->value);
		gtk_adjustment_set_value(GTK_ADJUSTMENT(p->adj_widget), d);
		g_snprintf(buffer, 10, "(%4.2f)", *((double *)p->value));
		gtk_label_set_text(GTK_LABEL(p->value_widget), buffer);
		break;
	    }

	    case BIG_SLIDER:  {
		d = 10 * *((double *)p->value);
		gtk_adjustment_set_value(GTK_ADJUSTMENT(p->adj_widget), d);
		g_snprintf(buffer, 10, "(%4.2f)", *((double *)p->value));
		gtk_label_set_text(GTK_LABEL(p->value_widget), buffer);
		break;
	    }

	    case LOG_SLIDER:  {

		d = 100.0 + (10.0 * log10(*((double *)p->value)));
		gtk_adjustment_set_value(GTK_ADJUSTMENT(p->adj_widget), d);
		g_snprintf(buffer, 10, "(%3.2e)", *((double *)p->value));

		gtk_label_set_text(GTK_LABEL(p->value_widget), buffer);
		break;
	    }

	    case INTEGER_SLIDER:  {
                d = *((int *)p->value) / 5;
		gtk_adjustment_set_value(GTK_ADJUSTMENT(p->adj_widget), d);
		g_snprintf(buffer, 10, "(%3d)", *((int *)p->value));
		gtk_label_set_text(GTK_LABEL(p->value_widget), buffer);
		break;
	    }

	    case MATHS_SELECTOR:  {
                gtk_option_menu_set_history(GTK_OPTION_MENU(p->value_widget), *((int *)p->value));
		break;
	    }

	    case TASK_SELECTOR:  {
                gtk_option_menu_set_history(GTK_OPTION_MENU(p->value_widget), *((int *)p->value));
		break;
	    }

	    case BOOLEAN_SELECTOR:  {
                gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(p->adj_widget), ((*(FILE **)p->value) != NULL));
		break;
	    }
	}
    }
}

/*----------------------------------------------------------------------------*/

static void make_params_widgets(ContentionScheduler *cs, GtkWidget *hbox)
{
    GtkWidget *vbox, *bbox, *tmp;

    vbox = gtk_vbox_new(TRUE, 0);
    gtk_container_add(GTK_CONTAINER(hbox), vbox);
    gtk_widget_show(vbox);

    bbox = gtk_hbox_new(TRUE, 200);
    gtk_box_pack_start(GTK_BOX(vbox), bbox, FALSE, FALSE, 0);
    gtk_widget_show(bbox);

    create_button(bbox, "Default", read_default_parameters, cs);
    create_button(bbox, "Save", write_default_parameters, cs);

    create_param_popup_widget(vbox, "Task:", &(cs->task), TASK_SELECTOR);

    tmp = gtk_hseparator_new();
    gtk_box_pack_start(GTK_BOX(vbox), tmp, FALSE, FALSE, 4);
    gtk_widget_show(tmp);

    create_param_popup_widget(vbox, "Rest Activation: ", &restAct, SMALL_SLIDER);
    create_param_popup_widget(vbox, "Select Thresh.:  ", &selectionThreshold, SMALL_SLIDER);
    create_param_popup_widget(vbox, "Persistence:     ", &persistence, SMALL_SLIDER);
    create_param_popup_widget(vbox, "Goal Precision:  ", &goalPrecision, BIG_SLIDER);
    create_param_popup_widget(vbox, "Goal Cutoff:     ", &goalCutoff, SMALL_SLIDER);
    create_param_popup_widget(vbox, "Maths System:    ", &maths, MATHS_SELECTOR);
    create_param_popup_widget(vbox, "Interventions:   ", &interventions, BOOLEAN_SELECTOR);

    tmp = gtk_hseparator_new();
    gtk_box_pack_start(GTK_BOX(vbox), tmp, FALSE, FALSE, 4);
    gtk_widget_show(tmp);

    create_param_popup_widget(vbox, "Show Warnings:       ", &(messages[WARNING]), BOOLEAN_SELECTOR);
    create_param_popup_widget(vbox, "Show Selections:     ", &(messages[SELECTION]), BOOLEAN_SELECTOR);
    create_param_popup_widget(vbox, "Show Bind Messages:  ", &(messages[BIND]), BOOLEAN_SELECTOR);
    create_param_popup_widget(vbox, "Show Run Messages:   ", &(messages[RUN]), BOOLEAN_SELECTOR);
    create_param_popup_widget(vbox, "Show World:          ", &(messages[WORLD]), BOOLEAN_SELECTOR);
    create_param_popup_widget(vbox, "Show SAS Monitoring: ", &(messages[MONITOR]), BOOLEAN_SELECTOR);
    create_param_popup_widget(vbox, "Evolve Graph:        ", &(xg.evolve_graphs), BOOLEAN_SELECTOR);

    tmp = gtk_vseparator_new();
    gtk_box_pack_start(GTK_BOX(hbox), tmp, FALSE, FALSE, 4);
    gtk_widget_show(tmp);

    vbox = gtk_vbox_new(TRUE, 0);
    gtk_container_add(GTK_CONTAINER(hbox), vbox);
    gtk_widget_show(vbox);

    /* Network activation parameters: */

    create_param_popup_widget(vbox, "Schema Int:   ", &(parental[SCHEMA]), SMALL_SLIDER);
    create_param_popup_widget(vbox, "Schema Ext:   ", &(environmental[SCHEMA]), SMALL_SLIDER);
    create_param_popup_widget(vbox, "Schema S:     ", &(self[SCHEMA]), SMALL_SLIDER);
    create_param_popup_widget(vbox, "Schema L:     ", &(lateral[SCHEMA]), SMALL_SLIDER);
    create_param_popup_widget(vbox, "Schema Noise: ", &noiseSD[SCHEMA], TINY_SLIDER);

    create_param_popup_widget(vbox, "Object Int:   ", &(environmental[OBJECT]), SMALL_SLIDER);
    create_param_popup_widget(vbox, "Object Ext:  ", &(parental[OBJECT]), SMALL_SLIDER);
    create_param_popup_widget(vbox, "Object S:     ", &(self[OBJECT]), SMALL_SLIDER);
    create_param_popup_widget(vbox, "Object L:     ", &(lateral[OBJECT]), SMALL_SLIDER);
    create_param_popup_widget(vbox, "Object Noise: ", &noiseSD[OBJECT], SMALL_SLIDER);
//    create_param_popup_widget(vbox, "Object Noise: ", &noiseSD[OBJECT], TINY_SLIDER);

    create_param_popup_widget(vbox, "Effector Int: ", &(environmental[EFFECTOR]), SMALL_SLIDER);
    create_param_popup_widget(vbox, "Effector Ext: ", &(parental[EFFECTOR]), SMALL_SLIDER);
    create_param_popup_widget(vbox, "Effector S:   ", &(self[EFFECTOR]), SMALL_SLIDER);
    create_param_popup_widget(vbox, "Effector L:   ", &(lateral[EFFECTOR]), SMALL_SLIDER);
    create_param_popup_widget(vbox, "Effector Noise: ", &noiseSD[EFFECTOR], TINY_SLIDER);

    reset_parameter_widgets(xg.parameters);
}

/******************************************************************************/

static void draw_plot_line(GdkPixmap *canvas, double *values, Boolean highlighted, int time0, int count, double x_scale, double y_scale, int height)
{
    GdkGC *current_gc = (highlighted?xg.gc[3]:xg.gc[6]);
    short x0, y0, x1, y1;

    if (time0 < count) {
        x0 = (short) (time0 * x_scale) + xMin;
        y0 = (short) (height - (values[time0] * y_scale)) + yMax;

        while (time0++ < count) {
            x1 = (short) (time0 * x_scale) + xMin;
            y1 = (short) (height - values[time0] * y_scale) + yMax;
    	    gdk_draw_line(canvas, current_gc, x0, y0, x1, y1);
            x0 = x1;
            y0 = y1;
        }
    }
}

/*----------------------------------------------------------------------------*/

static Boolean expose_graph(GtkWidget *widget, GdkEventExpose *event, void *data)
{
    GdkPixmap *canvas = gtk_object_get_data(GTK_OBJECT(widget), "canvas");

    if (event == NULL) {
        gdk_draw_pixmap(widget->window,
                    widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
                    canvas,
                    0, 0, 0, 0,
                    widget->allocation.width, widget->allocation.height);
    }
    else {
        gdk_draw_pixmap(widget->window,
                    widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
                    canvas,
                    event->area.x, event->area.y,
                    event->area.x, event->area.y,
                    event->area.width, event->area.height);

    }
    return(FALSE);
}

static Boolean locate_coordinates(GtkWidget *widget, GdkEventButton *event, ContentionScheduler *cs, int *cycle, double *act)
{
    if (event->button == 1) {
        double denominator, x_scale, y_scale;
        int range, width, height;

        width = widget->allocation.width;
        height = widget->allocation.height;
        range = calculate_range(cs->cycle);
        denominator = maxAct - minAct;
        x_scale	= (width - (xMin - xMax)) / (double) range;
        y_scale	= (height - (yMin - yMax)) / denominator;

        *cycle = (int) ((event->x - xMin) / x_scale);
        *act = 1.0 - (event->y - yMin) / y_scale;
        return((*cycle < cs->cycle) && (*act >= 0.0) && (*act <= 1.0));
    }
    else {
        return(FALSE);
    }
}

static Boolean click_on_schema_graph(GtkWidget *widget, GdkEventButton *event, ContentionScheduler *cs)
{
    int cycle;
    double act;

    if (locate_coordinates(widget, event, cs, &cycle, &act)) {
        double delta = 5.0 / (double) (widget->allocation.height - (yMin - yMax));
        SchemaList *ss;
        for (ss = cs->schemas; ss; ss = ss->tail)    {
            if (fabs(ss->head->activation[cycle] - act) < delta) {
                fprintf(stderr, "%s: %f (at cycle %d)\n", ss->head->name, ss->head->activation[cycle], cycle);
            }
        }
    }
    return(FALSE);
}

static Boolean click_on_object_graph(GtkWidget *widget, GdkEventButton *event, ContentionScheduler *cs)
{
    int cycle;
    double act;

    if (locate_coordinates(widget, event, cs, &cycle, &act)) {
        double delta = 5.0 / (double) (widget->allocation.height - (yMin - yMax));
        ObjectList *os;
        int role = (int) gtk_object_get_data(GTK_OBJECT(widget), "role");
        for (os = cs->objects; os; os = os->tail)    {
            if (fabs(os->head->activation[role][cycle] - act) < delta) {
                fprintf(stderr, "%s: %f (at cycle %d)\n", os->head->name, os->head->activation[role][cycle], cycle);
            }
        }
    }
    return(FALSE);
}

static Boolean click_on_effector_graph(GtkWidget *widget, GdkEventButton *event, ContentionScheduler *cs)
{
    int cycle;
    double act;

    if (locate_coordinates(widget, event, cs, &cycle, &act)) {
        double delta = 5.0 / (double) (widget->allocation.height - (yMin - yMax));
        EffectorList *es;
        for (es = cs->effectors; es; es = es->tail)    {
            if (fabs(es->head->activation[cycle] - act) < delta) {
                fprintf(stderr, "%s: %f (at cycle %d)\n", es->head->name, es->head->activation[cycle], cycle);
            }
        }
    }
    return(FALSE);
}

/*----------------------------------------------------------------------------*/

static GdkPixmap *get_blank_graph(GtkWidget *widget)
{
    GdkPixmap *canvas = gtk_object_get_data(GTK_OBJECT(widget), "canvas");

    if (canvas != NULL) {
        gdk_pixmap_unref(canvas);
    }
    if (widget->window != NULL) {
        canvas = gdk_pixmap_new(widget->window, widget->allocation.width, widget->allocation.height, -1);
    }
    else {
        canvas = NULL;
    }
    if (canvas != NULL) {
        gdk_draw_rectangle(canvas, widget->style->white_gc, TRUE, 0, 0, widget->allocation.width, widget->allocation.height);
    }
    gtk_object_set_data(GTK_OBJECT(widget), "canvas", canvas);
    return(canvas);
}

/*----------------------------------------------------------------------------*/

static Boolean repaint_schema_plotter(GtkWidget *widget, GdkEventConfigure *event, ContentionScheduler *cs)
{
    GdkPixmap *canvas = get_blank_graph(widget);
    char *label = "Schema Activations";
    double denominator, x_scale, y_scale;
    SchemaList *ss;
    int range, width, height;

    if (canvas != NULL) {
        width = widget->allocation.width;
        height = widget->allocation.height;
        range = calculate_range(cs->cycle);
        denominator = maxAct - minAct;
        x_scale	= (width - (xMin - xMax)) / (double) range;
        y_scale	= (height - (yMin - yMax)) / denominator;

        draw_axes(widget, canvas, x_scale, range, y_scale, denominator);
        gdk_draw_text(canvas, xg.font_info[1], xg.gc[0], 5, height-5, label, strlen(label));

        for (ss = cs->schemas; ss; ss = ss->tail)    {
            Boolean highlight = ss->head->highlighted;
            double *act = ss->head->activation;

            draw_plot_line(canvas, act, highlight, 0, cs->cycle, x_scale, y_scale, height);
        }
        expose_graph(widget, NULL, NULL);
    }
    return(FALSE);
}

static Boolean repaint_object_plotter(GtkWidget *widget, GdkEventConfigure *event, ContentionScheduler *cs)
{
    GdkPixmap *canvas = get_blank_graph(widget);
    char *label = "Object Representation Activations";
    double denominator, x_scale, y_scale;
    int range, width, height, role;
    ObjectList *os;

    if (canvas != NULL) {
        width = widget->allocation.width;
        height = widget->allocation.height;
        range	= calculate_range(cs->cycle);
        denominator = maxAct - minAct;
        x_scale	= (width - (xMin - xMax)) / (double) range;
        y_scale	= (height - (yMin - yMax)) / denominator;

        draw_axes(widget, canvas, x_scale, range, y_scale, denominator);
        gdk_draw_text(canvas, xg.font_info[1], xg.gc[0], 5, height-5, label, strlen(label));

        role = (int) gtk_object_get_data(GTK_OBJECT(widget), "role");
        for (os = cs->objects; os; os = os->tail)    {
            Boolean highlight = os->head->highlighted;
            double *act = os->head->activation[role];
    
            draw_plot_line(canvas, act, highlight, os->head->time0, cs->cycle, x_scale, y_scale, height);
        }
        expose_graph(widget, NULL, NULL);
    }
    return(FALSE);
}

static Boolean repaint_effector_plotter(GtkWidget *widget, GdkEventConfigure *event, ContentionScheduler *cs)
{
    GdkPixmap *canvas = get_blank_graph(widget);
    double denominator, x_scale, y_scale;
    char *label = "Effector Activations";
    int range, width, height;
    EffectorList *es;

    if (canvas != NULL) {
        width = widget->allocation.width;
        height = widget->allocation.height;
        range	= calculate_range(cs->cycle);
        denominator = maxAct - minAct;
        x_scale	= (width - (xMin - xMax)) / (double) range;
        y_scale	= (height - (yMin - yMax)) / denominator;

        draw_axes(widget, canvas, x_scale, range, y_scale, denominator);
        gdk_draw_text(canvas, xg.font_info[1], xg.gc[0], 5, height-5, label, strlen(label));

        for (es = cs->effectors; es; es = es->tail)    {
            Boolean highlight = es->head->highlighted;
            double *act = es->head->activation;

            draw_plot_line(canvas, act, highlight, 0, cs->cycle, x_scale, y_scale, height);
        }
        expose_graph(widget, NULL, NULL);
    }
    return(FALSE);
}

static GtkWidget *create_graph(GtkWidget *vbox, int height, Boolean repaint_proc(), char *label, ContentionScheduler *cs)
{
    GtkWidget *graph = gtk_drawing_area_new();
//    gtk_widget_set_usize(graph, 0, height);
    gtk_signal_connect(GTK_OBJECT(graph), "expose_event", GTK_SIGNAL_FUNC(expose_graph), 0);
    gtk_signal_connect(GTK_OBJECT(graph), "configure_event", GTK_SIGNAL_FUNC(repaint_proc), cs);
    gtk_widget_set_events(graph, GDK_EXPOSURE_MASK | GDK_BUTTON_PRESS_MASK);
    gtk_box_pack_start(GTK_BOX(vbox), graph, TRUE, TRUE, 0);
    gtk_widget_show(graph);
    return(graph);
}

/******************************************************************************/
/* Highlight callback functions ***********************************************/

static void highlight_schema_none(GtkWidget *button, ContentionScheduler *cs)
{
    SchemaList *tmp;

    for (tmp = cs->schemas; tmp != NULL; tmp = tmp->tail) {
	tmp->head->highlighted = FALSE;
    }
    repaint_schema_plotter(xg.canvas[SCHEMA], NULL, cs);
}

static void highlight_schema(GtkWidget *button, ContentionScheduler *cs)
{
    Schema *tmp = gtk_object_get_data(GTK_OBJECT(button), "schema");

    tmp->highlighted = !(tmp->highlighted);
    repaint_schema_plotter(xg.canvas[SCHEMA], NULL, cs);
}

static void highlight_schema_all(GtkWidget *button, ContentionScheduler *cs)
{
    SchemaList *tmp;

    for (tmp = cs->schemas; tmp != NULL; tmp = tmp->tail) {
	tmp->head->highlighted = TRUE;
    }
    repaint_schema_plotter(xg.canvas[SCHEMA], NULL, cs);
}

/*............................................................................*/

static void highlight_object_none(GtkWidget *button, ContentionScheduler *cs)
{
    ObjectList *tmp;

    for (tmp = cs->objects; tmp != NULL; tmp = tmp->tail) {
	tmp->head->highlighted = FALSE;
    }
    repaint_object_plotter(xg.canvas[OBJECT], NULL, cs);
}

static void highlight_object(GtkWidget *button, ContentionScheduler *cs)
{
    Object *tmp = gtk_object_get_data(GTK_OBJECT(button), "object");

    tmp->highlighted = !(tmp->highlighted);
    repaint_object_plotter(xg.canvas[OBJECT], NULL, cs);
}

static void highlight_object_all(GtkWidget *button, ContentionScheduler *cs)
{
    ObjectList *tmp;

    for (tmp = cs->objects; tmp != NULL; tmp = tmp->tail) {
	tmp->head->highlighted = TRUE;
    }
    repaint_object_plotter(xg.canvas[OBJECT], NULL, cs);
}

/*............................................................................*/

static void highlight_effector_none(GtkWidget *button, ContentionScheduler *cs)
{
    EffectorList *tmp;

    for (tmp = cs->effectors; tmp != NULL; tmp = tmp->tail) {
	tmp->head->highlighted = FALSE;
    }
    repaint_effector_plotter(xg.canvas[EFFECTOR], NULL, cs);
}

static void highlight_effector(GtkWidget *button, ContentionScheduler *cs)
{
    Effector *tmp = gtk_object_get_data(GTK_OBJECT(button), "effector");

    tmp->highlighted = !(tmp->highlighted);
    repaint_effector_plotter(xg.canvas[EFFECTOR], NULL, cs);
}

static void highlight_effector_all(GtkWidget *button, ContentionScheduler *cs)
{
    EffectorList *tmp;

    for (tmp = cs->effectors; tmp != NULL; tmp = tmp->tail) {
	tmp->head->highlighted = TRUE;
    }
    repaint_effector_plotter(xg.canvas[EFFECTOR], NULL, cs);
}

/******************************************************************************/

static GtkWidget *create_schema_submenu(ContentionScheduler *cs, SchemaLevel level)
{
    GtkWidget *menu, *mi;
    SchemaList *tmp;
    char buffer[64];

    menu = gtk_menu_new();

    for (tmp = cs->schemas; tmp != NULL; tmp = tmp->tail) {
        if (tmp->head->level == level) {
            g_snprintf(buffer, 64, "%c %s", tmp->head->highlighted?'*':' ', tmp->head->name);
            mi = gtk_menu_item_new_with_label(buffer);
            gtk_signal_connect(GTK_OBJECT(mi), "activate", GTK_SIGNAL_FUNC(highlight_schema), cs);
            gtk_object_set_data(GTK_OBJECT(mi), "schema", tmp->head);
            gtk_widget_show(mi);
            gtk_menu_append(GTK_MENU(menu), mi);
        }
    }
    return(menu);
}

static Boolean schemaMenuGenerate(GtkWidget *item, GdkEvent *event, ContentionScheduler *cs)
{
    GtkWidget *menu, *mi;

    gtk_menu_item_remove_submenu(GTK_MENU_ITEM(item));
    menu = gtk_menu_new();

    mi = gtk_menu_item_new_with_label("None");
    gtk_signal_connect(GTK_OBJECT(mi), "activate", GTK_SIGNAL_FUNC(highlight_schema_none), cs);
    gtk_widget_show(mi);
    gtk_menu_append(GTK_MENU(menu), mi);

    mi = gtk_menu_item_new_with_label("High Level");
    gtk_menu_item_set_submenu(GTK_MENU_ITEM(mi), create_schema_submenu(cs, HIGH));
    gtk_menu_append(GTK_MENU(menu), mi);
    gtk_widget_show(mi);

    mi = gtk_menu_item_new_with_label("Medium Level");
    gtk_menu_item_set_submenu(GTK_MENU_ITEM(mi), create_schema_submenu(cs, MEDIUM));
    gtk_menu_append(GTK_MENU(menu), mi);
    gtk_widget_show(mi);

    mi = gtk_menu_item_new_with_label("Low Level");
    gtk_menu_item_set_submenu(GTK_MENU_ITEM(mi), create_schema_submenu(cs, LOW));
    gtk_menu_append(GTK_MENU(menu), mi);
    gtk_widget_show(mi);

    mi = gtk_menu_item_new_with_label(" All");
    gtk_signal_connect(GTK_OBJECT(mi), "activate", GTK_SIGNAL_FUNC(highlight_schema_all), cs);
    gtk_widget_show(mi);
    gtk_menu_append(GTK_MENU(menu), mi);

    gtk_menu_item_set_submenu(GTK_MENU_ITEM(item), menu);

    /* Continue processing the event... */

    return(FALSE);
}

static Boolean objectMenuGenerate(GtkWidget *item, GdkEvent *event, ContentionScheduler *cs)
{
    GtkWidget *menu, *mi;
    ObjectList *tmp;
    char buffer[64];

    gtk_menu_item_remove_submenu(GTK_MENU_ITEM(item));
    menu = gtk_menu_new();

    mi = gtk_menu_item_new_with_label(" None");
    gtk_signal_connect(GTK_OBJECT(mi), "activate", GTK_SIGNAL_FUNC(highlight_object_none), cs);
    gtk_widget_show(mi);
    gtk_menu_append(GTK_MENU(menu), mi);

    for (tmp = cs->objects; tmp != NULL; tmp = tmp->tail) {
        g_snprintf(buffer, 64, "%c %s", tmp->head->highlighted?'*':' ', tmp->head->name);
        mi = gtk_menu_item_new_with_label(buffer);
        gtk_signal_connect(GTK_OBJECT(mi), "activate", GTK_SIGNAL_FUNC(highlight_object), cs);
        gtk_object_set_data(GTK_OBJECT(mi), "object", tmp->head);
        gtk_widget_show(mi);
        gtk_menu_append(GTK_MENU(menu), mi);
    }

    mi = gtk_menu_item_new_with_label(" All");
    gtk_signal_connect(GTK_OBJECT(mi), "activate", GTK_SIGNAL_FUNC(highlight_object_all), cs);
    gtk_widget_show(mi);
    gtk_menu_append(GTK_MENU(menu), mi);

    gtk_menu_item_set_submenu(GTK_MENU_ITEM(item), menu);

    /* Continue processing the event... */

    return(FALSE);
}

static Boolean effectorMenuGenerate(GtkWidget *item, GdkEvent *event, ContentionScheduler *cs)
{
    GtkWidget *menu, *mi;
    EffectorList *tmp;
    char buffer[64];

    gtk_menu_item_remove_submenu(GTK_MENU_ITEM(item));
    menu = gtk_menu_new();

    mi = gtk_menu_item_new_with_label(" None");
    gtk_signal_connect(GTK_OBJECT(mi), "activate", GTK_SIGNAL_FUNC(highlight_effector_none), cs);
    gtk_widget_show(mi);
    gtk_menu_append(GTK_MENU(menu), mi);

    for (tmp = cs->effectors; tmp != NULL; tmp = tmp->tail) {
        g_snprintf(buffer, 64, "%c %s", tmp->head->highlighted?'*':' ', tmp->head->name);
        mi = gtk_menu_item_new_with_label(buffer);
        gtk_signal_connect(GTK_OBJECT(mi), "activate", GTK_SIGNAL_FUNC(highlight_effector), cs);
        gtk_object_set_data(GTK_OBJECT(mi), "effector", tmp->head);
        gtk_widget_show(mi);
        gtk_menu_append(GTK_MENU(menu), mi);
    }

    mi = gtk_menu_item_new_with_label(" All");
    gtk_signal_connect(GTK_OBJECT(mi), "activate", GTK_SIGNAL_FUNC(highlight_effector_all), cs);
    gtk_widget_show(mi);
    gtk_menu_append(GTK_MENU(menu), mi);

    gtk_menu_item_set_submenu(GTK_MENU_ITEM(item), menu);

    /* Continue processing the event... */

    return(FALSE);
}

/*----------------------------------------------------------------------------*/

static GtkWidget *create_menu(char *label, Boolean menu_generate(), void *data)
{
    GtkWidget *menu_bar = gtk_menu_bar_new();
    GtkWidget *menu = gtk_menu_new();
    GtkWidget *item = gtk_menu_item_new_with_label(label);

    gtk_menu_item_set_submenu(GTK_MENU_ITEM(item), menu);
    gtk_signal_connect(GTK_OBJECT(item), "enter_notify_event", GTK_SIGNAL_FUNC(menu_generate), data);
    gtk_widget_show(item);

    gtk_menu_bar_append(GTK_MENU_BAR(menu_bar), item);
    gtk_widget_show(menu_bar);
    return(menu_bar);
}

/******************************************************************************/

static void select_object_role_callback(GtkWidget *item, ContentionScheduler *cs)
{
    int role = (int) gtk_object_get_data(GTK_OBJECT(item), "role");

    gtk_object_set_data(GTK_OBJECT(xg.canvas[OBJECT]), "role", (void *)role);
    repaint_object_plotter(xg.canvas[OBJECT], NULL, cs);
}

/*----------------------------------------------------------------------------*/

static void make_graph_widgets(ContentionScheduler *cs, GtkWidget *vbox)
{
    GtkWidget   *hbox, *menu, *mi, *tmp, *label, *panes, *panel[3];
    GSList      *rb_grp;
    int          i;

    panes = gtk_vpaned_new();
    gtk_box_pack_start(GTK_BOX(vbox), panes, TRUE, TRUE, 0);
    gtk_widget_show(panes);

    panel[0] = gtk_vbox_new(FALSE, 5);
    gtk_paned_add1(GTK_PANED(panes), panel[0]);
    gtk_widget_show(panel[0]);

    tmp = gtk_vpaned_new();
    gtk_paned_add2(GTK_PANED(panes), tmp);
    gtk_widget_show(tmp);

    panel[1] = gtk_vbox_new(FALSE, 5);
    gtk_paned_add1(GTK_PANED(tmp), panel[1]);
    gtk_widget_show(panel[1]);

    panel[2] = gtk_vbox_new(FALSE, 5);
    gtk_paned_add2(GTK_PANED(tmp), panel[2]);
    gtk_widget_show(panel[2]);

    gtk_widget_set_usize(panel[0], 0, 200);
    gtk_widget_set_usize(panel[1], 0, 200);
    gtk_widget_set_usize(panel[2], 0, 200);

    /* Schema menu: */

    hbox = gtk_hbox_new(FALSE, 4);
    gtk_widget_set_usize(hbox, 0, 26);
    gtk_box_pack_start(GTK_BOX(panel[0]), hbox, FALSE, FALSE, 0);
    gtk_widget_show(hbox);

    menu = create_menu("Highlight", schemaMenuGenerate, cs);
    gtk_box_pack_start(GTK_BOX(hbox), GTK_WIDGET(menu), FALSE, FALSE, 0);

    /* A message stating the current cycle number: */

    xg.cycle_widget = gtk_label_new("      0 ");
    gtk_widget_show(xg.cycle_widget);
    gtk_box_pack_end(GTK_BOX(hbox), xg.cycle_widget, FALSE, FALSE, 0);

    label = gtk_label_new("Cycle:");
    gtk_widget_show(label);
    gtk_box_pack_end(GTK_BOX(hbox), label, FALSE, FALSE, 0);

    /* Schema graph */

    xg.canvas[SCHEMA] = create_graph(panel[0], canvasHeight, repaint_schema_plotter, "Schema Activations", cs);
    gtk_signal_connect(GTK_OBJECT(xg.canvas[SCHEMA]), "button_press_event", GTK_SIGNAL_FUNC(click_on_schema_graph), cs);

    /* Object menu */

    hbox = gtk_hbox_new(FALSE, 4);
    gtk_widget_set_usize(hbox, 0, 26);
    gtk_box_pack_start(GTK_BOX(panel[1]), hbox, FALSE, FALSE, 0);
    gtk_widget_show(hbox);

    menu = create_menu("Highlight", objectMenuGenerate, cs);
    gtk_box_pack_start(GTK_BOX(hbox), GTK_WIDGET(menu), FALSE, FALSE, 0);

    /* The object role menu: */

    label = gtk_label_new("   Role:");
    gtk_widget_show(label);
    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);

    rb_grp = NULL;
    for (i = 0; i < MaxRoles-1; i++) {
        mi = gtk_radio_button_new_with_label(rb_grp, roleName[i]);
        gtk_box_pack_start(GTK_BOX(hbox), GTK_WIDGET(mi), FALSE, FALSE, 0);
        gtk_signal_connect(GTK_OBJECT(mi), "toggled", GTK_SIGNAL_FUNC(select_object_role_callback), cs);
        gtk_object_set_data(GTK_OBJECT(mi), "role", (void *)i);
        gtk_widget_show(mi);
        rb_grp = gtk_radio_button_group(GTK_RADIO_BUTTON(mi));
    }

    /* Object graph */

    xg.canvas[OBJECT] = create_graph(panel[1], canvasHeight, repaint_object_plotter, "Object Representation Activations", cs);
    gtk_object_set_data(GTK_OBJECT(xg.canvas[OBJECT]), "role", (void *)SOURCE);
    gtk_signal_connect(GTK_OBJECT(xg.canvas[OBJECT]), "button_press_event", GTK_SIGNAL_FUNC(click_on_object_graph), cs);

    /* Effector menu */

    hbox = gtk_hbox_new(FALSE, 4);
    gtk_widget_set_usize(hbox, 0, 26);
    gtk_box_pack_start(GTK_BOX(panel[2]), hbox, FALSE, FALSE, 0);
    gtk_widget_show(hbox);

    menu = create_menu("Highlight", effectorMenuGenerate, cs);
    gtk_box_pack_start(GTK_BOX(hbox), GTK_WIDGET(menu), FALSE, FALSE, 0);

    /* Effector graph */

    xg.canvas[EFFECTOR] = create_graph(panel[2], canvasHeight, repaint_effector_plotter, "Effector Activations", cs);
    gtk_signal_connect(GTK_OBJECT(xg.canvas[EFFECTOR]), "button_press_event", GTK_SIGNAL_FUNC(click_on_effector_graph), cs);
}

/*----------------------------------------------------------------------------*/

static Schema *locate_schema_by_name(ContentionScheduler *cs, char *name)
{
    SchemaList *sl = cs->schemas;
    while ((sl != NULL) && (strcmp(sl->head->name, name) != 0)) {
        sl = sl->tail;
    }
    return((sl == NULL) ? NULL : sl->head);
}

static Object *locate_object_by_name(ContentionScheduler *cs, char *name)
{
    ObjectList *ol = cs->objects;
    while ((ol != NULL) && (strcmp(ol->head->name, name) != 0)) {
        ol = ol->tail;
    }
    return((ol == NULL) ? NULL : ol->head);
}

static Effector *locate_effector_by_name(ContentionScheduler *cs, char *name)
{
    EffectorList *el = cs->effectors;
    while ((el != NULL) && (strcmp(el->head->name, name) != 0)) {
        el = el->tail;
    }
    return((el == NULL) ? NULL : el->head);
}

static void sort_cl_log(GtkWidget *cl, int col, void *dummy)
{
    static Boolean order[9] = {TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE};
    GtkSortType sort_type = (order[col] ? GTK_SORT_ASCENDING : GTK_SORT_DESCENDING);

    order[col] = !order[col];

    gtk_clist_set_sort_type(GTK_CLIST(cl), sort_type);
    gtk_clist_set_sort_column(GTK_CLIST(cl), col);
    gtk_clist_set_compare_func(GTK_CLIST(cl), NULL);
    gtk_clist_sort(GTK_CLIST(cl));
}

static void analyse_excitation(GtkWidget *cl, int row, int column, GdkEventButton *event, ContentionScheduler *cs)
{
    extern void print_excitation_of_schema_by_objects(FILE *fp, ContentionScheduler *cs, Schema *s);
    extern void print_excitation_of_object_by_schemas(FILE *fp, ContentionScheduler *cs, Object *o);
    extern void print_excitation_of_effector_by_schemas(FILE *fp, ContentionScheduler *cs, Effector *e);

    char *label;
    Schema *s;
    Object *o;
    Effector *e;

    if (gtk_clist_get_text(GTK_CLIST(cl), row, 0, &label)) {
        fprintf(stdout, "Analysis of excitation for %s on cycle %d:\n", label, cs->cycle);
        if ((s = locate_schema_by_name(cs, label)) != NULL) {
            print_excitation_of_schema_by_objects(stdout, cs, s);
        }
        else if ((o = locate_object_by_name(cs, label)) != NULL) {
            print_excitation_of_object_by_schemas(stdout, cs, o);
        }
        else if ((e = locate_effector_by_name(cs, label)) != NULL) {
            print_excitation_of_effector_by_schemas(stdout, cs, e);
        }
        else {
            fprintf(stdout, "Aborting analysis: Cannot categorise %s\n", label);
        }
    }
    else {
        fprintf(stdout, "Excitation analysis: Cannot locate node name\n");
    }
}

static GtkWidget *make_cl_log(ContentionScheduler *cs, GtkWidget *vbox)
{
    GtkWidget *cl, *sw;
    char *col0 = " Name";
    char *col1 = " ";
    char *col2 = " Before";
    char *col3 = " Self";
    char *col4 = " Lat.";
    char *col5 = " Int.";
    char *col6 = " Ext.";
    char *col7 = " Net.";
    char *col8 = " After";
    char *col9 = " Notes";

    char *row_text[10] = {col0, col1, col2, col3, col4, col5, col6, col7, col8, col9};

    sw = gtk_scrolled_window_new(NULL, NULL);
    gtk_container_add(GTK_CONTAINER(vbox), sw);
    gtk_widget_show(sw);

    cl = gtk_clist_new_with_titles(10, row_text);
    gtk_clist_set_column_width(GTK_CLIST(cl), 0, 150);
    gtk_clist_set_column_width(GTK_CLIST(cl), 1, 10);
    gtk_clist_set_column_justification(GTK_CLIST(cl), 1, GTK_JUSTIFY_CENTER);
    gtk_clist_set_column_width(GTK_CLIST(cl), 2, 45);
    gtk_clist_set_column_justification(GTK_CLIST(cl), 2, GTK_JUSTIFY_CENTER);
    gtk_clist_set_column_width(GTK_CLIST(cl), 3, 45);
    gtk_clist_set_column_justification(GTK_CLIST(cl), 3, GTK_JUSTIFY_CENTER);
    gtk_clist_set_column_width(GTK_CLIST(cl), 4, 45);
    gtk_clist_set_column_justification(GTK_CLIST(cl), 4, GTK_JUSTIFY_CENTER);
    gtk_clist_set_column_width(GTK_CLIST(cl), 5, 45);
    gtk_clist_set_column_justification(GTK_CLIST(cl), 5, GTK_JUSTIFY_CENTER);
    gtk_clist_set_column_width(GTK_CLIST(cl), 6, 45);
    gtk_clist_set_column_justification(GTK_CLIST(cl), 6, GTK_JUSTIFY_CENTER);
    gtk_clist_set_column_width(GTK_CLIST(cl), 7, 45);
    gtk_clist_set_column_justification(GTK_CLIST(cl), 7, GTK_JUSTIFY_CENTER);
    gtk_clist_set_column_width(GTK_CLIST(cl), 8, 65);
    gtk_clist_set_column_justification(GTK_CLIST(cl), 8, GTK_JUSTIFY_CENTER);

    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
    gtk_signal_connect(GTK_OBJECT(cl), "click_column", GTK_SIGNAL_FUNC(sort_cl_log), NULL);
    gtk_signal_connect(GTK_OBJECT(cl), "select_row", GTK_SIGNAL_FUNC(analyse_excitation), cs);
    gtk_container_add(GTK_CONTAINER(sw), cl);
    gtk_widget_show(cl);
    return(cl);
}

static void record_cl_activity(GtkWidget *cl, char *name, Boolean selected, double before, double s, double l, double p, double e, double in, double after, char *note)
{
    char col1[2], col2[10], col3[10], col4[10], col5[10], col6[10], col7[10], col8[10];
    char *row_text[10] = {name, col1, col2, col3, col4, col5, col6, col7, col8, note};

    g_snprintf(col1, 2, "%s", (selected ? "*" : " "));
    g_snprintf(col2, 10, "%+5.3f", before);
    g_snprintf(col3, 10, "%+5.3f", s);
    g_snprintf(col4, 10, "%+5.3f", l);
    g_snprintf(col5, 10, "%+5.3f", p);
    g_snprintf(col6, 10, "%+5.3f", e);
    g_snprintf(col7, 10, "%+5.3f", in);
    g_snprintf(col8, 10, "%7.5f", after);

    gtk_clist_append(GTK_CLIST(cl), row_text);
}

void record_schema_activity(ContentionScheduler *cs, char *name, Boolean selected, double before, double s, double l, double p, double e, double in, double after, char *note)
{
    record_cl_activity(cs->cl_schemas, name, selected, before, s, l, p, e, in, after, note);
}

void record_object_activity(ContentionScheduler *cs, char *name, int r, Boolean selected, double before, double s, double l, double p, double e, double in, double after, char *note)
{
    record_cl_activity(cs->cl_objects[r], name, selected, before, s, l, p, e, in, after, note);
}

void record_effector_activity(ContentionScheduler *cs, char *name, Boolean selected, double before, double s, double l, double p, double e, double in, double after, char *note)
{
    record_cl_activity(cs->cl_effectors, name, selected, before, s, l, p, e, in, after, note);
}

/******************************************************************************/

extern void list_schema_activation(ContentionScheduler *cs, Schema *this);
extern void list_object_activation(ContentionScheduler *cs, Object *this, ObjectRole role);
extern void list_effector_activation(ContentionScheduler *cs, Effector *this);

/*----------------------------------------------------------------------------*/

static void listSchemaActivations(ContentionScheduler *cs)
{
    SchemaList *ss;

    gtk_clist_freeze(GTK_CLIST(cs->cl_schemas));
    gtk_clist_clear(GTK_CLIST(cs->cl_schemas));
    for (ss = cs->schemas; ss != NULL; ss = ss->tail) {
	list_schema_activation(cs, ss->head);
    }
    gtk_clist_thaw(GTK_CLIST(cs->cl_schemas));
}

/*----------------------------------------------------------------------------*/

static void listObjectActivations(ContentionScheduler *cs, ObjectRole r)
{
    ObjectList *os;

    gtk_clist_freeze(GTK_CLIST(cs->cl_objects[r]));
    gtk_clist_clear(GTK_CLIST(cs->cl_objects[r]));
    for (os = cs->objects; os != NULL; os = os->tail) {
        list_object_activation(cs, os->head, r);
    }
    gtk_clist_thaw(GTK_CLIST(cs->cl_objects[r]));
}

/*----------------------------------------------------------------------------*/

static void listEffectorActivations(ContentionScheduler *cs)
{
    EffectorList *es;

    gtk_clist_freeze(GTK_CLIST(cs->cl_effectors));
    gtk_clist_clear(GTK_CLIST(cs->cl_effectors));
    for (es = cs->effectors; es != NULL; es = es->tail) {
	list_effector_activation(cs, es->head);
    }
    gtk_clist_thaw(GTK_CLIST(cs->cl_effectors));
}

/*----------------------------------------------------------------------------*/

static Boolean update_cl_schemas(GtkWidget *clist, GdkEventExpose *event, ContentionScheduler *cs)
{
// fprintf(stdout, "Schema list updated on cycle %d\n", cs->cycle);
    cs->cycle--;
    listSchemaActivations(cs);
    cs->cycle++;
    return(FALSE);
}

static Boolean update_cl_objects(GtkWidget *clist, GdkEventExpose *event, ContentionScheduler *cs)
{
    ObjectRole r;

// fprintf(stdout, "Object list updated on cycle %d\n", cs->cycle);
    cs->cycle--;
    for (r = 0; r < MaxRoles-1; r++) {
        if (cs->cl_objects[r] == clist) {
            listObjectActivations(cs, r);
        }
    }
    cs->cycle++;
    return(FALSE);
}

static Boolean update_cl_effectors(GtkWidget *clist, GdkEventExpose *event, ContentionScheduler *cs)
{
// fprintf(stdout, "Effector list updated on cycle %d\n", cs->cycle);
    cs->cycle--;
    listEffectorActivations(cs);
    cs->cycle++;
    return(FALSE);
}

void post_cycle_trace(ContentionScheduler *cs)
{
    int i;

    if (GTK_WIDGET_MAPPED(cs->cl_schemas)) {
        update_cl_schemas(cs->cl_schemas, NULL, cs);
    }
    for (i = 0; i < MaxRoles-1; i++) {
        if (GTK_WIDGET_MAPPED(cs->cl_objects[i])) {
            update_cl_objects(cs->cl_objects[i], NULL, cs);
        }
    }
    if (GTK_WIDGET_MAPPED(cs->cl_effectors)) {
        update_cl_effectors(cs->cl_effectors, NULL, cs);
    }
}

/*----------------------------------------------------------------------------*/

static void make_action_log_widgets(ContentionScheduler *cs, GtkWidget *vbox)
{
    GtkWidget *cl, *sw;
    char *col0 = " Cycle";
    char *col1 = " Selections";
    char *col2 = " Action";
    char *col3 = " Messages";
    char *row_text[4] = {col0, col1, col2, col3};

    sw = gtk_scrolled_window_new(NULL, NULL);
    gtk_container_add(GTK_CONTAINER(vbox), sw);
    gtk_widget_show(sw);

    cl = gtk_clist_new_with_titles(4, row_text);
    gtk_clist_set_column_width(GTK_CLIST(cl), 0, 50);
    gtk_clist_set_column_width(GTK_CLIST(cl), 1, 200);
    gtk_clist_set_column_width(GTK_CLIST(cl), 2, 80);

    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
    gtk_container_add(GTK_CONTAINER(sw), cl);
    gtk_widget_show(cl);

    cs->cl_actions = cl;
}

void action_log_add_action(ContentionScheduler *cs, char *act, char *message)
{
    if (cs->cl_actions != NULL) {
        char col0[10];
        char col2[80];
        char col3[80];
        char *row_text[4] = {col0, "", col2, col3};

        g_snprintf(col0, 10, "%d", cs->cycle);

        g_snprintf(col2, 80, "%s", act);
        g_snprintf(col3, 80, "%s", message);
        gtk_clist_append(GTK_CLIST(cs->cl_actions), row_text);
    }
}

void action_log_add_warning(ContentionScheduler *cs, char *type, char *warning)
{
    if (cs->cl_actions != NULL) {
        char col0[10];
        char col1[80];
        char col3[80];
        char *row_text[4] = {col0, col1, "", col3};

        g_snprintf(col0, 10, "%d", cs->cycle);
        g_snprintf(col1, 80, "%s", type);
        g_snprintf(col3, 80, "%s", warning);
        gtk_clist_append(GTK_CLIST(cs->cl_actions), row_text);
    }
}

void action_log_add_selection(ContentionScheduler *cs, char *schema_name, Boolean on)
{
    if (cs->cl_actions != NULL) {
        char col0[10];
        char col1[80];
        char *row_text[4] = {col0, col1, "", ""};

        g_snprintf(col0, 10, "%d", cs->cycle);
        g_snprintf(col1, 80, "%c%s", (on ? '+' : '-'), schema_name);
        gtk_clist_append(GTK_CLIST(cs->cl_actions), row_text);
    }
}

/*----------------------------------------------------------------------------*/

static void make_analysis_widgets(ContentionScheduler *cs, GtkWidget *vbox)
{
    GtkWidget *cl, *sw;
    char *col0 = " Statistic";
    char *col1 = " Value";
    char *row_text[2] = {col0, col1};

    sw = gtk_scrolled_window_new(NULL, NULL);
    gtk_container_add(GTK_CONTAINER(vbox), sw);
    gtk_widget_show(sw);

    cl = gtk_clist_new_with_titles(2, row_text);
    gtk_clist_set_column_width(GTK_CLIST(cl), 0, 150);
    gtk_clist_set_column_width(GTK_CLIST(cl), 1, 200);

    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
    gtk_container_add(GTK_CONTAINER(sw), cl);
    gtk_widget_show(cl);

    cs->cl_analysis = cl;
}

static void analysis_add_entry(ContentionScheduler *cs, char *stat, char *val)
{
    if (cs->cl_analysis != NULL) {
        char *row_text[2] = {stat, val};
        gtk_clist_append(GTK_CLIST(cs->cl_analysis), row_text);
    }
}

/*----------------------------------------------------------------------------*/

static void populate_confusability_matrix(ContentionScheduler *cs)
{
    ObjectList *os, *os2;
    GList *children;
    GtkWidget *cl;
    int n = 0, i = 0;
    double ol;
    char **row_text;

    /* First clear out any old matrix: */

    children = gtk_container_children(GTK_CONTAINER(xg.conf_matrix));
    while (children != NULL) {
        GList *tmp = g_list_next(children);
        gtk_container_remove(GTK_CONTAINER(xg.conf_matrix), (GtkWidget *)children->data);        children = tmp;
    }

    /* Count the objects (so we know how big to make it): */

    for (os = cs->objects; os; os = os->tail) {
        n++;
    }

    /* Now construct the row labels, the matrix and populate: */

    if ((row_text = (char **)malloc((n + 1) * sizeof(char *))) != NULL) {
        row_text[0] = ""; i = 0;
        for (os = cs->objects; os; os = os->tail) {
            row_text[++i] = os->head->name;
        }        
        cl = gtk_clist_new_with_titles(n+1, row_text);
        gtk_clist_set_column_width(GTK_CLIST(cl), 0, 150);

        for (i = 0; i < n; i++) {
            gtk_clist_set_column_width(GTK_CLIST(cl), i+1, 100);
            gtk_clist_set_column_justification(GTK_CLIST(cl), i+1, GTK_JUSTIFY_CENTER);
            row_text[i+1] = (char *)malloc(12 * sizeof(char));
        }

        for (os = cs->objects; os; os = os->tail)    {
            row_text[0] = os->head->name; i = 0;
            for (os2 = cs->objects; os2; os2 = os2->tail) {
                if ((ol = feature_overlap(os->head->features, os2->head->features)) > 0.0) {
                    sprintf(row_text[++i], "%6.4f", ol);
                }
                else {
                    strcpy(row_text[++i], "");
                }
            }
            gtk_clist_append(GTK_CLIST(cl), row_text);
        }        

        gtk_container_add(GTK_CONTAINER(xg.conf_matrix), cl);
        gtk_widget_show(cl);
    }
}

static void make_confusability_widgets(ContentionScheduler *cs, GtkWidget *vbox)
{
    xg.conf_matrix = gtk_scrolled_window_new(NULL, NULL);
    gtk_container_add(GTK_CONTAINER(vbox), xg.conf_matrix);
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(xg.conf_matrix), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
    gtk_widget_show(xg.conf_matrix);

    populate_confusability_matrix(cs);
}

/*----------------------------------------------------------------------------*/

static void update_nb_page(GtkWidget *nb, GtkNotebookPage *page, gint page_num, ContentionScheduler *cs)
{
    if (page_num == 3) {                          // Switch to the schema list
        update_cl_schemas(NULL, NULL, cs);
    }
    else if (page_num == 8) {                      // Switch to the effector list
        update_cl_effectors(NULL, NULL, cs);
    }
    else if ((page_num > 3) && (page_num < 8)) {  // Switch to an object list
        update_cl_objects(NULL, NULL, cs);
    }
}

/*----------------------------------------------------------------------------*/

GtkWidget *make_widgets(ContentionScheduler *cs)
{
    GtkWidget *frame, *vbox, *hbox, *button, *nb;
    int i;

    xg.cs = cs;

    frame = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_container_set_border_width(GTK_CONTAINER(frame), 4);
    gtk_window_set_policy(GTK_WINDOW(frame), TRUE, TRUE, FALSE);
    gtk_widget_set_usize(frame, frameWidth, 0);
    gtk_window_set_title(GTK_WINDOW(frame), "Contention Scheduling: Generalised Version 1.0");
    gtk_signal_connect(GTK_OBJECT(frame), "destroy", GTK_SIGNAL_FUNC(exit_xvcs), 0);
    gtk_window_set_position(GTK_WINDOW(frame), GTK_WIN_POS_CENTER);
    gtk_widget_realize(frame);

    create_gcs(frame);

    vbox = gtk_vbox_new(FALSE, 4);
    gtk_container_add(GTK_CONTAINER(frame), vbox);
    gtk_widget_show(vbox);

    /* The top box of buttons: */

    hbox = gtk_hbox_new(TRUE, 4);
    gtk_widget_set_usize(hbox, 0, 26);
    gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
    gtk_widget_show(hbox);

    /* Row 1: The control panel  */

    create_button(hbox, "Reinitialise", reinitialise_callback, cs);
    create_button(hbox, "Run", run_callback, cs);
    button = create_button(hbox, "1 Step", step_callback, cs);
    gtk_object_set_data(GTK_OBJECT(button), "steps", (void *)1);
    button = create_button(hbox, "10 Steps", step_callback, cs);
    gtk_object_set_data(GTK_OBJECT(button), "steps", (void *)10);
    button = create_button(hbox, "50 Steps ", step_callback, cs);
    gtk_object_set_data(GTK_OBJECT(button), "steps", (void *)50);
    button = create_button(hbox, "250 Steps ", step_callback, cs);
    gtk_object_set_data(GTK_OBJECT(button), "steps", (void *)250);
    button = create_button(hbox, "1000 Steps ", step_callback, cs);
    gtk_object_set_data(GTK_OBJECT(button), "steps", (void *)1000);
    if (FALSE) {
        create_button(hbox, "Dump Xgr", dump_callback, cs);
    }
    create_button(hbox, "Quit", quit_callback, frame);

    /* The notebook containing graphs etc.: */

    nb = gtk_notebook_new();
    gtk_box_pack_start(GTK_BOX(vbox), nb, TRUE, TRUE, 0);
    gtk_widget_show(nb);

    /* The parameters page of the notebook: */

    vbox = gtk_hbox_new(FALSE, 4);
    gtk_container_set_border_width(GTK_CONTAINER(vbox), 4);
    make_params_widgets(cs, vbox);
    gtk_widget_show(vbox);
    gtk_notebook_append_page(GTK_NOTEBOOK(nb), vbox, gtk_label_new("Parameters"));

    /* The graphs page of the notebook: */

    vbox = gtk_vbox_new(FALSE, 4);
    gtk_container_set_border_width(GTK_CONTAINER(vbox), 4);
    make_graph_widgets(cs, vbox);
    gtk_widget_show(vbox);
    gtk_notebook_append_page(GTK_NOTEBOOK(nb), vbox, gtk_label_new("Graphs"));

    /* The "actions" page of the notebook: */

    vbox = gtk_vbox_new(FALSE, 4);
    gtk_container_set_border_width(GTK_CONTAINER(vbox), 4);
    make_action_log_widgets(cs, vbox);
    gtk_widget_show(vbox);
    gtk_notebook_append_page(GTK_NOTEBOOK(nb), vbox, gtk_label_new("Action Log"));

    /* The "schemas" page of the notebook: */

    vbox = gtk_vbox_new(FALSE, 4);
    gtk_container_set_border_width(GTK_CONTAINER(vbox), 4);
    cs->cl_schemas = make_cl_log(cs, vbox);
    gtk_widget_show(vbox);
    gtk_notebook_append_page(GTK_NOTEBOOK(nb), vbox, gtk_label_new("Schemas"));
    gtk_signal_connect(GTK_OBJECT(cs->cl_schemas), "expose_event", GTK_SIGNAL_FUNC(update_cl_schemas), cs);

    /* The "object" pages of the notebook: */

    for (i = 0; i < MaxRoles-1; i++) {
        vbox = gtk_vbox_new(FALSE, 4);
        gtk_container_set_border_width(GTK_CONTAINER(vbox), 4);
        cs->cl_objects[i] = make_cl_log(cs, vbox);
        gtk_widget_show(vbox);
        gtk_notebook_append_page(GTK_NOTEBOOK(nb), vbox, gtk_label_new(roleName[i]));
        gtk_signal_connect(GTK_OBJECT(cs->cl_objects[i]), "expose_event", GTK_SIGNAL_FUNC(update_cl_objects), cs);
    } 

    /* The "effectors" page of the notebook: */

    vbox = gtk_vbox_new(FALSE, 4);
    gtk_container_set_border_width(GTK_CONTAINER(vbox), 4);
    cs->cl_effectors = make_cl_log(cs, vbox);
    gtk_widget_show(vbox);
    gtk_notebook_append_page(GTK_NOTEBOOK(nb), vbox, gtk_label_new("Effectors"));
    gtk_signal_connect(GTK_OBJECT(cs->cl_effectors), "expose_event", GTK_SIGNAL_FUNC(update_cl_effectors), cs);

    /* The "analysis" page of the notebook: */

    vbox = gtk_vbox_new(FALSE, 4);
    gtk_container_set_border_width(GTK_CONTAINER(vbox), 4);
    make_analysis_widgets(cs, vbox);
    gtk_widget_show(vbox);
    gtk_notebook_append_page(GTK_NOTEBOOK(nb), vbox, gtk_label_new("Analyses"));

    /* The "confusability" page of the notebook: */

    vbox = gtk_vbox_new(FALSE, 4);
    gtk_container_set_border_width(GTK_CONTAINER(vbox), 4);
    make_confusability_widgets(cs, vbox);
    gtk_widget_show(vbox);
    gtk_notebook_append_page(GTK_NOTEBOOK(nb), vbox, gtk_label_new("Similarity"));

    /* The callback to ensure schema/object/effector activations are updated: */
    gtk_signal_connect(GTK_OBJECT(nb), "switch_page", GTK_SIGNAL_FUNC(update_nb_page), cs);

    /* The "busy" cursor: */

    xg.busy = gdk_cursor_new(GDK_WATCH);
    xg.frame = frame;

    return(frame);
}

/******************************************************************************/

void update_cycle_display(ContentionScheduler *cs)
{
    static char count[7];

    g_snprintf(count, 7, "%5d", cs->cycle);
    gtk_label_set_text(GTK_LABEL(xg.cycle_widget), count);
}

/******************************************************************************/

void attach_data(ContentionScheduler *cs)
{
    repaint_schema_plotter(xg.canvas[SCHEMA], NULL, cs);
    repaint_object_plotter(xg.canvas[OBJECT], NULL, cs);
    repaint_effector_plotter(xg.canvas[EFFECTOR], NULL, cs);
}

/******************************************************************************/

void process_gtk_events()
{
    while (gtk_events_pending()) {
        gtk_main_iteration();
    }
}

/******************************************************************************/
/******************************************************************************/
/******************************************************************************/
